@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1997-2020, University of Amsterdam
+    Copyright (c)  1997-2021, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -47,6 +48,14 @@
     prolog:deprecated//1,	    % Deprecated features
     prolog:message_location//1,     % (File) location of error messages
     prolog:message_line_element/2.  % Extend printing
+:- '$hide'((
+    prolog:message//1,
+    prolog:error_message//1,
+    prolog:message_context//1,
+    prolog:deprecated//1,
+    prolog:message_location//1,
+    prolog:message_line_element/2)).
+
 :- discontiguous
     prolog_message/3.
 
@@ -60,17 +69,19 @@
 %   Translate a message Term into message lines. The produced lines
 %   is a list of
 %
-%       * nl
-%       Emit a newline
-%       * Fmt-Args
-%       Emit the result of format(Fmt, Args)
-%       * Fmt
-%       Emit the result of format(Fmt)
-%       * flush
-%       Used only as last element of the list.   Simply flush the
-%       output instead of producing a final newline.
-%       * at_same_line
-%       Start the messages at the same line (instead of using ~N)
+%       - nl
+%         Emit a newline
+%       - Fmt-Args
+%         Emit the result of format(Fmt, Args)
+%       - Fmt
+%         Emit the result of format(Fmt)
+%       - ansi(Code, Fmt, Args)
+%         Use ansi_format/3 for color output.
+%       - flush
+%         Used only as last element of the list.   Simply flush the
+%         output instead of producing a final newline.
+%       - at_same_line
+%         Start the messages at the same line (instead of using ~N)
 
 translate_message(Term) -->
     translate_message2(Term),
@@ -127,6 +138,8 @@ term_message(Term) -->
 term_message(Term) -->
     [ 'Unknown error term: ~p'-[Term] ].
 
+iso_message(resource_error(c_stack)) -->
+    out_of_c_stack.
 iso_message(resource_error(Missing)) -->
     [ 'Not enough resources: ~w'-[Missing] ].
 iso_message(type_error(evaluable, Actual)) -->
@@ -156,6 +169,8 @@ iso_message(existence_error(procedure, Proc)) -->
     unknown_proc_msg(Proc).
 iso_message(existence_error(answer_variable, Var)) -->
     [ '$~w was not bound by a previous query'-[Var] ].
+iso_message(existence_error(matching_rule, Goal)) -->
+    [ 'No rule matches ~p'-[Goal] ].
 iso_message(existence_error(Type, Object)) -->
     [ '~w `~p'' does not exist'-[Type, Object] ].
 iso_message(existence_error(Type, Object, In)) --> % not ISO
@@ -220,6 +235,10 @@ unknown_proc_msg((:-)/2) -->
     !,
     [nl, '  Rules must be loaded from a file'],
     faq('ToplevelMode').
+unknown_proc_msg((=>)/2) -->
+    !,
+    [nl, '  Rules must be loaded from a file'],
+    faq('ToplevelMode').
 unknown_proc_msg((:-)/1) -->
     !,
     [nl, '  Directives must be loaded from a file'],
@@ -236,6 +255,16 @@ unknown_proc_msg(Proc) -->
     ;   []
     ).
 
+dependency_error(shared(Shared), private(Private)) -->
+    [ 'Shared table for ~p may not depend on private ~p'-[Shared, Private] ].
+dependency_error(Dep, monotonic(On)) -->
+    { '$pi_head'(PI, Dep),
+      '$pi_head'(MPI, On)
+    },
+    [ 'Dependent ~p on monotonic predicate ~p is not monotonic or incremental'-
+      [PI, MPI]
+    ].
+
 faq(Page) -->
     [nl, '  See FAQ at https://www.swi-prolog.org/FAQ/', Page, '.txt' ].
 
@@ -370,8 +399,8 @@ swi_message(thread_error(TID, false)) -->
 swi_message(thread_error(TID, exception(Error))) -->
     [ 'Thread ~p died abnormally:'-[TID], nl ],
     translate_message(Error).
-swi_message(idg_dependency_error(Shared, Private)) -->
-    [ 'Shared table for ~p may not depend on private ~p'-[Shared, Private] ].
+swi_message(dependency_error(Tabled, DependsOn)) -->
+    dependency_error(Tabled, DependsOn).
 swi_message(shell(execute, Cmd)) -->
     [ 'Could not execute `~w'''-[Cmd] ].
 swi_message(shell(signal(Sig), Cmd)) -->
@@ -428,8 +457,32 @@ swi_message(initialization_error(failed, Goal, File:Line)) -->
 swi_message(initialization_error(Error, Goal, File:Line)) -->
     [ '~w:~w: ~p '-[File, Line, Goal] ],
     translate_message(Error).
+swi_message(determinism_error(PI, det, Found, property)) -->
+    (   { '$pi_head'(user:PI, Head),
+          predicate_property(Head, det)
+        }
+    ->  [ 'Deterministic procedure ~p'-[PI] ]
+    ;   [ 'Procedure ~p called from a deterministic procedure'-[PI] ]
+    ),
+    det_error(Found).
+swi_message(determinism_error(PI, det, fail, guard)) -->
+    [ 'Procedure ~p failed after $-guard'-[PI] ].
+swi_message(determinism_error(PI, det, fail, guard_in_caller)) -->
+    [ 'Procedure ~p failed after $-guard in caller'-[PI] ].
+swi_message(determinism_error(Goal, det, fail, goal)) -->
+    [ 'Goal ~p failed'-[Goal] ].
+swi_message(determinism_error(Goal, det, nondet, goal)) -->
+    [ 'Goal ~p succeeded with a choice point'-[Goal] ].
 swi_message(qlf_format_error(File, Message)) -->
     [ '~w: Invalid QLF file: ~w'-[File, Message] ].
+swi_message(goal_expansion_error(bound, Term)) -->
+    [ 'Goal expansion bound a variable to ~p'-[Term] ].
+
+det_error(nondet) -->
+    [ ' succeeded with a choicepoint'- [] ].
+det_error(fail) -->
+    [ ' failed'- [] ].
+
 
 cond_location(File:Line) -->
     { file_base_name(File, Base) },
@@ -940,6 +993,29 @@ resolve_overflow(enlarge) -->
       '?- set_prolog_flag(stack_limit, ~I). to double the limit.'-[NewLimit]
     ].
 
+%!  out_of_c_stack
+%
+%   The thread's C-stack limit was exceeded. Give  some advice on how to
+%   resolve this.
+
+out_of_c_stack -->
+    { statistics(c_stack, Limit), Limit > 0 },
+    !,
+    [ 'C-stack limit (~D bytes) exceeded.'-[Limit], nl ],
+    resolve_c_stack_overflow(Limit).
+out_of_c_stack -->
+    { statistics(c_stack, Limit), Limit > 0 },
+    [ 'C-stack limit exceeded.'-[Limit], nl ],
+    resolve_c_stack_overflow(Limit).
+
+resolve_c_stack_overflow(_Limit) -->
+    { thread_self(main) },
+    [ 'Use the shell command ' ], code('~w', 'ulimit -s size'),
+    [ ' to enlarge the limit.' ].
+resolve_c_stack_overflow(_Limit) -->
+    [ 'Use the ' ], code('~w', 'c_stack(KBytes)'),
+    [ ' option of '], code(thread_create/3), [' to enlarge the limit.' ].
+
 
                  /*******************************
                  *        MAKE/AUTOLOAD         *
@@ -1138,6 +1214,12 @@ prolog_message(close_on_abort(Stream)) -->
     [ 'Abort: closed stream ~p'-[Stream] ].
 prolog_message(cancel_halt(Reason)) -->
     [ 'Halt cancelled: ~p'-[Reason] ].
+prolog_message(on_error(halt(Status))) -->
+    { statistics(errors, Errors),
+      statistics(warnings, Warnings)
+    },
+    [ 'Halting with status ~w due to ~D errors and ~D warnings'-
+      [Status, Errors, Warnings] ].
 
 prolog_message(query(QueryResult)) -->
     query_result(QueryResult).
@@ -1402,12 +1484,12 @@ prolog_message(nospy(Head)) -->
     { goal_to_predicate_indicator(Head, Pred)
     },
     [ 'Spy point removed from ~p'-[Pred] ].
-prolog_message(trace_mode(Bool)) -->
-    [ 'Trace mode switched to ~w'-[Bool] ].
-prolog_message(debug_mode(Bool)) -->
-    [ 'Debug mode switched to ~w'-[Bool] ].
-prolog_message(debugging(Bool)) -->
-    [ 'Debug mode is ~w'-[Bool] ].
+prolog_message(trace_mode(OnOff)) -->
+    [ 'Trace mode switched to ~w'-[OnOff] ].
+prolog_message(debug_mode(OnOff)) -->
+    [ 'Debug mode switched to ~w'-[OnOff] ].
+prolog_message(debugging(OnOff)) -->
+    [ 'Debug mode is ~w'-[OnOff] ].
 prolog_message(spying([])) -->
     !,
     [ 'No spy points' ].
@@ -1457,9 +1539,16 @@ prolog_message(frame(Frame, choice, PC)) -->
 prolog_message(frame(_, cut_call, _)) --> !, [].
 prolog_message(frame(Goal, trace(Port))) -->
     !,
+    thread_context,
     [ ' T ' ],
     port(Port),
     goal(Goal).
+prolog_message(frame(Goal, trace(Port, Id))) -->
+    !,
+    thread_context,
+    [ ' T ' ],
+    port(Port, Id),
+    goal(Goal).
 prolog_message(frame(Frame, Port, _PC)) -->
     frame_flags(Frame),
     port(Port),
@@ -1514,6 +1603,11 @@ frame_flags(Frame) -->
     },
     [ '~w~w '-[T, S] ].
 
+% trace/1,2 context handling
+port(Port, _Id-Level) -->
+    [ '[~d] '-Level ],
+    port(Port).
+
 port(Port) -->
     { port_name(Port, Name)
     },
@@ -1568,7 +1662,7 @@ prolog_message(pack(attached(Pack, BaseDir))) -->
     [ 'Attached package ~w at ~q'-[Pack, BaseDir] ].
 prolog_message(pack(duplicate(Entry, OldDir, Dir))) -->
     [ 'Package ~w already attached at ~q.'-[Entry,OldDir], nl,
-      '\tIgnoring version from ~q'- [Entry, OldDir, Dir]
+      '\tIgnoring version from ~q'- [Dir]
     ].
 prolog_message(pack(no_arch(Entry, Arch))) -->
     [ 'Package ~w: no binary for architecture ~w'-[Entry, Arch] ].
@@ -1615,7 +1709,7 @@ tripwire_message(Wire, Context) -->
     tripwire_context(Wire, Context).
 
 tripwire_context(_, ATrie) -->
-    { '$is_answer_trie'(ATrie),
+    { '$is_answer_trie'(ATrie, _),
       !,
       '$tabling':atrie_goal(ATrie, QGoal),
       user_predicate_indicator(QGoal, Goal)
@@ -1667,6 +1761,7 @@ default_theme(message(Level),         Attrs) :-
     prolog:message_prefix_hook/2.
 :- thread_local
     user:thread_message_hook/3.
+:- '$hide'((push_msg/1,pop_msg/0)).
 
 %!  print_message(+Kind, +Term)
 %
@@ -1699,7 +1794,8 @@ pop_msg :-
     (   nb_current('$inprint_message', [_|Messages]),
         Messages \== []
     ->  b_setval('$inprint_message', Messages)
-    ;   nb_delete('$inprint_message')
+    ;   nb_delete('$inprint_message'),              % delete history
+        b_setval('$inprint_message', [])
     ).
 
 print_message_guarded(Level, Term) :-
@@ -1711,12 +1807,25 @@ print_message_guarded(Level, Term) :-
                 ;   notrace(user:message_hook(Term, Level, Lines))
                 )
             ->  true
-            ;   print_system_message(Term, Level, Lines)
+            ;   '$inc_message_count'(Level),
+                print_system_message(Term, Level, Lines),
+                maybe_halt_on_error(Level)
             )
         )
     ;   true
     ).
 
+maybe_halt_on_error(error) :-
+    current_prolog_flag(on_error, halt),
+    !,
+    halt(1).
+maybe_halt_on_error(warning) :-
+    current_prolog_flag(on_warning, halt),
+    !,
+    halt(1).
+maybe_halt_on_error(_).
+
+
 %!  print_system_message(+Term, +Kind, +Lines)
 %
 %   Print the message if the user did not intecept the message.
@@ -1764,7 +1873,7 @@ print_system_message(_, Kind, Lines) :-
     user:message_property/2.
 
 msg_property(Kind, Property) :-
-    user:message_property(Kind, Property),
+    notrace(user:message_property(Kind, Property)),
     !.
 msg_property(Kind, prefix(Prefix)) :-
     msg_prefix(Kind, Prefix),
@@ -1968,6 +2077,14 @@ actions_to_format([ansi(_Attrs, Fmt0, Args0)|Tail], Fmt, Args) :-
     actions_to_format(Tail, Fmt1, Args1),
     atom_concat(Fmt0, Fmt1, Fmt),
     append_args(Args0, Args1, Args).
+actions_to_format([url(Pos)|Tail], Fmt, Args) :-
+    !,
+    actions_to_format(Tail, Fmt1, Args1),
+    url_actions_to_format(url(Pos), Fmt1, Args1, Fmt, Args).
+actions_to_format([url(URL, Label)|Tail], Fmt, Args) :-
+    !,
+    actions_to_format(Tail, Fmt1, Args1),
+    url_actions_to_format(url(URL, Label), Fmt1, Args1, Fmt, Args).
 actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :-
     !,
     actions_to_format(Tail, Fmt1, Args1),
@@ -1992,14 +2109,33 @@ action_skip(flush).
 action_skip(begin(_Level, _Ctx)).
 action_skip(end(_Ctx)).
 
+url_actions_to_format(url(File:Line:Column), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w:~d:~d', Fmt1, Fmt),
+    append_args([File,Line,Column], Args1, Args).
+url_actions_to_format(url(File:Line), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w:~d', Fmt1, Fmt),
+    append_args([File,Line], Args1, Args).
+url_actions_to_format(url(File), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w', Fmt1, Fmt),
+    append_args([File], Args1, Args).
+url_actions_to_format(url(_URL, Label), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w', Fmt1, Fmt),
+    append_args([Label], Args1, Args).
+
+
 append_args(M:Args0, Args1, M:Args) :-
     !,
     strip_module(Args1, _, A1),
-    '$append'(Args0, A1, Args).
+    to_list(Args0, Args01),
+    '$append'(Args01, A1, Args).
 append_args(Args0, Args1, Args) :-
     strip_module(Args1, _, A1),
-    '$append'(Args0, A1, Args).
-
+    to_list(Args0, Args01),
+    '$append'(Args01, A1, Args).
 
                  /*******************************
                  *    MESSAGES TO PRINT ONCE    *
