@@ -39,10 +39,13 @@
        opt_help/2       %+OptsSpec, -Help
     ]).
 
-:- use_module(library(apply)).
-:- use_module(library(lists)).
-:- use_module(library(option)).
-:- use_module(library(error)).
+:- autoload(library(apply),[maplist/3]).
+:- autoload(library(debug),[assertion/1]).
+:- autoload(library(error),[must_be/2, current_type/3]).
+:- autoload(library(lists),[member/2,max_list/2,reverse/2,append/3]).
+:- autoload(library(option),[merge_options/3,option/3]).
+
+
 :- set_prolog_flag(double_quotes, codes).
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% EXPORTS
 
@@ -439,8 +442,7 @@ opt_help(OptsSpec, Help, HelpOptions0) :-
 %{{{ %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% OPT_PARSE
 
 opt_parse_(OptsSpec0, Args0, Opts, PositionalArgs, ParseOptions) :-
-    assertion(ground(Args0)),
-    assertion(is_list_of_atoms(Args0)),
+    must_be(list(atom), Args0),
 
     check_opts_spec(OptsSpec0, ParseOptions, OptsSpec),
 
@@ -845,6 +847,7 @@ name_rest([]) --> [].
 name_rest([C|Cs]) --> [C], {name_char(C)}, name_rest(Cs).
 name_1st(C) :- char_type(C, alpha).
 name_char(C) :- char_type(C, alpha).
+name_char( 0'_ ).
 name_char( 0'- ). %}}}
 
 
