@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2007-2015, University of Amsterdam
+    Copyright (c)  2007-2018, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -58,6 +59,7 @@ typedef struct chunked_context
 { IOSTREAM	   *stream;		/* Original stream */
   IOSTREAM	   *chunked_stream;	/* Stream I'm handle of */
   int		    close_parent;	/* close parent on close */
+  int		    eof_seen;		/* We saw end-of-file */
   IOENC		    parent_encoding;	/* Saved encoding of parent */
   size_t	    avail;		/* data available */
 } chunked_context;
@@ -94,6 +96,9 @@ static ssize_t				/* decode */
 chunked_read(void *handle, char *buf, size_t size)
 { chunked_context *ctx = handle;
 
+  if ( ctx->eof_seen )
+    return 0;
+
   for(;;)
   { if ( ctx->avail > 0 )			/* data waiting */
     { size_t  max_rd = ctx->avail < size ? ctx->avail : size;
@@ -137,7 +142,9 @@ chunked_read(void *handle, char *buf, size_t size)
 	  { s = Sfgets(hdr, sizeof(hdr), ctx->stream);
 	  } while ( s && strcmp(s, "\r\n") != 0 );
 	  if ( s )
+	  { ctx->eof_seen = TRUE;
 	    return 0;
+	  }
 	  Sseterr(ctx->chunked_stream, 0, "Bad end-of-stream");
 	  return -1;
 	}
@@ -175,9 +182,15 @@ chunked_control(void *handle, int op, void *data)
     case SIO_SETENCODING:
       return 0;				/* allow switching encoding */
     default:
-      if ( ctx->stream->functions->control )
-	return (*ctx->stream->functions->control)(ctx->stream->handle, op, data);
+    { IOSTREAM *s;
+      IOFUNCTIONS *sf;
+
+      if ( (s=ctx->stream) &&
+	   (sf=s->functions) &&
+	   s->magic == SIO_MAGIC )
+	return (*sf->control)(s->handle, op, data);
       return -1;
+    }
   }
 }
 
