@@ -35,7 +35,6 @@
 
 #include <string.h>
 #include <stdlib.h>
-#define PL_ARITY_AS_SIZE 1
 #include <SWI-Stream.h>
 #include <SWI-Prolog.h>
 #include <config.h>
@@ -248,6 +247,7 @@ This code is copied from our GNU libreadline wrapper.
 
 typedef struct
 { int			signo;		/* number of the signal */
+  int			prepared;	/* Is currently prepared */
   struct sigaction	old_state;	/* old state for the signal */
 } sigstate;
 
@@ -265,6 +265,7 @@ static sigstate el_signals[] =
   { SIGTERM },
   { SIGQUIT },
   { SIGWINCH },
+  { SIGUSR2 },				/* SWI-Prolog thread alert */
   { -1 }
 };
 
@@ -280,11 +281,14 @@ static sigstate cont_signals[] =
 static void
 prepare_signals(sigstate *s)
 { for(; s->signo != -1; s++)
-  { struct sigaction new;
+  { if ( !s->prepared )
+    { struct sigaction new;
 
-    memset(&new, 0, sizeof(new));
-    new.sa_handler = el_sighandler;
-    sigaction(s->signo, &new, &s->old_state);
+      memset(&new, 0, sizeof(new));
+      new.sa_handler = el_sighandler;
+      sigaction(s->signo, &new, &s->old_state);
+      s->prepared = TRUE;
+    }
   }
 }
 
@@ -293,6 +297,7 @@ static void
 restore_signals(sigstate *s)
 { for(; s->signo != -1; s++)
   { sigaction(s->signo, &s->old_state, NULL);
+    s->prepared = FALSE;
   }
 }
 
@@ -530,9 +535,7 @@ do_read(el_context *ctx, int fd, char *buf, size_t size)
   int oreader = ctx->reader;
 
   ctx->reader = PL_thread_self();
-  do
-  { rc = read(fd, buf, size);
-  } while ( rc < 0 && errno == EINTR );
+  rc = read(fd, buf, size);
   ctx->reader = oreader;
 
   return rc;
@@ -624,6 +627,8 @@ read_char(EditLine *el, el_char_t *cp)
       *cp = (el_char_t)'\0';
       return -1;
     }
+    if ( e == EINTR )
+      continue;
 
     if ( !tried && read__fixio(fileno(in), e) == 0 )
     { errno = save_errno;
@@ -893,7 +898,7 @@ pl_is_wrapped(term_t tin)
     else
       rc = FALSE;
 
-    PL_release_stream(in);
+    PL_release_stream_noerror(in);
   }
 
   return rc;
@@ -917,7 +922,7 @@ get_el_context(term_t tin, el_context **ctxp)
     { rc = PL_domain_error("libedit_input", tin);
     }
 
-    PL_release_stream(in);
+    PL_release_stream_noerror(in);
   }
 
   return rc;
@@ -1326,6 +1331,59 @@ pl_deletestr(term_t tin, term_t count)
   return FALSE;
 }
 
+static foreign_t
+pl_getc(term_t tin, term_t c)
+{ el_context *ctx;
+
+  if ( get_el_context(tin, &ctx) )
+  { wchar_t wc;
+    switch(el_wgetc(ctx->el, &wc))
+    { case 1:
+	return PL_unify_integer(c, wc);
+      case 0:
+	return PL_unify_integer(c, -1);
+      case -1:
+      default:
+	Sdprintf("el_getc(): I/O error\n");
+	return PL_unify_integer(c, -1);
+    }
+  }
+
+  return FALSE;
+}
+
+static foreign_t
+pl_push(term_t tin, term_t c)
+{ el_context *ctx;
+  int ic;
+
+  if ( PL_get_char_ex(c, &ic, FALSE) &&
+       get_el_context(tin, &ctx) )
+  { wchar_t wc[2];
+    wc[0] = ic;
+    wc[1] = 0;
+    el_wpush(ctx->el, wc);
+
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+static foreign_t
+pl_editmode(term_t tin, term_t on)
+{ el_context *ctx;
+  int m;
+
+  if ( PL_get_bool_ex(on, &m) &&
+       get_el_context(tin, &ctx) )
+  { el_set(ctx->el, EL_EDITMODE, m);
+
+    return TRUE;
+  }
+
+  return FALSE;
+}
 
 
 		 /*******************************
@@ -1539,4 +1597,7 @@ install_libedit4pl(void)
   PL_register_foreign("el_read_history",  2, pl_read_history,  0);
   PL_register_foreign("el_history_events",2, pl_history_events,0);
   PL_register_foreign("el_history",       2, pl_history,       0);
+  PL_register_foreign("el_getc",          2, pl_getc,          0);
+  PL_register_foreign("el_push",          2, pl_push,          0);
+  PL_register_foreign("el_editmode",      2, pl_editmode,      0);
 }
