@@ -73,7 +73,7 @@ NewMetaString(char *init_str)
     META_MALLOC(s->str, s->bufsize, char);
     assert( s->str != NULL );
 
-    strncpy(s->str, init_str, s->length + 1);
+    strcpy(s->str, init_str);
     s->free_string_on_destroy = 1;
 
     return s;
