@@ -41,7 +41,8 @@
 :- use_module(library(debug)).
 
 :- multifile
-    blob_rendering//3.              % +Type, +Blob, +Options
+    blob_rendering//3,              % +Type, +Blob, +Options
+    portray//2.                     % +Term, +Options
 
 /** <module> Represent Prolog terms as HTML
 
@@ -56,9 +57,12 @@ terms as structured HTML.
 %   passed to write_term/3. In addition,   the following options are
 %   processed:
 %
+%     - format(+Format)
+%     Used for atomic values.  Typically this is used to
+%     render a single value.
 %     - float_format(+Format)
 %     If a float is rendered, it is rendered using
-%     `format(string(S), Format, [Float])`%
+%     `format(string(S), Format, [Float])`
 %
 %   @tbd    Cyclic terms.
 %   @tbd    Attributed terms.
@@ -74,15 +78,23 @@ term(Term, Options) -->
                       depth(0)
                     ],
                     Options1),
-      dict_create(Dict, _, Options1)
+      dict_options(Dict, Options1)
     },
-    any(Term, Dict).
+    any(Term, Dict),
+    finalize_term(Term, Dict).
 
 
 any(_, Options) -->
     { Options.depth >= Options.max_depth },
     !,
     html(span(class('pl-ellipsis'), ...)).
+any(Term, Options) -->
+    (   {   nonvar(Term)
+        ;   attvar(Term)
+        }
+    ->  portray(Term, Options)
+    ),
+    !.
 any(Term, Options) -->
     { primitive(Term, Class0),
       !,
@@ -397,12 +409,12 @@ end_code_type(List, Type, _) :-
     Type = punct.
 end_code_type(OpTerm, Type, Options) :-
     compound_name_arity(OpTerm, Name, 1),
-    is_op1(Name, Type, Pri, ArgPri, Options),
+    is_op1(Name, OpType, Pri, ArgPri, Options),
     \+ Options.get(ignore_ops) == true,
     !,
     (   Pri > Options.priority
     ->  Type = punct
-    ;   (   Type == prefix
+    ;   (   OpType == prefix
         ->  end_code_type(Name, Type, Options)
         ;   arg(1, OpTerm, Arg),
             arg_options(Options, ArgOptions),
@@ -503,6 +515,11 @@ quote_atomic(Float, String, Options) :-
     Format = Options.get(float_format),
     !,
     format(string(String), Format, [Float]).
+quote_atomic(Plain, String, Options) :-
+    atomic(Plain),
+    Format = Options.get(format),
+    !,
+    format(string(String), Format, [Plain]).
 quote_atomic(Plain, String, Options) :-
     rational(Plain),
     \+ integer(Plain),
@@ -562,6 +579,24 @@ primitive_class('pl-atom', Atom, String, Class) :-
 primitive_class(Class, _, _, Class).
 
 
+%!  finalize_term(+Term, +Dict)// is det.
+%
+%   Handle the full_stop(Bool) and nl(Bool) options.
+
+finalize_term(Term, Dict) -->
+    (   { true == Dict.get(full_stop) }
+    ->  space(Term, '.', Dict, Dict),
+        (   { true == Dict.get(nl) }
+        ->  html(['.', br([])])
+        ;   html('. ')
+        )
+    ;   (   { true == Dict.get(nl) }
+        ->  html(br([]))
+        ;   []
+        )
+    ).
+
+
                  /*******************************
                  *             HOOKS            *
                  *******************************/
