@@ -188,7 +188,7 @@ syntax_error(IOSTREAM *in, const char *msg)
     { c = Sgetcode(in);
     } while(c != '\n' && c != -1);
 
-    return PL_raise_exception(ex);
+    return PL_raise_exception(ex),FALSE;
   }
 
   return FALSE;
@@ -417,7 +417,7 @@ get_iri_code(IOSTREAM *in, int *cp)
 
 static int
 read_uniref(IOSTREAM *in, term_t subject, int *cp)
-{ int c;
+{ int c = -1;
   string_buffer buf;
 
   initBuf(&buf);
@@ -558,9 +558,6 @@ read_subject(IOSTREAM *in, term_t subject, int *cp)
       return syntax_error(in, "subject expected");
   }
 
-  if ( rc && !is_ws(*cp) )
-    return syntax_error(in, "subject not followed by whitespace");
-
   return rc;
 }
 
@@ -578,9 +575,6 @@ read_predicate(IOSTREAM *in, term_t predicate, int *cp)
       return syntax_error(in, "predicate expected");
   }
 
-  if ( rc && !is_ws(*cp) )
-    return syntax_error(in, "predicate not followed by whitespace");
-
   return rc;
 }
 
@@ -642,7 +636,7 @@ get_string_code(IOSTREAM *in, int *cp)
 
 static int
 read_literal(IOSTREAM *in, term_t literal, int *cp)
-{ int c;
+{ int c = -1;
   string_buffer buf;
 
   initBuf(&buf);
