@@ -276,7 +276,7 @@ typedef struct literal
     int64_t	integer;
     double	real;
     struct
-    { record_t  record;
+    { char *    record;
       size_t	len;
     } term;				/* external record */
   } value;
@@ -397,7 +397,6 @@ typedef union triple_element
 typedef struct triple_array
 { triple_element *blocks[MAX_TBLOCKS];	/* Dynamic array starts */
   triple_element *freelist;		/* free buckets */
-  size_t	preinit;		/* Preinitialized size */
   size_t	size;			/* current (allocated) size */
 } triple_array;
 #endif /*COMPACT*/
