@@ -47,7 +47,7 @@ xmlns *		xmlns_push(dtd_parser *p, const ichar *ns, const ichar *url);
 void		update_xmlns(dtd_parser *p, dtd_element *e,
 			     int natts, sgml_attribute *atts);
 int		xmlns_resolve_attribute(dtd_parser *p, dtd_symbol *id,
-					const ichar **local, const ichar **url);
+                                        const ichar **local, const ichar **url, const ichar **prefix);
 int		xmlns_resolve_element(dtd_parser *p,
                                       const ichar **local, const ichar **url, const ichar **prefix);
 
