@@ -39,6 +39,12 @@
 #include <SWI-Prolog.h>
 #include <openssl/hmac.h>
 #include <openssl/rand.h>
+#ifdef HAVE_OPENSSL_CORE_NAMES_H
+#include <openssl/core_names.h>
+#endif
+#ifdef HAVE_OPENSSL_PARAM_BUILD_H
+#include <openssl/param_build.h>
+#endif
 #ifdef HAVE_OPENSSL_KDF_H
 #include <openssl/kdf.h>
 #endif
@@ -81,6 +87,17 @@ typedef enum
 { RSA_MODE, EVP_MODE
 } crypt_mode_t;
 
+#if defined HAVE_EVP_PKEY_NEW && defined HAVE_EVP_PKEY_FREE && defined HAVE_EVP_PKEY_GET_BN_PARAM && defined HAVE_EVP_PKEY_GET_OCTET_STRING_PARAM && defined HAVE_EVP_PKEY_GET_SIZE && defined HAVE_EVP_PKEY_DECRYPT && defined HAVE_EVP_PKEY_ENCRYPT && defined HAVE_EVP_PKEY_SIGN && defined HAVE_EVP_PKEY_VERIFY && defined HAVE_EVP_PKEY_Q_KEYGEN && defined HAVE_OSSL_PARAM_CONSTRUCT_UTF8_STRING && defined HAVE_BN_CHECK_PRIME && defined HAVE_OSSL_PARAM_BLD_NEW
+#define USE_EVP_API 1
+#endif
+
+#ifdef USE_EVP_API
+#define RSAKEY EVP_PKEY
+#else
+#define RSAKEY RSA
+#endif
+
+
 
                  /***************************
                  *       RANDOM BYTES       *
@@ -131,16 +148,24 @@ typedef struct hash_context
   int             close_parent;
 
   EVP_MD_CTX     *ctx;
-  HMAC_CTX       *hmac_ctx;
-  char           *hmac_key;
+#if defined USE_EVP_API
+  EVP_MAC         *mac;
+  EVP_MAC_CTX     *mac_ctx;
+#else
+  HMAC_CTX       *mac_ctx;
+#endif
+  char           *mac_key;
 } PL_CRYPTO_HASH_CONTEXT;
 
 static void
 free_crypto_hash_context(PL_CRYPTO_HASH_CONTEXT *c)
 { EVP_MD_CTX_free(c->ctx);
-  free(c->hmac_key);
-#ifdef HAVE_HMAC_CTX_FREE
-  HMAC_CTX_free(c->hmac_ctx);
+  free(c->mac_key);
+#ifdef USE_EVP_API
+  EVP_MAC_free(c->mac);
+  EVP_MAC_CTX_free(c->mac_ctx);
+#elif defined HAVE_HMAC_CTX_FREE
+  HMAC_CTX_free(c->mac_ctx);
 #endif
   free(c);
 }
@@ -310,7 +335,7 @@ hash_options(term_t options, PL_CRYPTO_HASH_CONTEXT *result)
         if ( !PL_get_nchars(a, &key_len, &key,
                       CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
           return FALSE;
-        result->hmac_key = ssl_strdup(key);
+        result->mac_key = ssl_strdup(key);
       } else if ( aname == ATOM_close_parent )
       { if ( !PL_get_bool_ex(a, &result->close_parent) )
           return FALSE;
@@ -345,8 +370,8 @@ pl_crypto_hash_context_new(term_t tcontext, term_t options)
 
   context->magic    = HASH_CONTEXT_MAGIC;
   context->ctx      = NULL;
-  context->hmac_ctx = NULL;
-  context->hmac_key = NULL;
+  context->mac_ctx  = NULL;
+  context->mac_key  = NULL;
 
   context->parent_stream = NULL;
   context->hash_stream   = NULL;
@@ -354,19 +379,41 @@ pl_crypto_hash_context_new(term_t tcontext, term_t options)
   if ( !hash_options(options, context) )
     return FALSE;
 
-#ifdef HAVE_HMAC_CTX_NEW
-  if ( context->hmac_key )
-  { context->hmac_ctx = HMAC_CTX_new();
-    if ( !HMAC_Init_ex(context->hmac_ctx,
-                       context->hmac_key, strlen(context->hmac_key),
+#ifdef USE_EVP_API
+  if ( context->mac_key )
+  { OSSL_PARAM params[2];
+    context->mac = EVP_MAC_fetch(NULL, "HMAC", NULL);
+    if ( context->mac == NULL )
+    { return FALSE;
+    }
+    context->mac_ctx = EVP_MAC_CTX_new(context->mac);
+    if ( context->mac_ctx == NULL )
+    { EVP_MAC_free(context->mac);
+      return FALSE;
+    }
+    params[0] = OSSL_PARAM_construct_utf8_string("digest", (char *)EVP_MD_name(context->algorithm), 0);
+    params[1] = OSSL_PARAM_construct_end();
+    if ( !EVP_MAC_init(context->mac_ctx,
+                       (unsigned char*)context->mac_key, strlen(context->mac_key),
+                       params) )
+    { EVP_MAC_CTX_free(context->mac_ctx);
+      EVP_MAC_free(context->mac);
+      return FALSE;
+    }
+  }
+#elif defined HAVE_HMAC_CTX_NEW
+  if ( context->mac_key )
+  { context->mac_ctx = HMAC_CTX_new();
+    if ( !HMAC_Init_ex(context->mac_ctx,
+                       context->mac_key, strlen(context->mac_key),
                        context->algorithm, NULL) )
-    { HMAC_CTX_free(context->hmac_ctx);
+    { HMAC_CTX_free(context->mac_ctx);
       return FALSE;
     }
   }
 #endif
 
-  if ( !context->hmac_ctx )
+  if ( !context->mac_ctx )
   { context->ctx = EVP_MD_CTX_new();
     if ( !EVP_DigestInit_ex(context->ctx, context->algorithm, NULL) )
     { EVP_MD_CTX_free(context->ctx);
@@ -392,7 +439,7 @@ pl_crypto_hash_context_copy(term_t tin, term_t tout)
     return FALSE;
 
   out->magic = HASH_CONTEXT_MAGIC;
-  out->hmac_key = ssl_strdup(in->hmac_key);
+  out->mac_key = ssl_strdup(in->mac_key);
 
   out->encoding = in->encoding;
   out->algorithm = in->algorithm;
@@ -405,21 +452,27 @@ pl_crypto_hash_context_copy(term_t tin, term_t tout)
     }
     rc = EVP_MD_CTX_copy_ex(out->ctx, in->ctx);
   }
+#if defined(USE_EVP_API) && defined(USE_EVP_API)
+  out->mac = in->mac;
+  if ( in->mac != NULL )
+  { EVP_MAC_up_ref(in->mac);
+  }
+  out->mac_ctx = in->mac_ctx ? EVP_MAC_CTX_dup(in->mac_ctx) : NULL;
+  rc = TRUE;
+#elif defined(HAVE_HMAC_CTX_NEW) && defined(HAVE_HMAC_CTX_FREE)
+  out->mac_ctx = in->mac_ctx ? HMAC_CTX_new() : NULL;
 
-#if defined(HAVE_HMAC_CTX_NEW) && defined(HAVE_HMAC_CTX_FREE)
-  out->hmac_ctx = in->hmac_ctx ? HMAC_CTX_new() : NULL;
-
-  if ( out->hmac_ctx )
-  { if ( !HMAC_Init_ex(out->hmac_ctx,
-                       out->hmac_key, strlen(out->hmac_key),
+  if ( out->mac_ctx )
+  { if ( !HMAC_Init_ex(out->mac_ctx,
+                       out->mac_key, strlen(out->mac_key),
                        out->algorithm, NULL) )
-    { HMAC_CTX_free(out->hmac_ctx);
+    { HMAC_CTX_free(out->mac_ctx);
       return FALSE;
     }
-    rc = HMAC_CTX_copy(out->hmac_ctx, in->hmac_ctx);
+    rc = HMAC_CTX_copy(out->mac_ctx, in->mac_ctx);
   }
 #else
-  out->hmac_ctx = NULL;
+  out->mac_ctx = NULL;
 #endif
 
   return unify_hash_context(tout, out) && rc;
@@ -429,8 +482,14 @@ pl_crypto_hash_context_copy(term_t tin, term_t tout)
 static int
 hash_append(PL_CRYPTO_HASH_CONTEXT *context, void *data, size_t size)
 {
-  if ( context->hmac_ctx )
-    return HMAC_Update(context->hmac_ctx, data, size);
+  if ( context->mac_ctx )
+  {
+#ifdef USE_EVP_API
+    return  EVP_MAC_update(context->mac_ctx, data, size);
+#else
+    return HMAC_Update(context->mac_ctx, data, size);
+#endif
+  }
 
   return EVP_DigestUpdate(context->ctx, data, size);
 }
@@ -459,15 +518,23 @@ pl_crypto_hash_context_hash(term_t tcontext, term_t hash)
 {
   PL_CRYPTO_HASH_CONTEXT *context = NULL;
   unsigned char digest[EVP_MAX_MD_SIZE];
-  unsigned int len;
-
+  size_t len;
   if ( !get_hash_context(tcontext, &context) )
     return FALSE;
 
-  if ( context->hmac_ctx )
-  { HMAC_Final(context->hmac_ctx, digest, &len);
+  if ( context->mac_ctx )
+  {
+#ifdef USE_EVP_API
+    EVP_MAC_final(context->mac_ctx, digest, &len, EVP_MAX_MD_SIZE);
+#else
+    unsigned int ulen;
+    HMAC_Final(context->mac_ctx, digest, &ulen);
+    len = ulen;
+#endif
   } else
-  { EVP_DigestFinal_ex(context->ctx, digest, &len);
+  { unsigned int ulen;
+    EVP_DigestFinal_ex(context->ctx, digest, &ulen);
+    len = ulen;
   }
 
   return PL_unify_list_ncodes(hash, len, (char *) digest);
@@ -677,9 +744,9 @@ pl_crypto_data_hkdf(term_t tkey, term_t tsalt, term_t tinfo, term_t talg,
 
   if ( (EVP_PKEY_derive_init(pctx) > 0) &&
        (EVP_PKEY_CTX_set_hkdf_md(pctx, alg) > 0) &&
-       (EVP_PKEY_CTX_set1_hkdf_salt(pctx, salt, saltlen) > 0) &&
-       (EVP_PKEY_CTX_set1_hkdf_key(pctx, key, keylen) > 0) &&
-       (EVP_PKEY_CTX_add1_hkdf_info(pctx, info, infolen) > 0) &&
+       (EVP_PKEY_CTX_set1_hkdf_salt(pctx, (unsigned char*)salt, saltlen) > 0) &&
+       (EVP_PKEY_CTX_set1_hkdf_key(pctx, (unsigned char*)key, keylen) > 0) &&
+       (EVP_PKEY_CTX_add1_hkdf_info(pctx, (unsigned char*)info, infolen) > 0) &&
        (EVP_PKEY_derive(pctx, out, &outlen) > 0) )
   { int rc = PL_unify_list_ncodes(tout, outlen, (char *) out);
     free(out);
@@ -720,10 +787,17 @@ get_bn_arg(int a, term_t t, BIGNUM **bn)
 }
 
 #ifndef OPENSSL_NO_EC
+
+#ifdef USE_EVP_API
+#define ECKEY EVP_PKEY
+#else
+#define ECKEY EC_KEY
+#endif
+
 static int
-recover_ec(term_t t, EC_KEY **rec)
+recover_ec(term_t t, ECKEY **rec)
 {
-  EC_KEY *ec;
+  ECKEY *key;
   BIGNUM *privkey = NULL;
   term_t pubkey;
   unsigned char *codes;
@@ -734,44 +808,75 @@ recover_ec(term_t t, EC_KEY **rec)
   if ( !(tcurve &&
          PL_get_arg(3, t, tcurve) &&
          PL_get_chars(tcurve, &curve, CVT_ATOM|CVT_STRING|CVT_EXCEPTION) &&
-         (ec = EC_KEY_new_by_curve_name(OBJ_sn2nid(curve)))) )
+#ifdef USE_EVP_API
+         (key = EVP_EC_gen(curve))
+#else
+         (key = EC_KEY_new_by_curve_name(OBJ_sn2nid(curve)))
+#endif
+     ) )
     return FALSE;
 
   if ( !get_bn_arg(1, t, &privkey) )
-  { EC_KEY_free(ec);
+  {
+#ifdef USE_EVP_API
+    EVP_PKEY_free(key);
+#else
+    EC_KEY_free(key);
+#endif
     return FALSE;
   }
 
   if ( privkey )
-    EC_KEY_set_private_key(ec, privkey);
+  {
+#ifdef USE_EVP_API
+    EVP_PKEY_set_bn_param(key, "priv", privkey);
+#else
+    EC_KEY_set_private_key(key, privkey);
+#endif
+  }
 
   if ( (pubkey=PL_new_term_ref()) &&
        PL_get_arg(2, t, pubkey) &&
        PL_get_nchars(pubkey, &codes_len, (char **) &codes,
                      CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) &&
-       (ec = o2i_ECPublicKey(&ec, (const unsigned char**) &codes, codes_len)) )
-  { *rec = ec;
+#ifdef USE_EVP_API
+        EVP_PKEY_set_octet_string_param(key, "pub", (const unsigned char*) codes, codes_len)
+#else
+       (key = o2i_ECPublicKey(&key, (const unsigned char**) &codes, codes_len))
+#endif
+  )
+  { *rec = key;
     return TRUE;
   }
 
-  EC_KEY_free(ec);
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
+  EC_KEY_free(key);
+#endif
   return FALSE;
 }
 #endif
 
+
 static int
-recover_rsa(term_t t, RSA** rsap)
-{ RSA *rsa = RSA_new();
+recover_rsa(term_t t, RSAKEY** keyp)
+{
+#ifdef USE_EVP_API
+  RSAKEY* key = EVP_PKEY_new();
+#else
+  RSAKEY *key = RSA_new();
+#endif
 
 #if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
-  if ( get_bn_arg(1, t, &rsa->n) &&
-       get_bn_arg(2, t, &rsa->e) &&
-       get_bn_arg(3, t, &rsa->d) &&
-       get_bn_arg(4, t, &rsa->p) &&
-       get_bn_arg(5, t, &rsa->q) &&
-       get_bn_arg(6, t, &rsa->dmp1) &&
-       get_bn_arg(7, t, &rsa->dmq1) &&
-       get_bn_arg(8, t, &rsa->iqmp)
+  if ( get_bn_arg(1, t, &key->n) &&
+       get_bn_arg(2, t, &key->e) &&
+       get_bn_arg(3, t, &key->d) &&
+       get_bn_arg(4, t, &key->p) &&
+       get_bn_arg(5, t, &key->q) &&
+       get_bn_arg(6, t, &key->dmp1) &&
+       get_bn_arg(7, t, &key->dmq1) &&
+       get_bn_arg(8, t, &key->iqmp)
      )
   {
 #else
@@ -787,25 +892,71 @@ recover_rsa(term_t t, RSA** rsap)
        get_bn_arg(7, t, &dmq1) &&
        get_bn_arg(8, t, &iqmp) )
   {
-    if ( !RSA_set0_key(rsa, n, e, d) ||
-         ( (p || q) && !RSA_set0_factors(rsa, p, q) ) ||
+#ifdef USE_EVP_API
+  OSSL_PARAM_BLD *param_builder;
+  OSSL_PARAM *params = NULL;
+  EVP_PKEY_CTX* ctx;
+  ctx = EVP_PKEY_CTX_new_from_name(NULL, "RSA", NULL);
+  if (!ctx)
+  { return FALSE;
+  }
+  param_builder = OSSL_PARAM_BLD_new();
+  if ( ! ( OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_N, n) &&
+           OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_E, e) &&
+           OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_D, d) ) ||
+        ( ( p || q ) && ! ( OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_FACTOR1, p) &&
+                            OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_FACTOR2, q) ) ) ||
+        ( ( dmp1 || dmq1 || iqmp ) && ! ( OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_EXPONENT1, dmp1) &&
+                                          OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_EXPONENT2, dmq1) &&
+                                          OSSL_PARAM_BLD_push_BN(param_builder, OSSL_PKEY_PARAM_RSA_COEFFICIENT1, iqmp) ) ) )
+  { EVP_PKEY_free(key);
+    OSSL_PARAM_BLD_free(param_builder);
+    return raise_ssl_error(ERR_get_error());
+  }
+  params = OSSL_PARAM_BLD_to_param(param_builder);
+  if (!params)
+  { EVP_PKEY_CTX_free(ctx);
+    return FALSE;
+  }
+	OSSL_PARAM_BLD_free(param_builder);
+  if (EVP_PKEY_fromdata_init(ctx) <= 0)
+  { EVP_PKEY_CTX_free(ctx);
+    OSSL_PARAM_free(params);
+    return raise_ssl_error(ERR_get_error());
+  }
+  if (EVP_PKEY_fromdata(ctx, &key, EVP_PKEY_KEYPAIR, params) <= 0)
+  { EVP_PKEY_CTX_free(ctx);
+    OSSL_PARAM_free(params);
+    return raise_ssl_error(ERR_get_error());
+  }
+  EVP_PKEY_CTX_free(ctx);
+  OSSL_PARAM_free(params);
+  *keyp = key;
+  return TRUE;
+#else
+    if ( !RSA_set0_key(key, n, e, d) ||
+         ( (p || q) && !RSA_set0_factors(key, p, q) ) ||
          ( (dmp1 || dmq1 || iqmp) &&
-           !RSA_set0_crt_params(rsa, dmp1, dmq1, iqmp)) )
-    { RSA_free(rsa);
+           !RSA_set0_crt_params(key, dmp1, dmq1, iqmp)) )
+    { RSA_free(key);
       return FALSE;
     }
 #endif
-    *rsap = rsa;
+#endif
+    *keyp = key;
     return TRUE;
   }
-
-  RSA_free(rsa);
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
+  RSA_free(key);
+#endif
   return FALSE;
 }
 
 
 static int
-recover_private_key(term_t t, RSA** rsap)
+recover_private_key(term_t t, RSAKEY** rsap)
 { if ( PL_is_functor(t, FUNCTOR_private_key1) )
   { term_t arg;
 
@@ -821,7 +972,7 @@ recover_private_key(term_t t, RSA** rsap)
 
 
 static int
-recover_public_key(term_t t, RSA** rsap)
+recover_public_key(term_t t, RSAKEY** rsap)
 { if ( PL_is_functor(t, FUNCTOR_public_key1) )
   { term_t arg;
 
@@ -848,7 +999,9 @@ get_padding(term_t t, crypt_mode_t mode, int *padding)
   { if      ( a == ATOM_pkcs1 && mode == RSA_MODE )      *padding = RSA_PKCS1_PADDING;
     else if ( a == ATOM_pkcs1_oaep && mode == RSA_MODE ) *padding = RSA_PKCS1_OAEP_PADDING;
     else if ( a == ATOM_none && mode == RSA_MODE  )      *padding = RSA_NO_PADDING;
+#ifdef RSA_SSLV23_PADDING    /* removed from OpenSSL 3.0 */
     else if ( a == ATOM_sslv23  && mode == RSA_MODE )    *padding = RSA_SSLV23_PADDING;
+#endif
     else if ( a == ATOM_none  && mode == EVP_MODE )      *padding = 0;
     else if ( a == ATOM_block  && mode == EVP_MODE )     *padding = 1;
     else return PL_domain_error("padding", t);
@@ -922,20 +1075,36 @@ pl_ecdsa_sign(term_t Private, term_t Data, term_t Enc, term_t Signature)
 #ifndef OPENSSL_NO_ECDSA
   unsigned char *data;
   size_t data_len;
-  EC_KEY *key;
-  ECDSA_SIG *sig;
+  ECKEY *key;
   unsigned char *signature = NULL;
-  int signature_len, rc;
+  int rc;
+#ifdef USE_EVP_API
+  size_t signature_len;
+#else
+  ECDSA_SIG *sig;
+  unsigned int signature_len;
+#endif
 
   if ( !recover_ec(Private, &key) ||
        !get_enc_text(Data, Enc, &data_len, &data) )
     return FALSE;
 
+#ifdef USE_EVP_API
+  signature_len = EVP_PKEY_get_size(key);
+  EVP_PKEY_CTX *sign_ctx = EVP_PKEY_CTX_new(key, NULL);
+  EVP_PKEY_sign_init(sign_ctx);
+  rc = EVP_PKEY_sign(sign_ctx,
+		                 signature, &signature_len,
+		                 data, (unsigned int)data_len);
+  EVP_PKEY_CTX_free(sign_ctx);
+  if (!rc)
+    return raise_ssl_error(ERR_get_error());
+#else
   sig = ECDSA_do_sign(data, (unsigned int)data_len, key);
   EC_KEY_free(key);
-
   if ( (signature_len = i2d_ECDSA_SIG(sig, &signature)) < 0 )
     return raise_ssl_error(ERR_get_error());
+#endif
 
   rc = unify_bytes_hex(Signature, signature_len, signature);
   OPENSSL_free(signature);
@@ -952,7 +1121,7 @@ pl_ecdsa_verify(term_t Public, term_t Data, term_t Enc, term_t Signature)
 #ifndef OPENSSL_NO_ECDSA
   unsigned char *data;
   size_t data_len;
-  EC_KEY *key;
+  ECKEY *key;
   ECDSA_SIG *sig;
   unsigned char *signature;
   const unsigned char *copy;
@@ -969,9 +1138,19 @@ pl_ecdsa_verify(term_t Public, term_t Data, term_t Enc, term_t Signature)
   if ( !(sig = d2i_ECDSA_SIG(NULL, &copy, signature_len)) )
     return FALSE;
 
+#ifdef USE_EVP_API
+  EVP_PKEY_CTX *verify_ctx = EVP_PKEY_CTX_new(key, NULL);
+  EVP_PKEY_verify_init(verify_ctx);
+  rc = EVP_PKEY_verify(verify_ctx,
+                       signature, (unsigned int)signature_len,
+                       data, (unsigned int)data_len);
+  EVP_PKEY_CTX_free(verify_ctx);
+  EVP_PKEY_free(key);
+#else
   rc = ECDSA_do_verify(data, data_len, sig, key);
 
   EC_KEY_free(key);
+#endif
   ECDSA_SIG_free(sig);
 
   if (rc == 0 || rc == 1 )
@@ -996,8 +1175,12 @@ pl_rsa_private_decrypt(term_t private_t, term_t cipher_t,
 { size_t cipher_length;
   unsigned char* cipher;
   unsigned char* plain;
+#ifdef USE_EVP_API
+  size_t outsize;
+#else
   int outsize;
-  RSA* key;
+#endif
+  RSAKEY* key;
   int rep = REP_UTF8;
   int padding = RSA_PKCS1_PADDING;
   int retval;
@@ -1010,11 +1193,36 @@ pl_rsa_private_decrypt(term_t private_t, term_t cipher_t,
     return FALSE;
   if ( !recover_private_key(private_t, &key) )
     return FALSE;
-
+#ifdef USE_EVP_API
+  outsize = EVP_PKEY_get_size(key);
+#else
   outsize = RSA_size(key);
+#endif
   ssl_deb(1, "Output size is going to be %d", outsize);
   plain = PL_malloc(outsize);
   ssl_deb(1, "Allocated %d bytes for plaintext", outsize);
+#ifdef USE_EVP_API
+  EVP_PKEY_CTX *ctx = EVP_PKEY_CTX_new(key, NULL);
+  if ( !ctx )
+  { EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  if ( EVP_PKEY_decrypt_init(ctx) <= 0 || EVP_PKEY_CTX_set_rsa_padding(ctx, padding) <= 0)
+  { EVP_PKEY_CTX_free(ctx);
+    EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  if ( EVP_PKEY_decrypt(ctx, plain, &outsize, cipher, cipher_length) <= 0)
+  { ssl_deb(1, "Failure to decrypt!");
+    EVP_PKEY_CTX_free(ctx);
+    EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  EVP_PKEY_CTX_free(ctx);
+#else
   if ((outsize = RSA_private_decrypt((int)cipher_length, cipher,
 				     plain, key, padding)) <= 0)
   { ssl_deb(1, "Failure to decrypt!");
@@ -1022,9 +1230,14 @@ pl_rsa_private_decrypt(term_t private_t, term_t cipher_t,
     PL_free(plain);
     return raise_ssl_error(ERR_get_error());
   }
+#endif
   ssl_deb(1, "decrypted bytes: %d", outsize);
   ssl_deb(1, "Freeing RSA");
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
   RSA_free(key);
+#endif
   ssl_deb(1, "Assembling plaintext");
   retval = PL_unify_chars(plain_t, rep | PL_STRING, outsize, (char*)plain);
   ssl_deb(1, "Freeing plaintext");
@@ -1040,8 +1253,12 @@ pl_rsa_public_decrypt(term_t public_t, term_t cipher_t,
 { size_t cipher_length;
   unsigned char* cipher;
   unsigned char* plain;
+#ifdef USE_EVP_API
+  size_t outsize;
+#else
   int outsize;
-  RSA* key;
+#endif
+  RSAKEY* key;
   int rep = REP_UTF8;
   int padding = RSA_PKCS1_PADDING;
   int retval;
@@ -1053,11 +1270,38 @@ pl_rsa_public_decrypt(term_t public_t, term_t cipher_t,
     return FALSE;
   if ( !recover_public_key(public_t, &key) )
     return FALSE;
-
+#ifdef USE_EVP_API
+  outsize = EVP_PKEY_get_size(key);
+#else
   outsize = RSA_size(key);
+#endif
   ssl_deb(1, "Output size is going to be %d", outsize);
   plain = PL_malloc(outsize);
   ssl_deb(1, "Allocated %d bytes for plaintext", outsize);
+#ifdef USE_EVP_API
+  EVP_PKEY_CTX *ctx = EVP_PKEY_CTX_new(key, NULL);
+  if ( !ctx )
+  { EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  // EVP_PKEY_decrypt is only for private-key decryption.
+  // The equivalent to RSA_public_decrypt is actually EVP_PKEY_verify_recover
+  if ( EVP_PKEY_verify_recover_init(ctx) <= 0 || EVP_PKEY_CTX_set_rsa_padding(ctx, padding) <= 0)
+  { EVP_PKEY_CTX_free(ctx);
+    EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  if (EVP_PKEY_verify_recover(ctx, plain, &outsize, cipher, cipher_length) <= 0)
+  { ssl_deb(1, "Failure to decrypt!");
+    EVP_PKEY_CTX_free(ctx);
+    EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  EVP_PKEY_CTX_free(ctx);
+#else
   if ((outsize = RSA_public_decrypt((int)cipher_length, cipher,
                                     plain, key, padding)) <= 0)
   { ssl_deb(1, "Failure to decrypt!");
@@ -1065,9 +1309,14 @@ pl_rsa_public_decrypt(term_t public_t, term_t cipher_t,
     PL_free(plain);
     return raise_ssl_error(ERR_get_error());
   }
+#endif
   ssl_deb(1, "decrypted bytes: %d", outsize);
   ssl_deb(1, "Freeing RSA");
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
   RSA_free(key);
+#endif
   ssl_deb(1, "Assembling plaintext");
   retval = PL_unify_chars(plain_t, rep | PL_STRING, outsize, (char*)plain);
   ssl_deb(1, "Freeing plaintext");
@@ -1083,8 +1332,12 @@ pl_rsa_public_encrypt(term_t public_t,
 { size_t plain_length;
   unsigned char* cipher;
   unsigned char* plain;
+#ifdef USE_EVP_API
+  size_t outsize;
+#else
   int outsize;
-  RSA* key;
+#endif
+  RSAKEY* key;
   int rep = REP_UTF8;
   int padding = RSA_PKCS1_PADDING;
   int retval;
@@ -1100,10 +1353,36 @@ pl_rsa_public_encrypt(term_t public_t,
   if ( !recover_public_key(public_t, &key) )
     return FALSE;
 
+#ifdef USE_EVP_API
+  outsize = EVP_PKEY_get_size(key);
+#else
   outsize = RSA_size(key);
+#endif
   ssl_deb(1, "Output size is going to be %d\n", outsize);
   cipher = PL_malloc(outsize);
   ssl_deb(1, "Allocated %d bytes for ciphertext\n", outsize);
+#ifdef USE_EVP_API
+  EVP_PKEY_CTX *ctx = EVP_PKEY_CTX_new(key, NULL);
+  if ( !ctx )
+  { EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  if ( EVP_PKEY_encrypt_init(ctx) <= 0 || EVP_PKEY_CTX_set_rsa_padding(ctx, padding) <= 0)
+  { EVP_PKEY_CTX_free(ctx);
+    EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  if ( EVP_PKEY_encrypt(ctx, cipher, &outsize, plain, plain_length) <= 0 )
+  { ssl_deb(1, "Failure to encrypt!");
+    EVP_PKEY_CTX_free(ctx);
+    PL_free(cipher);
+    EVP_PKEY_free(key);
+    return raise_ssl_error(ERR_get_error());
+  }
+  EVP_PKEY_CTX_free(ctx);
+#else
   if ( (outsize = RSA_public_encrypt((int)plain_length, plain,
 				     cipher, key, padding)) <= 0)
   { ssl_deb(1, "Failure to encrypt!");
@@ -1111,9 +1390,14 @@ pl_rsa_public_encrypt(term_t public_t,
     RSA_free(key);
     return raise_ssl_error(ERR_get_error());
   }
+#endif
   ssl_deb(1, "encrypted bytes: %d\n", outsize);
   ssl_deb(1, "Freeing RSA");
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
   RSA_free(key);
+#endif
   ssl_deb(1, "Assembling plaintext");
   retval = PL_unify_chars(cipher_t, PL_STRING|REP_ISO_LATIN_1,
 			  outsize, (char*)cipher);
@@ -1131,8 +1415,12 @@ pl_rsa_private_encrypt(term_t private_t,
 { size_t plain_length;
   unsigned char* cipher;
   unsigned char* plain;
+#ifdef USE_EVP_API
+  size_t outsize;
+#else
   int outsize;
-  RSA* key;
+#endif
+  RSAKEY* key;
   int rep = REP_UTF8;
   int padding = RSA_PKCS1_PADDING;
   int retval;
@@ -1146,10 +1434,37 @@ pl_rsa_private_encrypt(term_t private_t,
   if ( !recover_private_key(private_t, &key) )
     return FALSE;
 
+#ifdef USE_EVP_API
+  outsize = EVP_PKEY_get_size(key);
+#else
   outsize = RSA_size(key);
+#endif
   ssl_deb(1, "Output size is going to be %d", outsize);
   cipher = PL_malloc(outsize);
   ssl_deb(1, "Allocated %d bytes for ciphertext", outsize);
+#ifdef USE_EVP_API
+  memset(cipher, 0, outsize);
+  EVP_PKEY_CTX *ctx = EVP_PKEY_CTX_new(key, NULL);
+  if ( !ctx )
+  { EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  // EVP_PKEY_ENCRYPT is only for public encryption.
+  if ( (EVP_PKEY_sign_init(ctx) <= 0) || (EVP_PKEY_CTX_set_rsa_padding(ctx, padding) <= 0))
+  { EVP_PKEY_CTX_free(ctx);
+    EVP_PKEY_free(key);
+    PL_free(plain);
+    return raise_ssl_error(ERR_get_error());
+  }
+  if ( EVP_PKEY_sign(ctx, cipher, &outsize, plain, plain_length) <= 0 )
+  { ssl_deb(1, "Failure to encrypt!");
+    EVP_PKEY_CTX_free(ctx);
+    PL_free(cipher);
+    EVP_PKEY_free(key);
+    return raise_ssl_error(ERR_get_error());
+  }
+#else
   if ((outsize = RSA_private_encrypt((int)plain_length, plain,
                                      cipher, key, padding)) <= 0)
   { ssl_deb(1, "Failure to encrypt!");
@@ -1157,9 +1472,14 @@ pl_rsa_private_encrypt(term_t private_t,
     RSA_free(key);
     return raise_ssl_error(ERR_get_error());
   }
+#endif
   ssl_deb(1, "encrypted bytes: %d", outsize);
   ssl_deb(1, "Freeing RSA");
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
   RSA_free(key);
+#endif
   ssl_deb(1, "Assembling plaintext");
   retval = PL_unify_chars(cipher_t, PL_STRING|REP_ISO_LATIN_1,
 			  outsize, (char*)cipher);
@@ -1198,9 +1518,13 @@ pl_rsa_sign(term_t Private, term_t Type, term_t Enc,
 	    term_t Data, term_t Signature)
 { unsigned char *data;
   size_t data_len;
-  RSA *key;
+  RSAKEY *key;
   unsigned char *signature;
+#ifdef USE_EVP_API
+  size_t signature_len;
+#else
   unsigned int signature_len;
+#endif
   int rc;
   int type;
 
@@ -1209,12 +1533,32 @@ pl_rsa_sign(term_t Private, term_t Type, term_t Enc,
        !get_digest_type(Type, &type) )
     return FALSE;
 
+#ifdef USE_EVP_API
+  signature_len = EVP_PKEY_get_size(key);
+#else
   signature_len = RSA_size(key);
+#endif
   signature = PL_malloc(signature_len);
+#if defined USE_EVP_API
+  OSSL_PARAM params[2];
+  params[0] = OSSL_PARAM_construct_utf8_string("digest", (char *)OBJ_nid2ln(type), 0);
+  params[1] = OSSL_PARAM_construct_end();
+  EVP_PKEY_CTX *sign_ctx = EVP_PKEY_CTX_new(key, NULL);
+  EVP_PKEY_sign_init_ex(sign_ctx, params);
+  rc = EVP_PKEY_sign(sign_ctx,
+		signature, &signature_len,
+		data, (unsigned int)data_len);
+  EVP_PKEY_CTX_free(sign_ctx);
+#else
   rc = RSA_sign(type,
 		data, (unsigned int)data_len,
 		signature, &signature_len, key);
+#endif
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
   RSA_free(key);
+#endif
   if ( rc != 1 )
   { PL_free(signature);
     return raise_ssl_error(ERR_get_error());
@@ -1230,7 +1574,7 @@ pl_rsa_verify(term_t Public, term_t Type, term_t Enc,
 	    term_t Data, term_t Signature)
 { unsigned char *data;
   size_t data_len;
-  RSA *key;
+  RSAKEY *key;
   unsigned char *signature;
   size_t signature_len;
   int rc;
@@ -1241,11 +1585,26 @@ pl_rsa_verify(term_t Public, term_t Type, term_t Enc,
        !get_digest_type(Type, &type) ||
        !PL_get_nchars(Signature, &signature_len, (char**)&signature, REP_ISO_LATIN_1|CVT_LIST|CVT_EXCEPTION) )
     return FALSE;
-
+#if defined USE_EVP_API
+  OSSL_PARAM params[2];
+  params[0] = OSSL_PARAM_construct_utf8_string("digest", (char *)OBJ_nid2ln(type), 0);
+  params[1] = OSSL_PARAM_construct_end();
+  EVP_PKEY_CTX *verify_ctx = EVP_PKEY_CTX_new(key, NULL);
+  EVP_PKEY_verify_init_ex(verify_ctx, params);
+  rc = EVP_PKEY_verify(verify_ctx,
+                       signature, (unsigned int)signature_len,
+                       data, (unsigned int)data_len);
+  EVP_PKEY_CTX_free(verify_ctx);
+#else
   rc = RSA_verify(type,
                   data, (unsigned int)data_len,
                   signature, (unsigned int)signature_len, key);
+#endif
+#ifdef USE_EVP_API
+  EVP_PKEY_free(key);
+#else
   RSA_free(key);
+#endif
 
   if ( rc == 0 || rc == 1 )
     return rc;
@@ -1499,7 +1858,16 @@ pl_crypto_is_prime(term_t tprime, term_t tnchecks)
 
   if ( ( ctx = BN_CTX_new() ) &&
        get_bn_arg(1, tprime, &prime) )
-  { ret = BN_is_prime_ex(prime, nchecks, ctx, NULL);
+  {
+#ifdef HAVE_BN_CHECK_PRIME
+  // Note that we ignore nchecks here. BN_check_prime is listed as the replacement for BN_is_prime_ex
+  // but I could not find any information *anywhere* about how to pass in nchecks. Looking at the code in openssl
+  // it appears that they take some steps to correctly determine the value of nchecks. See
+  //     hhttps://github.com/openssl/openssl/blob/26b3e44a661899f0d0cb709482170cc411a94233/crypto/bn/bn_prime.c#L247
+  ret = BN_check_prime(prime, ctx, NULL);
+#else
+  ret = BN_is_prime_ex(prime, nchecks, ctx, NULL);
+#endif
   }
 
   BN_free(prime);
@@ -1701,9 +2069,9 @@ pl_crypto_curve_generator(term_t tcurve, term_t tx, term_t ty)
 
   if ( ( x = BN_new() ) &&
        ( y = BN_new() ) &&
-       EC_POINT_get_affine_coordinates_GFp(curve->group,
-                                           EC_GROUP_get0_generator(curve->group),
-                                           x, y, curve->ctx) &&
+       EC_POINT_get_affine_coordinates(curve->group,
+                                       EC_GROUP_get0_generator(curve->group),
+                                       x, y, curve->ctx) &&
        ( xhex = BN_bn2hex(x) ) &&
        ( yhex = BN_bn2hex(y) ) )
   { rc = PL_unify_chars(tx, PL_STRING|REP_ISO_LATIN_1, strlen(xhex), xhex)
@@ -1745,12 +2113,12 @@ pl_crypto_curve_scalar_mult(term_t tcurve, term_t ts,
        get_bn_arg(1, tx, &x) &&
        get_bn_arg(1, ty, &y)  &&
        ( q = EC_POINT_new(curve->group) ) &&
-       EC_POINT_set_affine_coordinates_GFp(curve->group, q, x, y, curve->ctx) &&
+       EC_POINT_set_affine_coordinates(curve->group, q, x, y, curve->ctx) &&
        ( r = EC_POINT_new(curve->group) ) &&
        EC_POINT_mul(curve->group, r, NULL, q, s, curve->ctx) &&
        ( a = BN_new() ) &&
        ( b = BN_new() ) &&
-       EC_POINT_get_affine_coordinates_GFp(curve->group, r, a, b, curve->ctx) &&
+       EC_POINT_get_affine_coordinates(curve->group, r, a, b, curve->ctx) &&
        ( ahex = BN_bn2hex(a) ) &&
        ( bhex = BN_bn2hex(b) ) )
   { rc = PL_unify_chars(ta, PL_STRING|REP_ISO_LATIN_1, strlen(ahex), ahex)
