@@ -47,10 +47,10 @@
 inline ConsoleEdit *wid2con(QWidget *w) { return qobject_cast<ConsoleEdit*>(w); }
 
 pqMainWindow::pqMainWindow(QWidget *parent) :
-    QMainWindow(parent), menu2pl(0) {
+    QMainWindow(parent) {
 #ifndef __APPLE__
-  QMenuBar *mb = menuBar();
-  mb->setNativeMenuBar(false);
+    QMenuBar *mb = menuBar();
+    mb->setNativeMenuBar(false);
 #endif
 }
 
@@ -62,8 +62,6 @@ pqMainWindow::pqMainWindow(int argc, char *argv[]) {
     QMenuBar *mb = menuBar();
     mb->setNativeMenuBar(false);
 #endif
-    // dispatch signals indexed
-    menu2pl = new cSignalMapper;
 
     setCentralWidget(new ConsoleEdit(argc, argv));
 
@@ -93,6 +91,8 @@ void pqMainWindow::closeEvent(QCloseEvent *event) {
 
     if (!SwiPrologEngine::quit_request())
         event->ignore();
+else
+    qApp->quit();
 }
 
 /** pass initialization script to actual interface
@@ -185,27 +185,35 @@ void pqMainWindow::remConsole(ConsoleEdit *c) {
 /** handle menu dispatch by means of Qt signal mapper
  */
 void pqMainWindow::addActionPq(ConsoleEdit *ce, QMenu *cmmenu, QString label, QString action) {
-    QAction *a = cmmenu->addAction(label, menu2pl, SLOT(map()));
+    // dispatch signals indexed
+    QAction *a = cmmenu->addAction(label, menuMapper(ce), SLOT(map()));
     menu2pl->setMapping(a, action);
-    if (0 == menu2pl->receivers(SIGNAL(mapped(const QString &))))
-        connect(menu2pl, SIGNAL(mapped(const QString &)), ce, SLOT(onConsoleMenuActionMap(const QString &)));
 }
 
 /** ditto
  */
 QAction* pqMainWindow::add_action(ConsoleEdit *ce, QMenu *mn, QString Label, QString ctxtmod, QString Goal, QAction *before) {
-
     QAction *a;
     if (!before)
-        a = mn->addAction(Label, menu2pl, SLOT(map()));
+        a = mn->addAction(Label, menuMapper(ce), SLOT(map()));
     else {
         mn->insertAction(before, a = new QAction(Label, mn));
-        connect(a, SIGNAL(triggered()), menu2pl, SLOT(map()));
+        connect(a, SIGNAL(triggered()), menuMapper(ce), SLOT(map()));
     }
     menu2pl->setMapping(a, ctxtmod + ':' + Goal);
-
-    if (0 == menu2pl->receivers(SIGNAL(mapped(const QString &))))
-        connect(menu2pl, SIGNAL(mapped(const QString &)), ce, SLOT(onConsoleMenuActionMap(QString)));
-
     return a;
 }
+
+/** 20/12/2021 - factorize code subject to Qt versioning
+ */
+QSignalMapper* pqMainWindow::menuMapper(ConsoleEdit *ce) {
+    if (!menu2pl) {
+        menu2pl = new QSignalMapper;
+#if QT_VERSION < 0x051500
+        connect(menu2pl, SIGNAL(mapped(const QString &)), ce, SLOT(onConsoleMenuActionMap(const QString &)));
+#else
+        connect(menu2pl, &QSignalMapper::mappedString, ce, &ConsoleEdit::onConsoleMenuActionMap);
+#endif
+    }
+    return menu2pl;
+}
