@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker and Anjo Anjewierden
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (c)  1985-2020, University of Amsterdam
+    Copyright (c)  1985-2021, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -53,7 +54,6 @@
 	     delete_breakpoint/1,
 	     manpce/1,
 	     prolog_ide/1,
-	     rational/1,
 	     spypce/1,
 	     tracepce/1,
 	     atomic_list_concat/2,
@@ -77,7 +77,6 @@
 	     get_dict/3,
 	     nb_setarg/3,
              pi_head/2,
-	     rational/3,
 	     send_list/3,
 	     setup_call_cleanup/3,
 	     stream_position_data/3,
@@ -86,6 +85,12 @@
 	     get_dict/5,
 	     sub_string/5
 	   ]).
+:- if(current_prolog_flag(bounded, false)).
+:- require([ rational/1,
+             rational/3
+           ]).
+:- endif.
+
 :- autoload(library(pldoc/doc_process), [comment_modes/2]).
 
 resource(mode_pl_icon, image, image('32x32/doc_pl.xpm')).
@@ -278,7 +283,7 @@ expand_path(Term, D) :-
 :- pce_group(indent).
 
 :- pce_global(@prolog_neck_regex,
-              new(regex(':-|:->|:<-|-->'))).
+              new(regex(':-|:->|:<-|-->|=>|\\?=>'))).
 :- pce_global(@prolog_full_stop,
               new(regex('[^-#$&*+./:<=>?@\\\\^`~]\\.($|\\s)'))).
 :- pce_global(@prolog_decl_regex,
@@ -960,7 +965,18 @@ clause_pi((Head :- _), PI) :-
     callable(Head),
     Head \= (_:_),
     pi_head(PI, Head).
+clause_pi((Left => _), PI) :-
+    !,
+    nonvar(Left),
+    (   Left = (Head,_Guard)
+    ->  true
+    ;   Head = Left
+    ),
+    callable(Head),
+    Head \= (_:_),
+    pi_head(PI, Head).
 clause_pi((Head --> _), Name//Arity) :-
+    !,
     callable(Head),
     Head \= (_:_),
     pi_head(Name/Arity, Head).
@@ -2274,6 +2290,12 @@ set_fragment_context(Fragment, P, Context) :-
     ).
 set_fragment_context(_, _, _).
 
+:- if(\+ current_predicate(rational/1)).
+rational(_) :- fail.
+rational(_,_,_) :- fail.
+:- endif.
+
+
 context_arg(1, context).
 context_arg(2, context_2).
 
@@ -3007,6 +3029,18 @@ elements_to_string(List, String) :-
 element_to_string(Fmt-Args, String) :-
     !,
     format(string(String), Fmt, Args).
+element_to_string(ansi(_Style, Fmt, Args), String) :-
+    !,
+    format(string(String), Fmt, Args).
+element_to_string(url(File:Line:Pos), String) :-
+    !,
+    format(string(String), '~w:~w:~w', [File, Line, Pos]).
+element_to_string(url(File:Line), String) :-
+    !,
+    format(string(String), '~w:~w', [File, Line]).
+element_to_string(url(File), String) :-
+    !,
+    format(string(String), '~w', [File]).
 element_to_string(nl, '\n') :- !.
 element_to_string(Atom, Atom).
 
