@@ -2,7 +2,7 @@
 
     Author:        Jan Wielemaker and Anjo Anjewierden
     E-mail:        J.Wielemaker@vu.nl
-    WWW:           http://www.swi-prolog.org/projects/xpce/
+    WWW:           http://www.swi-prolog.org/packages/xpce/
     Copyright (c)  2001-2016, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
@@ -549,6 +549,7 @@ clause_end(ClauseRef, File, CharA, CharZ) :-
     pce_clause_info(ClauseRef, File, TPos, _),
     nonvar(TPos),
     arg(2, TPos, CharA),
+    nonvar(CharA),
     CharZ is CharA + 1.
 
 head_pos(Ref, Pos, HPos) :-
