@@ -58,8 +58,8 @@
 		 *******************************/
 
 #ifndef __pce_export
-#if defined(PCE_INCLUDED) && (defined(WIN32) || defined(__CYWIN32__))
-#define __pce_export __declspec(dllexport)
+#if defined(PCE_INCLUDED) && defined(WIN32)
+#define __pce_export __declspec(dllexport) extern
 #else
 #define __pce_export extern
 #endif /*PCE_INCLUDED*/
