@@ -73,8 +73,8 @@
 #define PUBLIC_GLOBAL extern		/* exported global variables */
 #endif
 
-#if defined(WIN32) || defined(__CYGWIN__)
-#define __pce_export __declspec(dllexport)
+#if defined(WIN32)
+#define __pce_export __declspec(dllexport) extern
 #else
 #define __pce_export extern
 #endif
@@ -293,6 +293,8 @@ cpdata(to, from, type, n)
 #define toString	toStringPCE		/* SWI-Prolog name-conflict */
 #define valReal		valPceReal		/* and another */
 
+#undef hyper
+
 		 /*******************************
 		 *	    BASIC TYPES		*
 		 *******************************/
@@ -504,7 +506,7 @@ test, conversion and computation macro's are provided.
   string _s_ ## name ## _hdr; \
   void  *_s_ ## name ## _buf = (void *)alloca(iswide ? (size) * sizeof(charW) \
 						   : (size) * sizeof(charA)); \
-  String name = fstr_inithdr(&_s_ ## name ## _hdr, iswide, _s_ ## name ## _buf, size)
+  PceString name = fstr_inithdr(&_s_ ## name ## _hdr, iswide, _s_ ## name ## _buf, size)
 
 		/********************************
 		*         OBJECT HEADER		*
@@ -553,8 +555,8 @@ test, conversion and computation macro's are provided.
 #define F_ISHOSTDATA		makeFlag(23) /* instanceOf(x, ClassHostData) */
 #define F_NOTANY		makeFlag(24) /* Not acceptable to any/object */
 
-#define OBJ_MAGIC		((uintptr_t)(0x94L << 25))
-#define OBJ_MAGIC_MASK		((uintptr_t)(0xfeL << 25))
+#define OBJ_MAGIC		((uintptr_t)0x14 << 25)
+#define OBJ_MAGIC_MASK		((uintptr_t)0x7e << 25)
 
 #define hasObjectMagic(obj)	((((Instance)(obj))->flags&OBJ_MAGIC_MASK) == \
 					OBJ_MAGIC)
@@ -1436,7 +1438,7 @@ struct class_definition
 #ifndef NO_BUILT_IN_DECL
 extern struct name builtin_names[];	/* object-array of built-in's */
 #endif
-#include "names.ih"			/* #defines for code used names */
+#include "h/names.ih"			/* #defines for code used names */
 
 #define isName(name)	(isObject(name) && onFlag((name), F_ISNAME))
 #define notName(name)	(!isName(name))
