@@ -701,9 +701,14 @@ postEventWindow(PceWindow sw, EventObj ev)
     goto out;
   }
 
-  if ( sw->focus != (Graphical) sw || notNil(sw->focus_recogniser) )
-  { rval = send(sw, NAME_event, ev, EAV);
-  }
+  /* This code looks a bit awkward, but prevents a -Warray-bounds
+   * warning from gcc-11
+   */
+  if ( notNil(sw->focus) )
+  { if ( sw->focus != (Graphical) sw || notNil(sw->focus_recogniser) )
+      rval = send(sw, NAME_event, ev, EAV);
+  } else
+    rval = send(sw, NAME_event, ev, EAV);
 
   if ( !rval )
   { ScrollBar sb;
