@@ -35,10 +35,11 @@
 :- module(test_aggregate,
 	  [ test_aggregate/0
 	  ]).
-
-
 :- use_module(library(plunit)).
 :- use_module(library(aggregate)).
+:- use_module(library(dif)).
+:- use_module(library(hashtable)).
+:- use_module(library(lists)).
 
 test_aggregate :-
 	run_tests([ foreach,
@@ -51,6 +52,18 @@ test(forall, true) :-
 	foreach(between(1, 10, X), integer(X)).
 test(forall, fail) :-
 	foreach(between(1, 2, X), _=X).
+test(dif1, Y == 5) :-
+    foreach(between(1,4,X), dif(X,Y)), Y = 5.
+test(dif1, fail) :-
+    foreach(between(1,4,X), dif(X,Y)), Y = 3.
+test(member, L =@= [1,2,3,4|_]) :-
+    foreach(between(1,4,X), member(X,L)),
+    !.
+test(ht, Pairs == [1-1,2-2,3-3,4-4,5-5]) :-
+    ht_new(HT),
+    foreach(between(1,5,X),
+            ht_put(HT,X,X)),
+    ht_pairs(HT, Pairs).
 
 :- end_tests(foreach).
 
