@@ -43,6 +43,10 @@
 #include <unistd.h>
 #endif
 
+#ifdef _MSC_VER
+#pragma warning(disable : 4996)	/* deprecate open() etc */
+#endif
+
 #ifndef MAXPATHLEN
 #define MAXPATHLEN 1024
 #endif
@@ -51,16 +55,13 @@
 #include "pl-hash.c"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-This program creates pl-codetable.c, pl-jumptable.ic   and pl-vmi.h from
-pl-vmi.c.
+This program creates vmi-metadata.h from pl-vmi.c.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 const char *program;
 int verbose = 0;
 const char *vmi_file    = "pl-vmi.c";
-const char *ctable_file = "pl-codetable.ic";
-const char *jump_table  = "pl-jumptable.ic";
-const char *vmi_hdr	= "pl-vmi.h";
+const char *vmi_hdr	= "vmi-metadata.h";
 
 #define MAX_VMI 1000
 
@@ -69,10 +70,13 @@ typedef struct				/* VMI( */
   char *flags;				/* Flags (VIF_*) */
   char *argc;				/* Argument length (or VM_DYNARGC) */
   char *args;				/* Argument types (max 3) */
+  char *argn;				/* VMH ONLY: Argument names */
+  int is_vmh;				/* 0 for VMI, 1 for VMH */
 } vmi;					/* ) */
 
 vmi vmi_list[MAX_VMI];
 int vmi_count = 0;
+int vmh_count = 0;
 
 char *synopsis;
 size_t syn_size = 0;
@@ -111,11 +115,12 @@ skip_ws(const char *s)
 
 
 static char *
-skip_id(const char *s)
+skip_id(const char *s, int allow_lower)
 { if ( s )
   { for (; *s; s++)
     { if ( *s == '_' ||
 	   (*s >= 'A' && *s <= 'Z') ||
+	   (*s >= 'a' && *s <= 'z' && allow_lower) ||
 	   (*s >= '0' && *s <= '9') )
 	continue;
 
@@ -131,7 +136,7 @@ static char *
 skip_flags(const char *s)
 { if ( s )
   { for (; *s; s++)
-    { if ( *s == '_' || *s == '|' ||
+    { if ( *s == '_' || *s == '|' || *s == '+' ||
 	   (*s >= 'A' && *s <= 'Z') ||
 	   (*s >= '0' && *s <= '9') )
 	continue;
@@ -180,32 +185,42 @@ load_vmis(const char *file)
     while(fgets(buf, sizeof(buf), fd))
     { line++;
 
-      if ( strncmp(buf, "VMI(", 4) == 0 )
-      { const char *s1 = skip_ws(buf+4);
-	const char *e1 = skip_id(s1);
-	const char *s2 = skip_ws(skip_over(e1, ','));
-	const char *e2 = skip_flags(s2);
-	const char *s3 = skip_ws(skip_over(e2, ','));
+      if ( strncmp(buf, "VMI(", 4) == 0 || strncmp(buf, "VMH(", 4) == 0 )
+      { const int is_vmh = buf[2] == 'H';
+        const char *s1 = skip_ws(buf+4);
+	const char *e1 = skip_id(s1, is_vmh);
+	const char *s2 = is_vmh ? NULL : skip_ws(skip_over(e1, ','));
+	const char *e2 = is_vmh ? NULL : skip_flags(s2);
+	const char *s3 = skip_ws(skip_over(is_vmh ? e1 : e2, ','));
 	const char *e3 = skip_flags(s3);
 	const char *s4 = skip_over(skip_ws(skip_over(e3, ',')), '(');
 	const char *e4 = skip_over(s4, ')');
+	const char *s5 = is_vmh ? skip_over(skip_ws(skip_over(e4, ',')), '(') : NULL;
+	const char *e5 = is_vmh ? skip_over(s5, ')') : NULL;
 
-	if ( !e4 )
+	if ( !e4 || (is_vmh && !e5) )
 	{ fprintf(stderr, "Syntax error at %s:%d\n", file, line);
 	  exit(1);
 	} else
-	  e4--;				/* backspace over ) */
+	{ e4--;				/* backspace over ) */
+	  e5 -= is_vmh;
+	}
 
 	vmi_list[vmi_count].name  = my_strndup(s1, e1-s1);
-	vmi_list[vmi_count].flags = my_strndup(s2, e2-s2);
+	vmi_list[vmi_count].flags = is_vmh ? NULL : my_strndup(s2, e2-s2);
 	vmi_list[vmi_count].argc  = my_strndup(s3, e3-s3);
 	vmi_list[vmi_count].args  = my_strndup(s4, e4-s4);
+	vmi_list[vmi_count].argn  = is_vmh ? my_strndup(s5, e5-s5) : NULL;
+	vmi_list[vmi_count].is_vmh = is_vmh;
 
-	add_synopsis(s1, e1-s1);	/* flags (s2) isn't needed for VM signature */
-	add_synopsis(s3, e3-s3);
-	add_synopsis(s4, e4-s4);
+	if (!is_vmh)			/* VMH pseudo-instructions aren't part of signature */
+	{ add_synopsis(s1, e1-s1);	/* flags (s2) isn't needed for VM signature */
+	  add_synopsis(s3, e3-s3);
+	  add_synopsis(s4, e4-s4);
+	}
 
 	vmi_count++;
+	vmh_count += is_vmh;
       }
     }
 
@@ -272,102 +287,58 @@ mystrlwr(char *to, const char *from)	/* my*: Windows name conflict */
   return to;
 }
 
-
 static int
-emit_code_table(const char *to)
+emit_vmi_hdr(const char *to)
 { const char *tmp = "vmi.tmp";
   FILE *out = fopen(tmp, "w");
   int i;
 
   fprintf(out, "/*  File: %s\n\n", to);
-  fprintf(out, "    This file provides a description of the virtual machine instructions\n");
-  fprintf(out, "    and their arguments. It  is  used by  pl-comp.c  to facilitate the\n");
-  fprintf(out, "    compiler and decompiler, as well as pl-wic.c to save/reload sequences\n");
-  fprintf(out, "    of virtual machine instructions.\n");
+  fprintf(out, "    This file provides a set of defines describing the virtual machine\n");
+  fprintf(out, "    instructions and their arguments. These are used by other source files\n");
+  fprintf(out, "    to define various attributes and structures. See pl-vmi.h for more details.\n");
   fprintf(out, "\n");
   fprintf(out, "    Note: this file is generated by %s from %s.  DO NOT EDIT", program, vmi_file);
   fprintf(out, "    \n");
   fprintf(out, "*/\n\n");
-  fprintf(out, "#include \"pl-incl.h\"\n\n");
-  fprintf(out, "const code_info codeTable[] = {\n");
-  fprintf(out, "  /* {name, ID, flags, #args, argtype} */\n");
+
+  fprintf(out, "#ifndef _IN_PL_VMI_H\n");
+  fprintf(out, "#error \"Do not include <%s> directly! Include \\\"pl-vmi.h\\\"\"\n", to);
+  fprintf(out, "#endif\n\n");
+
+  fprintf(out, "#define VM_SIGNATURE 0x%x\n", MurmurHashAligned2(synopsis, syn_size, 0x12345678));
+#define PRINT_LIST(macro, condition) do \
+  { int written = 0; \
+    fprintf(out, "#define " #macro "(sep,f,...)"); \
+    for(i=0; i<vmi_count; i++) \
+    { if (condition) \
+        fprintf(out, "%s f(%s, ## __VA_ARGS__)", written++ > 0 ? " sep()" : "", vmi_list[i].name); \
+    } \
+    fprintf(out, "\n"); \
+  } while(0)
+  PRINT_LIST(FOREACH_VMI_CALL, !vmi_list[i].is_vmh);
+  PRINT_LIST(FOREACH_VMH_CALL, vmi_list[i].is_vmh);
+
+  fprintf(out, "/* Instruction data */\n\n");
 
   for(i=0; i<vmi_count; i++)
   { char name[100];
+    vmi *Vmi = vmi_list + i;
+    fprintf(out, "#define VMLCASE_%s     %s\n", Vmi->name, mystrlwr(name, Vmi->name));
 
-    fprintf(out, "  {\"%s\", %s, %s, %s, {%s}},\n",
-	    mystrlwr(name, vmi_list[i].name),
-	    vmi_list[i].name,
-	    vmi_list[i].flags,
-	    vmi_list[i].argc,
-	    vmi_list[i].args[0] ? vmi_list[i].args : "0");
+    if (Vmi->is_vmh)
+    { fprintf(out, "#define VMHDECL_%s     %s, %s, (%s), (%s)\n", Vmi->name, Vmi->name, Vmi->argc, Vmi->args, Vmi->argn);
+      fprintf(out, "#define VMHARGNAMES_%s %s\n", Vmi->name, Vmi->argn);
+    }
+    else
+    { fprintf(out, "#define VMIDECL_%s     %s, %s, %s, (%s)\n", Vmi->name, Vmi->name, Vmi->flags, Vmi->argc, Vmi->args);
+      fprintf(out, "#define VMIFLAGS_%s    %s\n", Vmi->name, Vmi->flags);
+    }
+    
+    fprintf(out, "#define VM%cARGCOUNT_%s %s\n", Vmi->is_vmh ? 'H' : 'I', Vmi->name, Vmi->argc);
+    fprintf(out, "#define VM%cARGTYPES_%s %s\n\n", Vmi->is_vmh ? 'H' : 'I', Vmi->name, Vmi->args);
   }
 
-  fprintf(out, "  { NULL, 0, 0, 0, {0} }\n");
-  fprintf(out, "};\n");
-  fclose(out);
-
-  return update_file(tmp, to);
-}
-
-
-static int
-emit_jump_table(const char *to)
-{  const char *tmp = "vmi.tmp";
-  FILE *out = fopen(tmp, "w");
-  int i;
-
-  fprintf(out, "/*  File: %s\n\n", to);
-  fprintf(out, "    This file provides the GCC-2 jump-labels to exploit GCC's\n");
-  fprintf(out, "    support for threaded code.\n");
-  fprintf(out, "\n");
-  fprintf(out, "    Note: this file is generated by %s from %s.  DO NOT EDIT", program, vmi_file);
-  fprintf(out, "    \n");
-  fprintf(out, "*/\n\n");
-
-  fprintf(out, "static void *jmp_table[] =\n");
-  fprintf(out, "{\n");
-
-
-  for(i=0; i<vmi_count; i++)
-  { fprintf(out, "  &&%s_LBL,\n", vmi_list[i].name);
-  }
-
-  fprintf(out, "  NULL\n");
-  fprintf(out, "};\n");
-
-  fclose(out);
-
-  return update_file(tmp, to);
-}
-
-
-static int
-emit_code_defs(const char *to)
-{  const char *tmp = "vmi.tmp";
-  FILE *out = fopen(tmp, "w");
-  int i;
-
-  fprintf(out, "/*  File: %s\n\n", to);
-  fprintf(out, "    This file provides the definition of type code.\n");
-  fprintf(out, "\n");
-  fprintf(out, "    Note: this file is generated by %s from %s.  DO NOT EDIT", program, vmi_file);
-  fprintf(out, "    \n");
-  fprintf(out, "*/\n\n");
-
-  fprintf(out, "typedef enum\n");
-  fprintf(out, "{\n");
-
-
-  for(i=0; i<vmi_count; i++)
-  { fprintf(out, "  %s,\n", vmi_list[i].name);
-  }
-
-  fprintf(out, "  VMI_END_LIST\n");
-  fprintf(out, "} vmi;\n\n");
-  fprintf(out, "#define I_HIGHEST ((int)VMI_END_LIST)\n");
-  fprintf(out, "#define VM_SIGNATURE 0x%x\n", MurmurHashAligned2(synopsis, syn_size, 0x12345678));
-
   fclose(out);
 
   return update_file(tmp, to);
@@ -387,7 +358,7 @@ main(int argc, char **argv)
     verbose = 1;
   }
 
-  if ( argc == 1 )
+  if ( argc >= 1 )
     snprintf(buf, sizeof(buf), "%s/%s", argv[0], vmi_file);
   else
     snprintf(buf, sizeof(buf), "%s", vmi_file);
@@ -395,11 +366,9 @@ main(int argc, char **argv)
 
   load_vmis(buf);
   if ( verbose )
-    fprintf(stderr, "Found %d VMs\n", vmi_count);
+    fprintf(stderr, "Found %d VMs and %d helpers\n", vmi_count - vmh_count, vmh_count);
 
-  if ( emit_code_table(ctable_file) == 0 &&
-       emit_jump_table(jump_table) == 0 &&
-       emit_code_defs(vmi_hdr) == 0 )
+  if ( emit_vmi_hdr(vmi_hdr) == 0 )
     return 0;
   else
     return 1;
