@@ -33,13 +33,10 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-#ifdef _WIN64
-#define _WIN32_WINNT 0x0501            /* get RtlCaptureContext() */
-#endif
-
 #define _GNU_SOURCE
-#include "pl-incl.h"
-#include "os/pl-cstack.h"
+#include "../pl-incl.h"
+#include "pl-cstack.h"
+#include "../pl-setup.h"
 #include <time.h>
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -209,9 +206,7 @@ print_backtrace(int last)		/* 1..SAVE_TRACES */
 
 void
 bstore_print_backtrace_named(btrace *bt, const char *why)
-{ int done = 0;
-
-  if ( bt )
+{ if ( bt )
   { int me = bt->current-1;
 
     for(;;)
@@ -219,15 +214,12 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 	me += SAVE_TRACES;
       if ( bt->dumps[me].name && strcmp(bt->dumps[me].name, why) == 0 )
       { print_trace(bt, me);
-	done++;
+	return;
       }
       if ( --me == bt->current-1 )
 	break;
     }
   }
-
-  if ( !done )
-    Sdprintf("No backtrace named %s\n", why);
 }
 
 #endif /*HAVE_LIBUNWIND*/
@@ -379,7 +371,7 @@ addr2line(const char *fname, uintptr_t offset, char *buf, size_t size)
 
       *o = '\0';
 
-      fclose(fd);
+      pclose(fd);
       return o > buf;
     }
   }
@@ -441,11 +433,9 @@ print_backtrace(int last)		/* 1..SAVE_TRACES */
 }
 
 
-void
+static void
 bstore_print_backtrace_named(btrace *bt, const char *why)
-{ int done = 0;
-
-  if ( bt )
+{ if ( bt )
   { int me = bt->current-1;
 
     for(;;)
@@ -453,15 +443,14 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 	me += SAVE_TRACES;
       if ( bt->why[me] && strcmp(bt->why[me], why) == 0 )
       { print_trace(bt, me);
-	done++;
+	return;
       }
       if ( --me == bt->current-1 )
+      { Sdprintf("No backtrace named %s\n", why);
 	break;
+      }
     }
   }
-
-  if ( !done )
-    Sdprintf("No backtrace named %s\n", why);
 }
 
 
@@ -477,7 +466,7 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 void
 initBackTrace(void)
 {
-#ifdef SIGSEGV
+#if defined(SIGSEGV) && !defined(O_C_STACK_GUARDED)
   PL_signal(SIGSEGV, sigCrashHandler);
 #endif
 #ifdef SIGILL
@@ -826,9 +815,7 @@ print_backtrace(int last)		/* 1..SAVE_TRACES */
 
 void
 bstore_print_backtrace_named(btrace *bt, const char *why)
-{ int done = 0;
-
-  if ( bt )
+{ if ( bt )
   { int me = bt->current-1;
 
     for(;;)
@@ -836,15 +823,12 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 	me += SAVE_TRACES;
       if ( bt->dumps[me].name && strcmp(bt->dumps[me].name, why) == 0 )
       { print_trace(bt, me);
-	done++;
+	return;
       }
       if ( --me == bt->current-1 )
 	break;
     }
   }
-
-  if ( !done )
-    Sdprintf("No backtrace named %s\n", why);
 }
 
 static LONG WINAPI crashHandler(PEXCEPTION_POINTERS pExceptionInfo)
@@ -896,7 +880,20 @@ sigCrashHandler(int sig)
   time_t now = time(NULL);
   char tbuf[48];
 
-  signal(sig, SIG_DFL);
+  signal(sig,     SIG_DFL);
+#ifdef SIGALRM
+  signal(SIGALRM, SIG_DFL);
+#endif
+#ifdef SIGABRT
+  signal(SIGABRT, SIG_DFL);
+#endif
+#ifdef SIGSEGV
+  signal(SIGSEGV, SIG_DFL);
+#endif
+#ifdef HAVE_ALARM
+  alarm(10);				/* try to avoid deadlocks */
+#endif
+
   tid = PL_thread_self();
   ctime_r(&now, tbuf);
   tbuf[24] = '\0';
@@ -908,6 +905,8 @@ sigCrashHandler(int sig)
 	   "received fatal signal %d (%s)\n",
 	   PL_thread_self(), name, tbuf, sig, signal_name(sig));
   print_c_backtrace("crash");
+  Sdprintf("Prolog stack:\n");
+  PL_backtrace(25, PL_BT_SAFE);
   Sdprintf("Running on_halt hooks with status %d\n", 128+sig);
   run_on_halt(&GD->os.exit_hooks, 128+sig);
 
