@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2015, University of Amsterdam
+    Copyright (c)  2011-2022, University of Amsterdam
                               VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -33,9 +34,11 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-#include "pl-incl.h"
-#include <ctype.h>
 #include "pl-ctype.h"
+#include "pl-text.h"
+#include "../pl-read.h"
+#include "../pl-fli.h"
+#include <ctype.h>
 #include <errno.h>
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -211,6 +214,7 @@ mkfunction(iswalpha)
 mkfunction(isascii)
 mkfunction(iswcntrl)
 mkfunction(iswdigit)
+mkfunction(iswprint)
 mkfunction(iswgraph)
 mkfunction(iswlower)
 mkfunction(iswupper)
@@ -231,6 +235,7 @@ static const char_type char_types[] =
   { ATOM_white,			     iswhite },
   { ATOM_cntrl,			     fiswcntrl },
   { ATOM_digit,			     fiswdigit },
+  { ATOM_print,			     fiswprint },
   { ATOM_graph,			     fiswgraph },
   { ATOM_lower,			     fiswlower },
   { ATOM_upper,			     fiswupper },
@@ -268,7 +273,7 @@ char_type_by_name(atom_t name, int arity)
 static int
 advanceGen(generator *gen)
 { if ( gen->do_enum & ENUM_CHAR )
-  { if ( ++gen->current == 256 )
+  { if ( ++gen->current > PLMAXWCHAR )
       fail;
   } else
   { gen->class++;
@@ -334,7 +339,7 @@ do_char_type(term_t chr, term_t class, control_t h, int how)
 
       if ( !(do_enum & ENUM_CLASS) )
       { if ( !PL_get_name_arity(class, &cn, &arity) ||
-	     !(cc = char_type_by_name(cn, arity)) )
+	     !(cc = char_type_by_name(cn, (int)arity)) )
 	  return PL_error("char_type", 2, NULL,
 			  ERR_TYPE, ATOM_char_type, class);
       }
@@ -524,8 +529,9 @@ get_chr_from_text(const PL_chars_t *t, size_t index)
 }
 
 
+#define modify_case_atom(in, out, down, text_type) LDFUNC(modify_case_atom, in, out, down, text_type)
 static foreign_t
-modify_case_atom(term_t in, term_t out, int down, int text_type ARG_LD)
+modify_case_atom(DECL_LD term_t in, term_t out, int down, int text_type)
 { PL_chars_t tin, tout;
 
   if ( !PL_get_text(in, &tin, CVT_ATOMIC|CVT_EXCEPTION) )
@@ -619,28 +625,28 @@ modify_case_atom(term_t in, term_t out, int down, int text_type ARG_LD)
 static
 PRED_IMPL("downcase_atom", 2, downcase_atom, 0)
 { PRED_LD
-  return modify_case_atom(A1, A2, TRUE, PL_ATOM PASS_LD);
+  return modify_case_atom(A1, A2, TRUE, PL_ATOM);
 }
 
 
 static
 PRED_IMPL("upcase_atom", 2, upcase_atom, 0)
 { PRED_LD
-  return modify_case_atom(A1, A2, FALSE, PL_ATOM PASS_LD);
+  return modify_case_atom(A1, A2, FALSE, PL_ATOM);
 }
 
 
 static
 PRED_IMPL("string_lower", 2, string_lower, 0)
 { PRED_LD
-  return modify_case_atom(A1, A2, TRUE, PL_STRING PASS_LD);
+  return modify_case_atom(A1, A2, TRUE, PL_STRING);
 }
 
 
 static
 PRED_IMPL("string_upper", 2, string_upper, 0)
 { PRED_LD
-  return modify_case_atom(A1, A2, FALSE, PL_STRING PASS_LD);
+  return modify_case_atom(A1, A2, FALSE, PL_STRING);
 }
 
 
@@ -881,20 +887,28 @@ initEncoding(void)
   { if ( !LD->encoding )
     { char *enc;
 
+#if __APPLE__
+      if ( !getenv("LANG") && !getenv("LC_CTYPE") )
+	putenv("LC_CTYPE=UTF-8");
+#endif
+
       if ( !init_locale() )
       { LD->encoding = ENC_ISO_LATIN_1;
       } else if ( (enc = setlocale(LC_CTYPE, NULL)) )
-      { LD->encoding = ENC_ANSI;		/* text encoding */
+      { char *ext = strchr(enc, '.');
+	const enc_map *m;
 
-	if ( (enc = strchr(enc, '.')) )
-	{ const enc_map *m;
-	  enc++;				/* skip '.' */
+	LD->encoding = ENC_ANSI;		/* text encoding */
 
-	  for ( m=map; m->name; m++ )
-	  { if ( strcmp(enc, m->name) == 0 )
-	    { LD->encoding = m->encoding;
-	      break;
-	    }
+	if ( ext )
+	  ext++;
+	else
+	  ext = enc;
+
+	for ( m=map; m->name; m++ )
+	{ if ( strcmp(enc, m->name) == 0 )
+	  { LD->encoding = m->encoding;
+	    break;
 	  }
 	}
       } else
