@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2017, University of Amsterdam
+    Copyright (c)  2011-2022, University of Amsterdam
                               VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -38,10 +39,15 @@ Formatted output (Prolog predicates format/[1,2,3]).   One  day,  the  C
 source should also use format() to produce error messages, etc.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#include "pl-incl.h"
-#include "pl-arith.h"
+#include "pl-fmt.h"
 #include "pl-ctype.h"
 #include "pl-utf8.h"
+#include "../pl-arith.h"
+#include "../pl-proc.h"
+#include "../pl-fli.h"
+#include "../pl-prims.h"
+#include "../pl-write.h"
+#include "../pl-pro.h"
 #include <ctype.h>
 #include <stdio.h>
 
@@ -165,7 +171,7 @@ oututf8(format_state *state, const char *s, size_t len)
   while(s<e)
   { int chr;
 
-    s = utf8_get_char(s, &chr);
+    PL_utf8_code_point(&s, e, &chr);
     if ( !outchr(state, chr) )
       return FALSE;
   }
@@ -210,7 +216,7 @@ static int	update_column(int, Char);
 static bool	do_format(IOSTREAM *fd, PL_chars_t *fmt,
 			  int ac, term_t av, Module m);
 static void	distribute_rubber(struct rubber *, int, int);
-static int	emit_rubber(format_state *state);
+static WUNUSED int emit_rubber(format_state *state);
 
 
 		/********************************
@@ -373,6 +379,66 @@ get_chr_from_text(const PL_chars_t *t, int index)
 }
 
 
+typedef struct sub_state
+{ char          buf[BUFSIZE];
+  char         *str;
+  size_t        bufsize;
+  format_state *fstate;
+  IOSTREAM     *old_stream;
+} sub_state;
+
+static int
+prepare_sub_format(sub_state *state, format_state *fstate, IOSTREAM *fd)
+{ state->fstate = fstate;
+
+  if ( !fstate->pending_rubber &&
+       fd->position &&
+       fd->position->linepos == fstate->column )
+  { state->old_stream = Scurout;
+
+    Scurout = fd;
+  } else
+  { state->old_stream = NULL;
+    state->str        = state->buf;
+    state->bufsize    = sizeof(state->buf);
+
+    tellString(&state->str, &state->bufsize, ENC_UTF8);
+    if ( true(fd, SIO_ISATTY) )
+      set(Scurout, SIO_ISATTY);
+  }
+
+  return TRUE;
+}
+
+static int
+end_sub_format(sub_state *state, int rc)
+{ int lp = Scurout->position->linepos;
+
+  if ( state->old_stream )
+  { Scurout = state->old_stream;
+
+    if ( rc )
+      state->fstate->column = lp;
+  } else
+  { int lp = Scurout->position->linepos;
+
+    toldString();
+    if ( rc )
+    { int c0 = state->fstate->column;
+      rc = oututf8(state->fstate, state->str, state->bufsize);
+      state->fstate->column = c0 + lp;
+    }
+    if ( state->str != state->buf )
+      free(state->str);
+  }
+
+  return rc;
+}
+
+
+
+
+
 		/********************************
 		*       ACTUAL FORMATTING	*
 		********************************/
@@ -427,7 +493,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	    }
 	  } else if ( c == '*' )
 	  { NEED_ARG;
-	    if ( PL_get_integer(argv, &arg) )
+	    if ( PL_get_integer(argv, &arg) && arg >= 0 )
 	    { SHIFT;
 	    } else
 	      FMT_ERROR("no or negative integer for `*' argument");
@@ -447,9 +513,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	       (proc = lookupHTable(format_predicates, (void*)((intptr_t)c))) )
 	  { size_t arity;
 	    term_t av;
-	    char buf[BUFSIZE];
-	    char *str = buf;
-	    size_t bufsize = BUFSIZE;
+	    sub_state sstate;
 	    int i;
 
 	    PL_predicate_info(proc, NULL, &arity, NULL);
@@ -466,13 +530,11 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	      SHIFT;
 	    }
 
-	    tellString(&str, &bufsize, ENC_UTF8);
+	    if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+	      goto out;
 	    rc = PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, proc, av);
-	    toldString();
-	    if ( rc )
-	      rc = oututf8(&state, str, bufsize);
-	    if ( str != buf )
-	      free(str);
+	    rc = end_sub_format(&sstate, rc);
+
 	    if ( !rc )
 	      goto out;
 
@@ -523,7 +585,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  PL_locale *l;
 
 		  NEED_ARG;
-		  if ( !valueExpression(argv, &n PASS_LD) )
+		  if ( !valueExpression(argv, &n) )
 		  { char f[2];
 
 		    f[0] = c;
@@ -557,7 +619,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  tmp_buffer b;
 
 		  NEED_ARG;
-		  if ( !valueExpression(argv, &i PASS_LD) ||
+		  if ( !valueExpression(argv, &i) ||
 		       !toIntegerNumber(&i, 0) )
 		  { char f[2];
 
@@ -600,7 +662,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		      FMT_EXEPTION();
 		  } else			/* r,R */
 		  { if ( arg == DEFAULT )
-		      FMT_ERROR("r,R requires radix specifier");
+		      arg = 8;
 		    if ( arg < 2 || arg > 36 )
 		    { term_t r = PL_new_term_ref();
 
@@ -623,10 +685,16 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		{ PL_chars_t txt;
 
 		  NEED_ARG;
-		  if ( !PL_get_text(argv, &txt, CVT_LIST|CVT_STRING) &&
-		       !PL_get_text(argv, &txt, CVT_ATOM) ) /* SICStus compat */
+		  if ( !PL_get_text(argv, &txt, CVT_ATOM|CVT_LIST|CVT_STRING) )
 		    FMT_ARG("s", argv);
+		  if ( arg != DEFAULT )
+		  { if ( arg < 0 )
+		      arg = 0;
+		    if ( arg < txt.length )
+		      txt.length = arg;
+		  }
 		  rc = outtext(&state, &txt);
+		  PL_free_text(&txt);
 		  SHIFT;
 		  if ( !rc )
 		    goto out;
@@ -640,8 +708,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  break;
 		}
 		{ Func f;
-		  char buf[BUFSIZE];
-		  char *str;
+		  sub_state sstate;
 
 	      case 'k':			/* write_canonical */
 		  f = pl_write_canonical;
@@ -657,121 +724,54 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  pl_common:
 
 		  NEED_ARG;
-		  if ( state.pending_rubber )
-		  { size_t bufsize = BUFSIZE;
 
-		    str = buf;
-		    tellString(&str, &bufsize, ENC_UTF8);
-		    rc = (*f)(argv);
-		    toldString();
-		    if ( rc )
-		      rc = oututf8(&state, str, bufsize);
-		    if ( str != buf )
-		      free(str);
-		    if ( !rc )
-		      goto out;
-		  } else
-		  { if ( fd->position &&
-			 fd->position->linepos == state.column )
-		    { IOSTREAM *old = Scurout;
+		  if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+		    goto out;
+		  rc = (int)(*f)(argv);
+		  rc = end_sub_format(&sstate, rc);
 
-		      Scurout = fd;
-		      rc = (int)(*f)(argv);
-		      Scurout = old;
-		      if ( !rc )
-			goto out;
+		  if ( !rc )
+		   goto out;
 
-		      state.column = fd->position->linepos;
-		    } else
-		    { size_t bufsize = BUFSIZE;
-
-		      str = buf;
-		      tellString(&str, &bufsize, ENC_UTF8);
-		      rc = (*f)(argv);
-		      toldString();
-		      if ( rc )
-			rc = oututf8(&state, str, bufsize);
-		      if ( str != buf )
-			free(str);
-		      if ( !rc )
-			goto out;
-		    }
-		  }
 		  SHIFT;
 		  here++;
 		  break;
 		}
 	      case 'W':			/* write_term(Value, Options) */
-	       { char buf[BUFSIZE];
-		 char *str;
+	       { sub_state sstate;
 
 		 if ( argc < 2 )
 		 { FMT_ERROR("not enough arguments");
 		 }
-		 if ( state.pending_rubber )
-		  { size_t bufsize = BUFSIZE;
 
-		    str = buf;
-		    tellString(&str, &bufsize, ENC_UTF8);
-		    rc = (int)pl_write_term(argv, argv+1);
-		    toldString();
-		    if ( rc )
-		      rc = oututf8(&state, str, bufsize);
-		    if ( str != buf )
-		      free(str);
-		    if ( !rc )
-		      goto out;
-		  } else
-		  { if ( fd->position &&
-			 fd->position->linepos == state.column )
-		    { IOSTREAM *old = Scurout;
+		 if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+		   goto out;
+		 rc = (int)pl_write_term(argv, argv+1);
+		 rc = end_sub_format(&sstate, rc);
 
-		      Scurout = fd;
-		      rc = (int)pl_write_term(argv, argv+1);
-		      Scurout = old;
-		      if ( !rc )
-			goto out;
+		 if ( !rc )
+		   goto out;
 
-		      state.column = fd->position->linepos;
-		    } else
-		    { size_t bufsize = BUFSIZE;
-
-		      str = buf;
-		      tellString(&str, &bufsize, ENC_UTF8);
-		      rc = (int)pl_write_term(argv, argv+1);
-		      toldString();
-		      if ( rc )
-			rc = oututf8(&state, str, bufsize);
-		      if ( str != buf )
-			free(str);
-		      if ( !rc )
-			goto out;
-		    }
-		  }
-		  SHIFT;
-		  SHIFT;
-		  here++;
-		  break;
+		 SHIFT;
+		 SHIFT;
+		 here++;
+		 break;
 	       }
 	      case '@':
-	        { char buf[BUFSIZE];
-		  char *str = buf;
-		  size_t bufsize = BUFSIZE;
-		  term_t ex = 0;
+	        { term_t ex = 0;
+		  sub_state sstate;
 
 		  if ( argc < 1 )
 		  { FMT_ERROR("not enough arguments");
 		  }
-		  tellString(&str, &bufsize, ENC_UTF8);
+
+		  if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+		    goto out;
 		  rc = callProlog(m, argv, PL_Q_CATCH_EXCEPTION, &ex);
-		  toldString();
-		  if ( rc )
-		    rc = oututf8(&state, str, bufsize);
-		  if ( str != buf )
-		    free(str);
+		  rc = end_sub_format(&sstate, rc);
 
 		  if ( !rc )
-		  { if ( ex )
+		  { if ( ex && !PL_exception(0) )
 		      rc = PL_raise_exception(ex);
 		    goto out;
 		  }
@@ -816,14 +816,22 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		}
 	      case '|':			/* set tab */
 		{ int stop;
+		  int nl_and_reindent;
 
 		  if ( arg == DEFAULT )
 		    arg = state.column;
+		  /*FALLTHROUGH*/
 	      case '+':			/* tab relative */
 		  if ( arg == DEFAULT )
 		    arg = 8;
 		  stop = (c == '+' ? tab_stop + arg : arg);
 
+		  if ( stop < state.column && mod_colon )
+		    nl_and_reindent = state.pending_rubber ?
+						state.rub[state.pending_rubber-1].pad : ' ';
+		  else
+		    nl_and_reindent = 0;
+
 		  if ( state.pending_rubber == 0 ) /* nothing to distribute */
 		  { state.rub[0].where = state.buffered;
 		    state.rub[0].pad = ' ';
@@ -832,7 +840,26 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  distribute_rubber(state.rub,
 				    state.pending_rubber,
 				    stop - state.column);
-		  emit_rubber(&state);
+		  if ( !(rc=emit_rubber(&state)) )
+		    goto out;
+
+		  if ( nl_and_reindent )
+		  { if ( Sputcode('\n', state.out) < 0 )
+		    { rc = FALSE;
+		      goto out;
+		    }
+		    state.column = update_column(state.column, '\n');
+
+		    state.rub[0].where = state.buffered;
+		    state.rub[0].pad = nl_and_reindent;
+		    state.pending_rubber++;
+
+		    distribute_rubber(state.rub,
+				      state.pending_rubber,
+				      stop - state.column);
+		    if ( !(rc=emit_rubber(&state)) )
+		      goto out;
+		  }
 
 		  state.column = tab_stop = stop;
 		  here++;
@@ -861,7 +888,9 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
   }
 
   if ( state.pending_rubber )		/* not closed ~t: flush out */
-    emit_rubber(&state);
+  { if ( !(rc=emit_rubber(&state)) )
+      goto out;
+  }
   if ( argc != 0 )
     FMT_ERROR("too many arguments");
 
@@ -892,7 +921,7 @@ distribute_rubber(struct rubber *r, int rn, int space)
 }
 
 
-static int
+static WUNUSED int
 emit_rubber(format_state *state)
 { const char *s = baseBuffer(&state->buffer, char);
   const char *e = &s[entriesBuffer(&state->buffer, char)];
@@ -915,7 +944,7 @@ emit_rubber(format_state *state)
     }
 
     if ( s < e )
-    { s = utf8_get_char(s, &chr);
+    { PL_utf8_code_point(&s, e, &chr);
       if ( Sputcode(chr, state->out) < 0 )
 	return FALSE;
     } else
