@@ -3,25 +3,38 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2013, VU University Amsterdam
+    Copyright (c)  2013-2015, VU University Amsterdam
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include "pl-incl.h"
 #include "pl-locale.h"
+#include "../pl-fli.h"
 #include <errno.h>
 
 #if defined(__sun) || __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__ < 1070
@@ -32,9 +45,6 @@
 
 #include <locale.h>
 
-#define LOCK()   PL_LOCK(L_LOCALE)	/* MT locking */
-#define UNLOCK() PL_UNLOCK(L_LOCALE)
-
 #undef LD				/* fetch LD once per function */
 #define LD LOCAL_LD
 
@@ -157,18 +167,15 @@ update_locale(PL_locale *l, int category, const char *locale)
 
 static int
 alias_locale(PL_locale *l, atom_t alias)
-{ int rc;
+{ GET_LD
+  int rc;
 
-  LOCK();
+  PL_LOCK(L_LOCALE);
 
   if ( !GD->locale.localeTable )
     GD->locale.localeTable = newHTable(16);
 
-  if ( addHTable(GD->locale.localeTable, (void*)alias, l) )
-  { l->alias = alias;
-    PL_register_atom(alias);
-    rc = TRUE;
-  } else
+  if ( lookupHTable(GD->locale.localeTable, (void*)alias) )
   { GET_LD
     term_t obj = PL_new_term_ref();
 
@@ -176,7 +183,13 @@ alias_locale(PL_locale *l, atom_t alias)
     rc = PL_error("locale_create", 2, "Alias name already taken",
 		  ERR_PERMISSION, ATOM_create, ATOM_locale, obj);
   }
-  UNLOCK();
+  else
+  { addNewHTable(GD->locale.localeTable, (void*)alias, l);
+    l->alias = alias;
+    PL_register_atom(alias);
+    rc = TRUE;
+  }
+  PL_UNLOCK(L_LOCALE);
 
   return rc;
 }
@@ -214,12 +227,12 @@ static int
 release_locale_ref(atom_t aref)
 { locale_ref *ref = PL_blob_data(aref, NULL, NULL);
 
-  LOCK();
+  PL_LOCK(L_LOCALE);
   if ( ref->data->references == 0 )
     free_locale(ref->data);
   else
     ref->data->symbol = 0;
-  UNLOCK();
+  PL_UNLOCK(L_LOCALE);
 
   return TRUE;
 }
@@ -298,10 +311,7 @@ getLocale(term_t t, PL_locale **lp)
     } else if ( (ref=PL_blob_data(a, NULL, &bt)) && bt == &locale_blob )
     { l = ref->data;
     } else if ( GD->locale.localeTable )
-    { Symbol s;
-
-      if ( (s=lookupHTable(GD->locale.localeTable, (void*)a)) )
-	l = s->value;
+    { l = lookupHTable(GD->locale.localeTable, (void*)a);
     }
 
     if ( l )
@@ -323,9 +333,11 @@ getLocaleEx(term_t t, PL_locale **lp)
     return TRUE;
 
   if ( PL_is_atom(t) )
-    return PL_existence_error("locale", t);
+    PL_existence_error("locale", t);
   else
-    return PL_type_error("locale", t);
+    PL_type_error("locale", t);
+
+  return FALSE;
 }
 
 
@@ -333,32 +345,36 @@ getLocaleEx(term_t t, PL_locale **lp)
 		 *	 PROLOG BINDING		*
 		 *******************************/
 
+#define locale_alias_property(l, prop) LDFUNC(locale_alias_property, l, prop)
 static int		/* locale_property(Mutex, alias(Name)) */
-locale_alias_property(PL_locale *l, term_t prop ARG_LD)
+locale_alias_property(DECL_LD PL_locale *l, term_t prop)
 { if ( l->alias )
     return PL_unify_atom(prop, l->alias);
 
   return FALSE;
 }
 
+#define locale_decimal_point_property(l, prop) LDFUNC(locale_decimal_point_property, l, prop)
 static int		/* locale_property(Locale, decimal_point(Atom)) */
-locale_decimal_point_property(PL_locale *l, term_t prop ARG_LD)
+locale_decimal_point_property(DECL_LD PL_locale *l, term_t prop)
 { if ( l->decimal_point && l->decimal_point[0] )
     return PL_unify_wchars(prop, PL_ATOM, (size_t)-1, l->decimal_point);
 
   return FALSE;
 }
 
+#define locale_thousands_sep_property(l, prop) LDFUNC(locale_thousands_sep_property, l, prop)
 static int		/* locale_property(Locale, thousands_sep(Atom)) */
-locale_thousands_sep_property(PL_locale *l, term_t prop ARG_LD)
+locale_thousands_sep_property(DECL_LD PL_locale *l, term_t prop)
 { if ( l->thousands_sep && l->thousands_sep[0] )
     return PL_unify_wchars(prop, PL_ATOM, (size_t)-1, l->thousands_sep);
 
   return FALSE;
 }
 
+#define locale_grouping_property(l, prop) LDFUNC(locale_grouping_property, l, prop)
 static int		/* locale_property(Locale, grouping(List)) */
-locale_grouping_property(PL_locale *l, term_t prop ARG_LD)
+locale_grouping_property(DECL_LD PL_locale *l, term_t prop)
 { if ( l->grouping && l->grouping[0] )
   { term_t tail = PL_copy_term_ref(prop);
     term_t head = PL_new_term_ref();
@@ -383,54 +399,22 @@ locale_grouping_property(PL_locale *l, term_t prop ARG_LD)
 }
 
 
-typedef struct
-{ functor_t functor;			/* functor of property */
-  int (*function)();			/* function to generate */
-} lprop;
-
-static const lprop lprop_list [] =
-{ { FUNCTOR_alias1,	    locale_alias_property },
-  { FUNCTOR_decimal_point1, locale_decimal_point_property },
-  { FUNCTOR_thousands_sep1, locale_thousands_sep_property },
-  { FUNCTOR_grouping1,      locale_grouping_property },
+static const tprop lprop_list [] =
+{ { FUNCTOR_alias1,	    LDFUNC_REF(locale_alias_property) },
+  { FUNCTOR_decimal_point1, LDFUNC_REF(locale_decimal_point_property) },
+  { FUNCTOR_thousands_sep1, LDFUNC_REF(locale_thousands_sep_property) },
+  { FUNCTOR_grouping1,      LDFUNC_REF(locale_grouping_property) },
   { 0,			    NULL }
 };
 
 typedef struct
 { TableEnum e;				/* Enumerator on mutex-table */
   PL_locale *l;				/* current locale */
-  const lprop *p;			/* Pointer in properties */
+  const tprop *p;			/* Pointer in properties */
   int enum_properties;			/* Enumerate the properties */
 } lprop_enum;
 
 
-static int
-get_prop_def(term_t t, atom_t expected, const lprop *list, const lprop **def)
-{ GET_LD
-  functor_t f;
-
-  if ( PL_get_functor(t, &f) )
-  { const lprop *p = list;
-
-    for( ; p->functor; p++ )
-    { if ( f == p->functor )
-      { *def = p;
-        return TRUE;
-      }
-    }
-
-    PL_error(NULL, 0, NULL, ERR_DOMAIN, expected, t);
-    return -1;
-  }
-
-  if ( PL_is_variable(t) )
-    return 0;
-
-  PL_error(NULL, 0, NULL, ERR_TYPE, expected, t);
-  return -1;
-}
-
-
 static int
 advance_lstate(lprop_enum *state)
 { if ( state->enum_properties )
@@ -441,10 +425,10 @@ advance_lstate(lprop_enum *state)
     state->p = lprop_list;
   }
   if ( state->e )
-  { Symbol s;
+  { PL_locale *l;
 
-    if ( (s = advanceTableEnum(state->e)) )
-    { state->l = s->value;
+    if ( advanceTableEnum(state->e, NULL, (void**)&l) )
+    { state->l = l;
 
       return TRUE;
     }
@@ -498,10 +482,10 @@ PRED_IMPL("locale_property", 2, locale_property, PL_FA_NONDETERMINISTIC)
 
 	    if ( state->p->functor == FUNCTOR_alias1 &&
 		 get_atom_arg(property, &alias) )
-	    { Symbol s;
+	    { PL_locale *l;
 
-	      if ( (s=lookupHTable(GD->locale.localeTable, (void*)alias)) )
-		return unifyLocale(locale, s->value, FALSE);
+	      if ( (l = lookupHTable(GD->locale.localeTable, (void*)alias)) )
+		return unifyLocale(locale, l, FALSE);
 	      else
 		return FALSE;
 	    }
@@ -546,11 +530,11 @@ PRED_IMPL("locale_property", 2, locale_property, PL_FA_NONDETERMINISTIC)
 
 enumerate:
   if ( !state->l )			/* first time, enumerating locales */
-  { Symbol s;
+  { PL_locale *l;
 
     assert(state->e);
-    if ( (s=advanceTableEnum(state->e)) )
-    { state->l = s->value;
+    if ( advanceTableEnum(state->e, NULL, (void**)&l) )
+    { state->l = l;
     } else
     { freeTableEnum(state->e);
       assert(state != &statebuf);
@@ -564,7 +548,7 @@ enumerate:
       _PL_get_arg(1, property, arg);
 
     for(;;)
-    { if ( (*state->p->function)(state->l, arg PASS_LD) )
+    { if ( LDFUNCP(*state->p->function)(state->l, arg) )
       { if ( state->enum_properties )
 	{ if ( !PL_unify_term(property,
 			      PL_FUNCTOR, state->p->functor,
@@ -629,7 +613,7 @@ get_group_size_ex(term_t t, int *s)
     { *s = i;
       return TRUE;
     }
-    return PL_domain_error("digit_group_size", t);
+    PL_domain_error("digit_group_size", t);
   }
 
   return FALSE;
@@ -687,32 +671,35 @@ set_grouping(term_t t, char **valp)
 static
 PRED_IMPL("locale_create", 3, locale_create, 0)
 { PRED_LD
-  PL_locale *def, *new;
+  PL_locale *def, *new = NULL;
   char *lname;
 
-  if ( PL_get_chars(A2, &lname, CVT_LIST|CVT_STRING|REP_MB) )
-  { const char *old;
-
-    LOCK();
-    if ( (old=setlocale(LC_NUMERIC, lname)) )
-    { new = new_locale(NULL);
-      setlocale(LC_NUMERIC, old);
-    } else
-    { assert(0);				/* keep compiler happy */
-      return FALSE;
-    }
-    UNLOCK();
-    if ( !old )
-    { if ( errno == ENOENT )
-	return PL_existence_error("locale", A2);
-      else
-	return PL_error(NULL, 0, MSG_ERRNO, ERR_SYSCALL, "setlocale");
-    }
-  } else
-  { if ( !getLocaleEx(A2, &def) )
-      return FALSE;
-    new = new_locale(def);
+  if ( getLocale(A2, &def) )
+  { new = new_locale(def);
     releaseLocale(def);
+  } else
+  { if ( PL_get_chars(A2, &lname, CVT_LIST|CVT_STRING|CVT_ATOM|REP_MB) )
+    { if ( strcmp(lname, "default") == 0 )
+      { new = new_locale(NULL);
+      } else
+      { const char *old;
+
+	PL_LOCK(L_LOCALE);
+	if ( (old=setlocale(LC_NUMERIC, lname)) )
+	{ new = new_locale(NULL);
+	  setlocale(LC_NUMERIC, old);
+	}
+	PL_UNLOCK(L_LOCALE);
+	if ( !old )
+	{ if ( errno == ENOENT )
+	    return PL_existence_error("locale", A2);
+	  else
+	    return PL_error(NULL, 0, MSG_ERRNO, ERR_SYSCALL, "setlocale");
+	}
+      }
+    } else
+    { return getLocaleEx(A2, &def);
+    }
   }
 
   if ( new )
@@ -723,7 +710,7 @@ PRED_IMPL("locale_create", 3, locale_create, 0)
 
     while(PL_get_list_ex(tail, head, tail))
     { atom_t pname;
-      int parity;
+      size_t parity;
 
       if ( !PL_get_name_arity(head, &pname, &parity) ||
 	   parity != 1 ||
@@ -764,19 +751,19 @@ PRED_IMPL("locale_create", 3, locale_create, 0)
 
 static
 PRED_IMPL("locale_destroy", 1, locale_destroy, 0)
-{ PL_locale *l;
+{ GET_LD
+  PL_locale *l;
 
   if ( getLocaleEx(A1, &l) )
   { if ( l->alias )
-    { Symbol s;
-      atom_t alias = l->alias;
+    { atom_t alias = l->alias;
 
-      LOCK();
-      if ( (s=lookupHTable(GD->locale.localeTable, (void*)alias)) )
-	deleteSymbolHTable(GD->locale.localeTable, s);
+      PL_LOCK(L_LOCALE);
+      if ( lookupHTable(GD->locale.localeTable, (void*)alias) )
+	deleteHTable(GD->locale.localeTable, (void*)alias);
       l->alias = 0;
       PL_unregister_atom(alias);
-      UNLOCK();
+      PL_UNLOCK(L_LOCALE);
     }
 
     releaseLocale(l);
@@ -897,9 +884,9 @@ initStreamLocale(IOSTREAM *s)
 
 PL_locale *
 acquireLocale(PL_locale *l)
-{ LOCK();
+{ PL_LOCK(L_LOCALE);
   l->references++;
-  UNLOCK();
+  PL_UNLOCK(L_LOCALE);
 
   return l;
 }
@@ -907,10 +894,10 @@ acquireLocale(PL_locale *l)
 
 void
 releaseLocale(PL_locale *l)
-{ LOCK();
+{ PL_LOCK(L_LOCALE);
   if ( --l->references == 0 && l->symbol == 0 && l->alias == 0 )
     free_locale(l);
-  UNLOCK();
+  PL_UNLOCK(L_LOCALE);
 }
 
 
