@@ -3,8 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2017, University of Amsterdam
+    Copyright (c)  2011-2021, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -34,15 +36,27 @@
 */
 
 /*#define O_DEBUG 1*/
-#include "pl-incl.h"
+#ifdef __WINDOWS__
+#include <winsock2.h>
+#include <windows.h>
+#include <process.h>			/* getpid() */
+#endif
+#include "pl-prologflag.h"
 #include "pl-ctype.h"
+#include "../pl-arith.h"
+#include "../pl-tabling.h"
+#include "../pl-fli.h"
+#include "../pl-write.h"
+#include "../pl-pro.h"
+#include "../pl-wam.h"
+#include "../pl-trace.h"
+#include "../pl-setup.h"
+#include "../pl-modul.h"
+#include "../pl-version.h"
 #include <ctype.h>
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #endif
-#ifdef __WINDOWS__
-#include <process.h>			/* getpid() */
-#endif
 #include <time.h>
 
 
@@ -87,7 +101,7 @@ static void initPrologFlagTable(void);
 
 typedef struct _prolog_flag
 { short		flags;			/* Type | Flags */
-  short		index;			/* index in PLFLAG_ mask */
+  short		index;			/* index in LD->prolog_flag.mask */
   union
   { atom_t	a;			/* value as atom */
     int64_t	i;			/* value as integer */
@@ -109,21 +123,6 @@ following arguments are to be provided:
     FT_TERM	a term
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int
-indexOfBoolMask(unsigned int mask)
-{ int i=1;
-
-  if ( !mask )
-    return -1;
-
-  while(!(mask & 0x1))
-  { i++;
-    mask >>= 1;
-  }
-  return i;
-}
-
-
 void
 setPrologFlag(const char *name, int flags, ...)
 { GET_LD
@@ -144,7 +143,7 @@ setPrologFlag(const char *name, int flags, ...)
       return;
   } else
   { f = allocHeapOrHalt(sizeof(*f));
-    f->index = -1;
+    f->index = 0;
     f->flags = flags;
     addNewHTable(GD->prolog_flag.table, (void *)an, f);
     first_def = TRUE;
@@ -154,25 +153,23 @@ setPrologFlag(const char *name, int flags, ...)
   switch(type)
   { case FT_BOOL:
     { int           val = va_arg(args, int);
-      unsigned int mask = va_arg(args, unsigned int);
+      unsigned int flag = va_arg(args, unsigned int);
 
-      if ( !first_def && mask && f->index < 0 )	/* type definition */
-      { f->index = indexOfBoolMask(mask);
+      if ( !first_def && flag && !f->index )	/* type definition */
+      { f->index = flag;
 	val = (f->value.a == ATOM_true);
       } else if ( first_def )			/* 1st definition */
-      { f->index = indexOfBoolMask(mask);
+      { f->index = flag;
 	DEBUG(MSG_PROLOG_FLAG,
-	      Sdprintf("Prolog flag %s at 0x%08lx\n", name, mask));
+	      Sdprintf("Prolog flag %s at %d\n", name, flag));
       }
 
       f->value.a = (val ? ATOM_true : ATOM_false);
-      if ( f->index >= 0 )
-      { mask = (unsigned int)1 << (f->index-1);
-
-	if ( val )
-	  setPrologFlagMask(mask);
+      if ( f->index )
+      { if ( val )
+	  setPrologFlagMask(f->index);
 	else
-	  clearPrologFlagMask(mask);
+	  clearPrologFlagMask(f->index);
       }
       break;
     }
@@ -332,6 +329,31 @@ setBackQuotes(atom_t a, unsigned int *flagp)
 }
 
 
+int
+setRationalSyntax(atom_t a, unsigned int *flagp)
+{ GET_LD
+  unsigned int flags;
+
+  if	  ( a == ATOM_natural )
+    flags = RAT_NATURAL;
+  else if ( a == ATOM_compatibility )
+    flags = RAT_COMPAT;
+  else
+  { term_t value = PL_new_term_ref();
+
+    PL_put_atom(value, a);
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN,
+		    ATOM_rational_syntax, value);
+  }
+
+  *flagp &= ~RAT_MASK;
+  *flagp |= flags;
+
+  succeed;
+}
+
+
+
 static int
 setUnknown(term_t value, atom_t a, Module m)
 { unsigned int flags = m->flags & ~(UNKNOWN_MASK);
@@ -368,17 +390,28 @@ setUnknown(term_t value, atom_t a, Module m)
 }
 
 
+static int
+checkOnError(term_t value, atom_t a, atom_t key)
+{ if ( a == ATOM_print || a == ATOM_halt || a == ATOM_status )
+    return TRUE;
+  return PL_error(NULL, 0, NULL, ERR_DOMAIN, key, value);
+}
+
+
+
 static int
 setFileNameCaseHandling(atom_t a)
 { GET_LD
 
   if ( a == ATOM_case_sensitive )
-  { setPrologFlagMask(PLFLAG_FILE_CASE|PLFLAG_FILE_CASE_PRESERVING);
+  { setPrologFlagMask(PLFLAG_FILE_CASE);
+    setPrologFlagMask(PLFLAG_FILE_CASE_PRESERVING);
   } else if ( a == ATOM_case_preserving )
   { setPrologFlagMask(PLFLAG_FILE_CASE_PRESERVING);
     clearPrologFlagMask(PLFLAG_FILE_CASE);
   } else if ( a == ATOM_case_insensitive )
-  { clearPrologFlagMask(PLFLAG_FILE_CASE|PLFLAG_FILE_CASE_PRESERVING);
+  { clearPrologFlagMask(PLFLAG_FILE_CASE);
+    clearPrologFlagMask(PLFLAG_FILE_CASE_PRESERVING);
   } else
   { term_t value = PL_new_term_ref();
 
@@ -395,16 +428,13 @@ static atom_t
 currentFileNameCaseHandling(void)
 { GET_LD
 
-  switch ( LD->prolog_flag.mask.flags &
-	   (PLFLAG_FILE_CASE|PLFLAG_FILE_CASE_PRESERVING) )
-  { case 0:
-      return ATOM_case_insensitive;
-    case PLFLAG_FILE_CASE_PRESERVING:
+  if ( truePrologFlag(PLFLAG_FILE_CASE) )
+  { if ( truePrologFlag(PLFLAG_FILE_CASE_PRESERVING) )
       return ATOM_case_preserving;
-    case PLFLAG_FILE_CASE|PLFLAG_FILE_CASE_PRESERVING:
-      return ATOM_case_sensitive;
-    default:
-      return ATOM_unknown;
+    else
+      return ATOM_case_insensitive;
+  } else
+  { return ATOM_case_sensitive;
   }
 }
 
@@ -512,6 +542,46 @@ setStreamTypeCheck(atom_t a)
   return TRUE;
 }
 
+
+static int
+setAutoload(atom_t a)
+{ GET_LD
+
+  if ( a == ATOM_false )
+    clearPrologFlagMask(PLFLAG_AUTOLOAD);
+  else if ( a == ATOM_explicit ||
+	    a == ATOM_true ||
+	    a == ATOM_user ||
+	    a == ATOM_user_or_explicit )
+    setPrologFlagMask(PLFLAG_AUTOLOAD);
+  else
+  { term_t value = PL_new_term_ref();
+
+    PL_put_atom(value, a);
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_autoload, value);
+  }
+
+  return TRUE;
+}
+
+
+#define propagateAutoload(val) LDFUNC(propagateAutoload, val)
+static int
+propagateAutoload(DECL_LD term_t val)
+{ if ( !GD->bootsession )
+  { predicate_t pred;
+    term_t av;
+
+    pred = PL_predicate("set_autoload", 1, "$autoload");
+    return ( (av=PL_new_term_refs(2)) &&
+	     PL_put_term(av+0, val) &&
+	     PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, pred, av) );
+  } else
+  { return TRUE;
+  }
+}
+
+
 #if O_XOS
 typedef struct access_id
 { char *name;
@@ -556,19 +626,12 @@ get_win_file_access_check(void)
 }
 #endif
 
+#define set_prolog_flag_unlocked(m, k, value, flags) LDFUNC(set_prolog_flag_unlocked, m, k, value, flags)
 static word
-set_prolog_flag_unlocked(term_t key, term_t value, int flags)
-{ GET_LD
-  atom_t k;
-  prolog_flag *f;
-  Module m = MODULE_parse;
+set_prolog_flag_unlocked(DECL_LD Module m, atom_t k, term_t value, int flags)
+{ prolog_flag *f;
   int rval = TRUE;
 
-  if ( !PL_strip_module(key, &m, key) )
-    return FALSE;
-  if ( !PL_get_atom(key, &k) )
-    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, key);
-
 					/* set existing Prolog flag */
 #ifdef O_PLMT
   if ( LD->prolog_flag.table &&
@@ -581,8 +644,18 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
   { if ( flags & FF_KEEP )
       return TRUE;
     if ( (f->flags&FF_READONLY) && !(flags&FF_FORCE) )
-      return PL_error(NULL, 0, NULL, ERR_PERMISSION,
-		      ATOM_modify, ATOM_flag, key);
+    { term_t key;
+
+      return ( (key = PL_new_term_ref()) &&
+	       PL_put_atom(key, k) &&
+	       PL_error(NULL, 0, NULL, ERR_PERMISSION,
+			ATOM_modify, ATOM_flag, key) );
+    }
+
+    if ( tbl_is_restraint_flag(k) )
+      return tbl_set_restraint_flag(value, k);
+    if ( is_arith_flag(k) )
+      return set_arith_flag(value, k);
 
 #ifdef O_PLMT
     if ( GD->statistics.threads_created > 1 )
@@ -610,7 +683,7 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
   anyway:
     PL_register_atom(k);
     f = allocHeapOrHalt(sizeof(*f));
-    f->index = -1;
+    f->index = 0;
 
     switch( (flags & FT_MASK) )
     { case FT_FROM_VALUE:
@@ -694,9 +767,13 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 
     if ( PL_current_prolog_flag(ATOM_user_flags, PL_ATOM, &how) )
     { if ( how == ATOM_error )
-	return PL_error(NULL, 0, NULL, ERR_EXISTENCE,
-			ATOM_prolog_flag, key);
-      else if ( how == ATOM_warning )
+      { term_t key;
+
+	return ( (key = PL_new_term_ref()) &&
+		 PL_put_atom(key, k) &&
+		 PL_error(NULL, 0, NULL, ERR_EXISTENCE,
+			  ATOM_prolog_flag, key) );
+      } else if ( how == ATOM_warning )
 	Sdprintf("WARNING: Flag %s: new Prolog flags must be created using "
 		 "create_prolog_flag/3\n", stringAtom(k));
     }
@@ -724,24 +801,20 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 	  clear(m, M_VARPREFIX);
       } else if ( k == ATOM_debug )
       { if ( val )
-	{ debugmode(DBG_ALL, NULL);
+	{ rval = debugmode(DBG_ALL, NULL);
 	} else
-	{ tracemode(FALSE, NULL);
-	  debugmode(DBG_OFF, NULL);
+	{ rval = ( tracemode(FALSE, NULL) &&
+		   debugmode(DBG_OFF, NULL) );
 	}
       } else if ( k == ATOM_debugger_show_context )
       { debugstatus.showContext = val;
 #ifdef O_PLMT
       } else if ( k == ATOM_threads )
-      { if ( val )
+      { if ( !!val != !!GD->thread.enabled )
 	{ rval = enableThreads(val);
-	  PL_LOCK(L_PLFLAG);
-	} else
-	{ PL_UNLOCK(L_PLFLAG);
-	  rval = enableThreads(val);
+	  if ( !rval )
+	    break;			/* do not change value */
 	}
-	if ( !rval )
-	  break;			/* don't change value */
 #endif
       } else if ( k == ATOM_tty_control )
       { if ( val != (f->value.a == ATOM_true) )
@@ -752,6 +825,8 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 	    PushTty(Sinput, &ttytab, TTY_SAVE);
 	  }
 	}
+      } else if ( k == ATOM_debug_on_interrupt )
+      {	rval = enable_debug_on_interrupt(val);
       } else if ( k == ATOM_protect_static_code )
       { if ( val != (f->value.a == ATOM_true) && val == FALSE )
 	{ term_t ex;
@@ -763,13 +838,11 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 	}
       }
 					/* set the flag value */
-      if ( f->index > 0 )
-      { unsigned int mask = (unsigned int)1 << (f->index-1);
-
-	if ( val )
-	  setPrologFlagMask(mask);
+      if ( f->index && rval )
+      { if ( val )
+	  setPrologFlagMask(f->index);
 	else
-	  clearPrologFlagMask(mask);
+	  clearPrologFlagMask(f->index);
       }
       f->value.a = (val ? ATOM_true : ATOM_false);
 
@@ -785,8 +858,12 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
       { rval = setDoubleQuotes(a, &m->flags);
       } else if ( k == ATOM_back_quotes )
       { rval = setBackQuotes(a, &m->flags);
+      } else if ( k == ATOM_rational_syntax )
+      { rval = setRationalSyntax(a, &m->flags);
       } else if ( k == ATOM_unknown )
       { rval = setUnknown(value, a, m);
+      } else if ( k == ATOM_on_error || k == ATOM_on_warning )
+      { rval = checkOnError(value, a, k);
       } else if ( k == ATOM_write_attributes )
       { rval = setWriteAttributes(a);
       } else if ( k == ATOM_occurs_check )
@@ -799,6 +876,10 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
       { rval = setStreamTypeCheck(a);
       } else if ( k == ATOM_file_name_case_handling )
       { rval = setFileNameCaseHandling(a);
+      } else if ( k == ATOM_autoload )
+      { rval = setAutoload(a);
+      } else if ( k == ATOM_table_monotonic )
+      { rval = setMonotonicMode(a);
 #if O_XOS
       } else if ( k == ATOM_win_file_access_check )
       { rval = set_win_file_access_check(value);
@@ -807,9 +888,11 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
       if ( !rval )
 	fail;
 
-      PL_unregister_atom(f->value.a);
-      f->value.a = a;
-      PL_register_atom(a);
+      if ( f->value.a != a )
+      { PL_unregister_atom(f->value.a);
+	f->value.a = a;
+	PL_register_atom(a);
+      }
       break;
     }
     case FT_INTEGER:
@@ -825,10 +908,26 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
       else
 #endif
       if ( k == ATOM_table_space )
-	LD->tabling.node_pool.limit = (size_t)i;
+      { if ( !LD->tabling.node_pool )
+	  LD->tabling.node_pool = new_alloc_pool("private_table_space", i);
+	else
+	  LD->tabling.node_pool->limit = (size_t)i;
+      }
+#ifdef O_PLMT
+      else if ( k == ATOM_shared_table_space )
+      { if ( !GD->tabling.node_pool )
+	{ alloc_pool *pool = new_alloc_pool("shared_table_space", i);
+	  if ( pool && !COMPARE_AND_SWAP_PTR(&GD->tabling.node_pool, NULL, pool) )
+	    free_alloc_pool(pool);
+	} else
+	  GD->tabling.node_pool->limit = (size_t)i;
+      }
+#endif
       else if ( k == ATOM_stack_limit )
       { if ( !set_stack_limit((size_t)i) )
 	  return FALSE;
+      } else if ( k == ATOM_string_stack_tripwire )
+      { LD->fli.string_buffers.tripwire = (unsigned int)i;
       }
       break;
     }
@@ -856,10 +955,22 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 
 int
 set_prolog_flag(term_t key, term_t value, int flags)
-{ int rc;
+{ GET_LD
+  atom_t k;
+  Module m = MODULE_parse;
+  int rc;
+
+  if ( !PL_strip_module(key, &m, key) ||
+       !PL_get_atom_ex(key, &k) )
+    return FALSE;
+
+  if ( k == ATOM_autoload && !propagateAutoload(value) )
+    return FALSE;
+  if ( k == ATOM_threads )
+    return set_prolog_flag_unlocked(m, k, value, flags);
 
   PL_LOCK(L_PLFLAG);
-  rc = set_prolog_flag_unlocked(key, value, flags);
+  rc = set_prolog_flag_unlocked(m, k, value, flags);
   PL_UNLOCK(L_PLFLAG);
 
   return rc;
@@ -1022,6 +1133,16 @@ unify_prolog_flag_value(Module m, atom_t key, prolog_flag *f, term_t val)
     else
       v = ATOM_symbol_char;
 
+    return PL_unify_atom(val, v);
+  } else if ( key == ATOM_rational_syntax )
+  { atom_t v;
+
+    switch(m->flags&RAT_MASK)
+    { case RAT_NATURAL: v = ATOM_natural;       break;
+      case RAT_COMPAT:  v = ATOM_compatibility; break;
+      default:		v = 0; assert(0);
+    }
+
     return PL_unify_atom(val, v);
   } else if ( key == ATOM_unknown )
   { atom_t v;
@@ -1060,15 +1181,16 @@ unify_prolog_flag_value(Module m, atom_t key, prolog_flag *f, term_t val)
   { return PL_unify_atom(val, accessLevel());
   } else if ( key == ATOM_stack_limit )
   { return PL_unify_int64(val, LD->stacks.limit);
+  } else if ( tbl_is_restraint_flag(key) )
+  { return tbl_get_restraint_flag(val, key) == TRUE;
+  } else if ( is_arith_flag(key) )
+  { return get_arith_flag(val, key) == TRUE;
   }
 
   switch(f->flags & FT_MASK)
   { case FT_BOOL:
-      if ( f->index >= 0 )
-      { unsigned int mask = (unsigned int)1 << (f->index-1);
-
-	return PL_unify_bool_ex(val, truePrologFlag(mask) != FALSE);
-      }
+      if ( f->index )
+	return PL_unify_bool_ex(val, truePrologFlag(f->index) != FALSE);
       /*FALLTHROUGH*/
     case FT_ATOM:
       return PL_unify_atom(val, f->value.a);
@@ -1264,6 +1386,37 @@ pl_prolog_flag(term_t name, term_t value, control_t h)
 }
 
 
+static void
+set_arch(void)
+{
+#ifdef __APPLE__
+#include <sys/sysctl.h>
+  uint32_t cputype;
+  size_t size = sizeof(cputype);
+
+  if ( strcmp(PLARCH, "fat-darwin") == 0 )
+  { setPrologFlag("apple_universal_binary", FT_BOOL|FF_READONLY, TRUE, 0);
+
+    if ( sysctlbyname("hw.cputype", &cputype, &size, NULL, 0) == 0 )
+    { switch(cputype&0xff)
+      { case 7:
+	  setPrologFlag("arch", FT_ATOM|FF_READONLY, "x86_64-darwin");
+          return;
+        case 12:
+	  setPrologFlag("arch", FT_ATOM|FF_READONLY, "arm64-darwin");
+          return;
+        default:
+	  Sdprintf("sysctlbyname() cputype = %d (unknown)\n", (int)cputype);
+      }
+    }
+  }
+#endif
+
+  setPrologFlag("arch", FT_ATOM|FF_READONLY, PLARCH);
+}
+
+
+
 		 /*******************************
 		 *	INITIALISE FEATURES	*
 		 *******************************/
@@ -1291,8 +1444,9 @@ initPrologFlagTable(void)
 void
 initPrologFlags(void)
 { GET_LD
+
   setPrologFlag("iso",  FT_BOOL, FALSE, PLFLAG_ISO);
-  setPrologFlag("arch", FT_ATOM|FF_READONLY, PLARCH);
+  set_arch();
 #if __WINDOWS__
   setPrologFlag("windows",	FT_BOOL|FF_READONLY, TRUE, 0);
   const char *wine_version;
@@ -1309,12 +1463,16 @@ initPrologFlags(void)
   setPrologFlag("dialect", FT_ATOM|FF_READONLY, "swi");
   if ( systemDefaults.home )
     setPrologFlag("home", FT_ATOM|FF_READONLY, systemDefaults.home);
+#ifdef PLSHAREDHOME
+  setPrologFlag("shared_home", FT_ATOM|FF_READONLY, PLSHAREDHOME);
+#endif
   if ( GD->paths.executable )
     setPrologFlag("executable", FT_ATOM|FF_READONLY, GD->paths.executable);
 #if defined(HAVE_GETPID) || defined(EMULATE_GETPID)
   setPrologFlag("pid", FT_INTEGER|FF_READONLY, getpid());
 #endif
   setPrologFlag("optimise", FT_BOOL, GD->cmdline.optimise, PLFLAG_OPTIMISE);
+  setPrologFlag("optimise_unify", FT_BOOL, TRUE, PLFLAG_OPTIMISE_UNIFY);
   setPrologFlag("optimise_debug", FT_ATOM, "default", 0);
   setPrologFlag("generate_debug_info", FT_BOOL,
 		truePrologFlag(PLFLAG_DEBUGINFO), PLFLAG_DEBUGINFO);
@@ -1341,7 +1499,10 @@ initPrologFlags(void)
 #ifdef O_ATOMGC
   setPrologFlag("agc_margin",FT_INTEGER,	       GD->atoms.margin);
 #endif
-  setPrologFlag("table_space", FT_INTEGER, LD->tabling.node_pool.limit);
+  setPrologFlag("table_space", FT_INTEGER, GD->options.tableSpace);
+#ifdef O_PLMT
+  setPrologFlag("shared_table_space", FT_INTEGER, GD->options.sharedTableSpace);
+#endif
   setPrologFlag("stack_limit", FT_INTEGER, LD->stacks.limit);
 #if defined(HAVE_DLOPEN) || defined(HAVE_SHL_LOAD) || defined(EMULATE_DLOPEN)
   setPrologFlag("open_shared_object",	  FT_BOOL|FF_READONLY, TRUE, 0);
@@ -1354,6 +1515,8 @@ initPrologFlags(void)
 #endif
 #ifdef O_PLMT
   setPrologFlag("threads",	FT_BOOL, !GD->options.nothreads, 0);
+  if ( GD->options.xpce >= 0 )
+    setPrologFlag("xpce",	FT_BOOL, GD->options.xpce, 0);
   setPrologFlag("system_thread_id", FT_INTEGER|FF_READONLY, 0, 0);
   setPrologFlag("gc_thread",    FT_BOOL,
 		!GD->options.nothreads &&
@@ -1375,11 +1538,14 @@ initPrologFlags(void)
   setPrologFlag("debug_on_error",	FT_BOOL, TRUE, PLFLAG_DEBUG_ON_ERROR);
   setPrologFlag("report_error",	FT_BOOL, TRUE, PLFLAG_REPORT_ERROR);
 #endif
+  setPrologFlag("on_error", FT_ATOM, GD->options.on_error);
+  setPrologFlag("on_warning", FT_ATOM, GD->options.on_warning);
   setPrologFlag("break_level", FT_INTEGER|FF_READONLY, 0, 0);
   setPrologFlag("user_flags", FT_ATOM, "silent");
   setPrologFlag("editor", FT_ATOM, "default");
   setPrologFlag("debugger_show_context", FT_BOOL, FALSE, 0);
-  setPrologFlag("autoload",  FT_BOOL, TRUE,  PLFLAG_AUTOLOAD);
+  setPrologFlag("autoload",  FT_ATOM, "true");
+  setPrologFlagMask(PLFLAG_AUTOLOAD);
 #ifndef O_GMP
   setPrologFlag("max_integer",	   FT_INT64|FF_READONLY, PLMAXINT);
   setPrologFlag("min_integer",	   FT_INT64|FF_READONLY, PLMININT);
@@ -1387,7 +1553,11 @@ initPrologFlags(void)
   setPrologFlag("max_tagged_integer", FT_INTEGER|FF_READONLY, PLMAXTAGGEDINT);
   setPrologFlag("min_tagged_integer", FT_INTEGER|FF_READONLY, PLMINTAGGEDINT);
 #ifdef O_GMP
-  setPrologFlag("bounded",		   FT_BOOL|FF_READONLY,	   FALSE, 0);
+  setPrologFlag("bounded",	      FT_BOOL|FF_READONLY,	   FALSE, 0);
+  setPrologFlag("prefer_rationals", FT_BOOL, O_PREFER_RATIONALS, PLFLAG_RATIONAL);
+  setPrologFlag("rational_syntax",  FT_ATOM,
+		O_RATIONAL_SYNTAX == RAT_NATURAL ? "natural" :
+		                                   "compatibility");
 #ifdef __GNU_MP__
   setPrologFlag("gmp_version",	   FT_INTEGER|FF_READONLY, __GNU_MP__);
 #endif
@@ -1398,10 +1568,14 @@ initPrologFlags(void)
     setPrologFlag("integer_rounding_function", FT_ATOM|FF_READONLY, "down");
   else
     setPrologFlag("integer_rounding_function", FT_ATOM|FF_READONLY, "toward_zero");
+  setPrologFlag("max_char_code", FT_INTEGER|FF_READONLY, PLMAXWCHAR);
   setPrologFlag("max_arity", FT_ATOM|FF_READONLY, "unbounded");
+  setPrologFlag("max_procedure_arity", FT_INTEGER|FF_READONLY, MAXARITY);
   setPrologFlag("answer_format", FT_ATOM, "~p");
   setPrologFlag("colon_sets_calling_context", FT_BOOL|FF_READONLY, TRUE, 0);
   setPrologFlag("character_escapes", FT_BOOL, TRUE, PLFLAG_CHARESCAPE);
+  setPrologFlag("character_escapes_unicode", FT_BOOL, TRUE,
+		PLFLAG_CHARESCAPE_UNICODE);
   setPrologFlag("var_prefix", FT_BOOL, FALSE, PLFLAG_VARPREFIX);
   setPrologFlag("char_conversion", FT_BOOL, FALSE, PLFLAG_CHARCONVERSION);
 #ifdef O_QUASIQUOTATIONS
@@ -1410,14 +1584,19 @@ initPrologFlags(void)
   setPrologFlag("write_attributes", FT_ATOM, "ignore");
   setPrologFlag("stream_type_check", FT_ATOM, "loose");
   setPrologFlag("occurs_check", FT_ATOM, "false");
+  setPrologFlag("shift_check", FT_BOOL, FALSE,  PLFLAG_SHIFT_CHECK);
   setPrologFlag("access_level", FT_ATOM, "user");
   setPrologFlag("double_quotes", FT_ATOM,
 		GD->options.traditional ? "codes" : "string");
   setPrologFlag("back_quotes", FT_ATOM,
 		GD->options.traditional ? "symbol_char" : "codes");
+  setPrologFlag("portable_vmi", FT_BOOL, TRUE, PLFLAG_PORTABLE_VMI);
   setPrologFlag("traditional", FT_BOOL|FF_READONLY, GD->options.traditional, 0);
   setPrologFlag("unknown", FT_ATOM, "error");
   setPrologFlag("debug", FT_BOOL, FALSE, 0);
+  setPrologFlag("debug_on_interrupt", FT_BOOL,
+		truePrologFlag(PLFLAG_DEBUG_ON_INTERRUPT),
+		PLFLAG_DEBUG_ON_INTERRUPT);
   setPrologFlag("verbose", FT_ATOM|FF_KEEP, GD->options.silent ? "silent" : "normal");
   setPrologFlag("verbose_load", FT_ATOM, "silent");
   setPrologFlag("verbose_autoload", FT_BOOL, FALSE, 0);
@@ -1429,9 +1608,14 @@ initPrologFlags(void)
 		PLFLAG_DOT_IN_ATOM);
   setPrologFlag("toplevel_var_size", FT_INTEGER, 1000);
   setPrologFlag("toplevel_print_anon", FT_BOOL, TRUE, 0);
+  setPrologFlag("toplevel_name_variables", FT_BOOL, TRUE, 0);
   setPrologFlag("toplevel_prompt", FT_ATOM, "~m~d~l~! ?- ");
   setPrologFlag("file_name_variables", FT_BOOL, FALSE, PLFLAG_FILEVARS);
   setPrologFlag("fileerrors", FT_BOOL, TRUE, PLFLAG_FILEERRORS);
+  setPrologFlag("determinism_error", FT_ATOM, "error");
+#ifdef O_DEBUG
+  setPrologFlag("prolog_debug", FT_BOOL|FF_READONLY, TRUE, 0);
+#endif
 #ifdef __EMSCRIPTEN__
   setPrologFlag("emscripten", FT_BOOL|FF_READONLY, TRUE, 0);
 #else
@@ -1441,6 +1625,12 @@ initPrologFlags(void)
 #ifdef __APPLE__
   setPrologFlag("apple", FT_BOOL|FF_READONLY, TRUE, 0);
 #endif
+#ifdef __ANDROID__
+  setPrologFlag("android", FT_BOOL|FF_READONLY, TRUE, 0);
+# ifdef __ANDROID_API__
+  setPrologFlag("android_api",FT_INTEGER|FF_READONLY, __ANDROID_API__);
+# endif
+#endif
 #endif
 
   setPrologFlag("encoding", FT_ATOM, stringAtom(encoding_to_atom(LD->encoding)));
@@ -1449,6 +1639,7 @@ initPrologFlags(void)
 		truePrologFlag(PLFLAG_TTY_CONTROL), PLFLAG_TTY_CONTROL);
   setPrologFlag("signals", FT_BOOL|FF_READONLY,
 		truePrologFlag(PLFLAG_SIGNALS), PLFLAG_SIGNALS);
+  setPrologFlag("packs", FT_BOOL, GD->cmdline.packs, 0);
 
 #if defined(__WINDOWS__) && defined(_DEBUG)
   setPrologFlag("kernel_compile_mode", FT_ATOM|FF_READONLY, "debug");
@@ -1464,7 +1655,13 @@ initPrologFlags(void)
 #ifdef O_MITIGATE_SPECTRE
   setPrologFlag("mitigate_spectre", FT_BOOL, FALSE, PLFLAG_MITIGATE_SPECTRE);
 #endif
+#ifdef POSIX_SHELL
   setPrologFlag("posix_shell", FT_ATOM, POSIX_SHELL);
+#endif
+
+  setPrologFlag("table_incremental", FT_BOOL, FALSE, PLFLAG_TABLE_INCREMENTAL);
+  setPrologFlag("table_subsumptive", FT_BOOL, FALSE, 0);
+  setPrologFlag("table_shared",      FT_BOOL, FALSE, PLFLAG_TABLE_SHARED);
 
   setTmpDirPrologFlag();
   setTZPrologFlag();
@@ -1517,6 +1714,10 @@ static void
 setTZPrologFlag(void)
 { tzset();
 
+#if defined(__WINDOWS__) || defined(__CYGWIN__)
+#define timezone _timezone
+#endif
+
   setPrologFlag("timezone", FT_INTEGER|FF_READONLY, timezone);
 }
 
@@ -1562,6 +1763,40 @@ setVersionPrologFlag(void)
   setGITVersion();
 }
 
+static int
+abi_version_dict(term_t dict)
+{ GET_LD
+  const atom_t keys[] = { ATOM_foreign_interface,
+			  ATOM_record,
+			  ATOM_qlf,
+			  ATOM_qlf_min_load,
+			  ATOM_vmi,
+			  ATOM_built_in };
+  term_t values = PL_new_term_refs(6);
+
+  return ( PL_unify_integer(values+0, PL_version(PL_VERSION_FLI)) &&
+	   PL_unify_integer(values+1, PL_version(PL_VERSION_REC)) &&
+	   PL_unify_integer(values+2, PL_version(PL_VERSION_QLF)) &&
+	   PL_unify_integer(values+3, PL_version(PL_VERSION_QLF_LOAD)) &&
+	   PL_unify_integer(values+4, PL_version(PL_VERSION_VM)) &&
+	   PL_unify_integer(values+5, PL_version(PL_VERSION_BUILT_IN)) &&
+
+	   PL_put_dict(dict, ATOM_abi, 6, keys, values) );
+}
+
+
+void
+setABIVersionPrologFlag(void)
+{ GET_LD
+  fid_t fid = PL_open_foreign_frame();
+  term_t t = PL_new_term_ref();
+
+  if ( abi_version_dict(t) )
+    setPrologFlag("abi_version", FF_READONLY|FT_TERM, t);
+
+  PL_discard_foreign_frame(fid);
+}
+
 
 void
 cleanupPrologFlags(void)
