@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2020, University of Amsterdam
+    Copyright (c)  1985-2021, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -34,10 +35,22 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
+#define EMIT_ALLOC_INLINES 1
 #include "pl-incl.h"
 #include "os/pl-cstack.h"
 #include "pl-dict.h"
+#include "pl-arith.h"
+#include "pl-variant.h"
+#include "pl-prims.h"
+#include "pl-gvar.h"
+#include "pl-gc.h"
+#include "pl-fli.h"
+#include "pl-setup.h"
+#include "pl-pro.h"
 #include <math.h>
+#ifdef HAVE_MALLOC_H
+#include <malloc.h>
+#endif
 #ifdef HAVE_SYS_MMAN_H
 #define MMAP_STACK 1
 #include <sys/mman.h>
@@ -59,6 +72,8 @@
 #define ALLOC_NEW_MAGIC  0xF9
 #endif
 
+/* Emit the non-inline definitions here */
+#include "pl-alloc-inline.h"
 
 		 /*******************************
 		 *	    USE BOEHM GC	*
@@ -308,14 +323,15 @@ typedef struct cycle_entry
 { LocalFrame frame;
 } cycle_entry;
 
+#define is_variant_frame(fr1, fr2) LDFUNC(is_variant_frame, fr1, fr2)
 static int
-is_variant_frame(LocalFrame fr1, LocalFrame fr2 ARG_LD)
+is_variant_frame(DECL_LD LocalFrame fr1, LocalFrame fr2)
 { if ( fr1->predicate == fr2->predicate )
   { size_t arity = fr1->predicate->functor->arity;
     size_t i;
 
     for(i=0; i<arity; i++)
-    { if ( !is_variant_ptr(argFrameP(fr1, i), argFrameP(fr2, i) PASS_LD) )
+    { if ( !is_variant_ptr(argFrameP(fr1, i), argFrameP(fr2, i)) )
 	return FALSE;
     }
 
@@ -326,11 +342,11 @@ is_variant_frame(LocalFrame fr1, LocalFrame fr2 ARG_LD)
 }
 
 
+#define non_terminating_recursion(fr0, ce, is_cycle) LDFUNC(non_terminating_recursion, fr0, ce, is_cycle)
 static int
-non_terminating_recursion(LocalFrame fr0,
+non_terminating_recursion(DECL_LD LocalFrame fr0,
 			  cycle_entry ce[MAX_CYCLE],
-			  int *is_cycle
-			  ARG_LD)
+			  int *is_cycle)
 { int depth, mindepth = 1, repeat;
   LocalFrame fr, ctx;
 
@@ -348,7 +364,7 @@ again:
   if ( !fr || depth >= MAX_CYCLE )
     return 0;
 
-  *is_cycle = is_variant_frame(fr0, fr PASS_LD);
+  *is_cycle = is_variant_frame(fr0, fr);
   ctx = fr;
 
   for(repeat=MIN_REPEAT; fr && --repeat > 0; )
@@ -377,15 +393,16 @@ again:
   return 0;
 }
 
+#define find_non_terminating_recursion(fr, ce, is_cycle) LDFUNC(find_non_terminating_recursion, fr, ce, is_cycle)
 static int
-find_non_terminating_recursion(LocalFrame fr, cycle_entry ce[MAX_CYCLE],
-			       int *is_cycle ARG_LD)
+find_non_terminating_recursion(DECL_LD LocalFrame fr, cycle_entry ce[MAX_CYCLE],
+			       int *is_cycle)
 { int max_pre_loop = MAX_PRE_LOOP;
 
   for(; fr && max_pre_loop; fr = parentFrame(fr), max_pre_loop--)
   { int len;
 
-    if ( (len=non_terminating_recursion(fr, ce, is_cycle PASS_LD)) )
+    if ( (len=non_terminating_recursion(fr, ce, is_cycle)) )
       return len;
   }
 
@@ -393,8 +410,9 @@ find_non_terminating_recursion(LocalFrame fr, cycle_entry ce[MAX_CYCLE],
 }
 
 
+#define top_of_stack(fr, ce, maxdepth) LDFUNC(top_of_stack, fr, ce, maxdepth)
 static int
-top_of_stack(LocalFrame fr, cycle_entry ce[MAX_CYCLE], int maxdepth ARG_LD)
+top_of_stack(DECL_LD LocalFrame fr, cycle_entry ce[MAX_CYCLE], int maxdepth)
 { int depth;
 
   for(depth = 0; fr && depth < maxdepth; fr = parentFrame(fr), depth++)
@@ -468,7 +486,7 @@ push_goal(LocalFrame fr)
 
 	if ( isList(*a) )
 	{ Word tail;
-	  intptr_t len = skip_list(a, &tail PASS_LD);
+	  intptr_t len = skip_list(a, &tail);
 
 	  *ad++ = FUNCTOR_dot2;
 	  deRef(tail);
@@ -533,8 +551,9 @@ push_cycle(cycle_entry ce[MAX_CYCLE], int depth)
 }
 
 
+#define push_stack(ce, depth, name, pp) LDFUNC(push_stack, ce, depth, name, pp)
 static void
-push_stack(cycle_entry ce[MAX_CYCLE], int depth, atom_t name, Word *pp ARG_LD)
+push_stack(DECL_LD cycle_entry ce[MAX_CYCLE], int depth, atom_t name, Word *pp)
 { word w;
   Word p = *pp;
 
@@ -578,7 +597,7 @@ push_overflow_context(Stack stack, int extra)
     }
     *p++ = consInt(env_frames(environment_frame));
     *p++ = ATOM_environments;
-    *p++ = consInt(choice_points(BFR));
+    *p++ = consInt(choice_points(LD->choicepoints));
     *p++ = ATOM_choicepoints;
     gTop = p;
 
@@ -586,19 +605,19 @@ push_overflow_context(Stack stack, int extra)
     { int is_cycle;
 
       if ( (depth=find_non_terminating_recursion(environment_frame, ce,
-						 &is_cycle PASS_LD)) )
+						 &is_cycle)) )
       { push_stack(ce, depth, is_cycle ? ATOM_cycle : ATOM_non_terminating,
-		   &p PASS_LD);
-      } else if ( (depth=top_of_stack(environment_frame, ce, 5 PASS_LD)) )
-      { push_stack(ce, depth, ATOM_stack, &p PASS_LD);
+		   &p);
+      } else if ( (depth=top_of_stack(environment_frame, ce, 5)) )
+      { push_stack(ce, depth, ATOM_stack, &p);
       }
-    } else if ( (depth=top_of_stack(environment_frame, ce, 5 PASS_LD)) )
-    { push_stack(ce, depth, ATOM_stack, &p PASS_LD);
+    } else if ( (depth=top_of_stack(environment_frame, ce, 5)) )
+    { push_stack(ce, depth, ATOM_stack, &p);
     }
 
     *dict = dict_functor((p-dict-2)/2);		/* final functor */
 
-    dict_order(dict, FALSE PASS_LD);
+    dict_order(dict, NULL);
 
     return consPtr(dict, STG_GLOBAL|TAG_COMPOUND);
   } else
@@ -680,7 +699,7 @@ outOfStack(void *stack, stack_overflow_action how)
 	gTop += 5;
 
 	*valTermRef(LD->exception.bin) = consPtr(p, TAG_COMPOUND|STG_GLOBAL);
-	freezeGlobal(PASS_LD1);
+	freezeGlobal();
       } else
       { Sdprintf("ERROR: Out of global-stack.\n"
 		 "ERROR: No room for exception term.  Aborting.\n");
@@ -728,7 +747,7 @@ raiseStackOverflow(int overflow)
 
 
 void
-pushArgumentStack__LD(Word p ARG_LD)
+f_pushArgumentStack(DECL_LD Word p)
 { Word *newbase;
   size_t newsize = nextStackSize((Stack)&LD->stacks.argument, 1);
 
@@ -769,7 +788,7 @@ inline  as  it is simple and usualy very time critical.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Word
-allocGlobal__LD(size_t n ARG_LD)
+allocGlobal(DECL_LD size_t n)
 { Word result;
 
   if ( !hasGlobalSpace(n) )
@@ -788,7 +807,7 @@ allocGlobal__LD(size_t n ARG_LD)
 }
 
 Word
-allocGlobalNoShift__LD(size_t n ARG_LD)
+allocGlobalNoShift(DECL_LD size_t n)
 { Word result;
 
   if ( gTop+n > gMax )
@@ -827,7 +846,7 @@ Return is one of:
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-put_int64(Word at, int64_t l, int flags ARG_LD)
+put_int64(DECL_LD Word at, int64_t l, int flags)
 { Word p;
   word r, m;
   int req;
@@ -899,7 +918,7 @@ Note that these functions can trigger GC
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Word
-allocString(size_t len ARG_LD)
+allocString(DECL_LD size_t len)
 { size_t lw = (len+sizeof(word))/sizeof(word);
   int pad = (int)(lw*sizeof(word) - len);
   Word p = allocGlobal(2 + lw);
@@ -919,7 +938,7 @@ allocString(size_t len ARG_LD)
 word
 globalString(size_t len, const char *s)
 { GET_LD
-  Word p = allocString(len+1 PASS_LD);
+  Word p = allocString(len+1);
 
   if ( p )
   { char *q = (char *)&p[1];
@@ -949,7 +968,7 @@ globalWString(size_t len, const pl_wchar_t *s)
   if ( p == e )				/* 8-bit string */
   { unsigned char *t;
 
-    if ( !(g = allocString(len+1 PASS_LD)) )
+    if ( !(g = allocString(len+1)) )
       return 0;
     t = (unsigned char *)&g[1];
     *t++ = 'B';
@@ -959,7 +978,7 @@ globalWString(size_t len, const pl_wchar_t *s)
   { char *t;
     pl_wchar_t *w;
 
-    if ( !(g = allocString((len+1)*sizeof(pl_wchar_t) PASS_LD)) )
+    if ( !(g = allocString((len+1)*sizeof(pl_wchar_t))) )
       return 0;
     t = (char *)&g[1];
     w = (pl_wchar_t*)t;
@@ -973,7 +992,7 @@ globalWString(size_t len, const pl_wchar_t *s)
 
 
 char *
-getCharsString__LD(word w, size_t *len ARG_LD)
+getCharsString(DECL_LD word w, size_t *len)
 { Word p = valPtr(w);
   word m = *p;
   size_t wn  = wsizeofInd(m);
@@ -994,7 +1013,7 @@ getCharsString__LD(word w, size_t *len ARG_LD)
 
 
 pl_wchar_t *
-getCharsWString__LD(word w, size_t *len ARG_LD)
+getCharsWString(DECL_LD word w, size_t *len)
 { Word p = valPtr(w);
   word m = *p;
   size_t wn  = wsizeofInd(m);
@@ -1026,7 +1045,7 @@ might not be properly aligned.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-put_double(Word at, double d, int flags ARG_LD)
+put_double(DECL_LD Word at, double d, int flags)
 { Word p;
   word m = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
 
@@ -1053,31 +1072,7 @@ put_double(Word at, double d, int flags ARG_LD)
 }
 
 
-		 /*******************************
-		 *	  64-BIT INTEGERS	*
-		 *******************************/
-
-#if ALIGNOF_INT64_T != ALIGNOF_VOIDP
-
-int64_t					/* take care of alignment! */
-valBignum__LD(word w ARG_LD)
-{ Word p = valIndirectP(w);
-  union
-  { int64_t i;
-    word w[WORDS_PER_INT64];
-  } val;
-
-#if ( SIZEOF_VOIDP == 4 )
-  val.w[0] = p[0];
-  val.w[1] = p[1];
-#else
-#error "Unsupported int64_t alignment conversion"
-#endif
-
-  return val.i;
-}
-
-#endif
+/* valBignum(DECL_LD word w) moved to pl-inline.h */
 
 		 /*******************************
 		 *  GENERIC INDIRECT OPERATIONS	*
@@ -1103,58 +1098,14 @@ equalIndirect(word w1, word w2)
   fail;
 }
 
-
-size_t					/* size in cells */
-gsizeIndirectFromCode(Code pc)
-{ return wsizeofInd(pc[0]) + 2;
-}
-
-
 word
 globalIndirectFromCode(Code *PC)
 { GET_LD
-  Code pc = *PC;
-  word m = *pc++;
-  size_t n = wsizeofInd(m);
-  Word p = allocGlobal(n+2);
-
-  if ( p )
-  { word r = consPtr(p, tag(m)|STG_GLOBAL);
-
-    *p++ = m;
-    while(n-- > 0)
-      *p++ = *pc++;
-    *p++ = m;
-
-    *PC = pc;
-    return r;
-  } else
-    return 0;
+  struct word_and_Code retval = VM_globalIndirectFromCode(*PC);
+  *PC = retval.code;
+  return retval.word;
 }
 
-
-static int				/* used in pl-wam.c */
-equalIndirectFromCode(word a, Code *PC)
-{ GET_LD
-  Word pc = *PC;
-  Word pa = addressIndirect(a);
-
-  if ( *pc == *pa )
-  { size_t n = wsizeofInd(*pc);
-
-    while(n-- > 0)
-    { if ( *++pc != *++pa )
-	fail;
-    }
-    pc++;
-    *PC = pc;
-    succeed;
-  }
-
-  fail;
-}
-
-
 		 /*******************************
 		 *	     GNU MALLOC		*
 		 *******************************/
@@ -1438,7 +1389,7 @@ tmp_nrealloc(void *mem, size_t req)
 }
 
 
-size_t
+static size_t
 tmp_malloc_size(void *mem)
 { if ( mem )
   { map_region *reg = (map_region *)((char*)mem-SA_OFFSET);
@@ -1692,6 +1643,7 @@ static int (*fMallocExtension_SetNumericProperty)(const char *, size_t);
 static void (*fMallocExtension_MarkThreadIdle)(void) = NULL;
 static void (*fMallocExtension_MarkThreadTemporarilyIdle)(void) = NULL;
 static void (*fMallocExtension_MarkThreadBusy)(void) = NULL;
+static void (*fMallocExtension_ReleaseFreeMemory)(void) = NULL;
 
 static const char* tcmalloc_properties[] =
 { "generic.current_allocated_bytes",
@@ -1834,18 +1786,39 @@ heapUsed(void)
 }
 
 
-int
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Try to initialize tcmalloc(). Note that we   get all the functions using
+dlsym() rather than as static symbols  because   we  do not know whether
+tcmalloc is really there. Even if the   symbols  are present the library
+may be overruled.
+
+Returns 0 if tcmalloc is not present or not enabled.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int is_tcmalloc = FALSE;
+
+static int
 initTCMalloc(void)
 { static int done = FALSE;
   int set = 0;
 
   if ( done )
-    return !!fMallocExtension_GetNumericProperty;
+    return is_tcmalloc;
   done = TRUE;
 
   if ( (fMallocExtension_GetNumericProperty =
 		PL_dlsym(NULL, "MallocExtension_GetNumericProperty")) )
-  { PL_register_foreign_in_module("system", "malloc_property", 1, malloc_property,
+  { size_t in_use;
+
+    if ( fMallocExtension_GetNumericProperty("generic.current_allocated_bytes", &in_use) &&
+	 in_use > 100000 )
+    { is_tcmalloc = TRUE;
+      PL_set_prolog_flag("malloc", PL_ATOM, "tcmalloc");
+    } else
+    { return 0;
+    }
+
+    PL_register_foreign_in_module("system", "malloc_property", 1, malloc_property,
 			PL_FA_NONDETERMINISTIC);
     set++;
   }
@@ -1861,10 +1834,78 @@ initTCMalloc(void)
     PL_dlsym(NULL, "MallocExtension_MarkThreadTemporarilyIdle");
   fMallocExtension_MarkThreadBusy =
     PL_dlsym(NULL, "MallocExtension_MarkThreadBusy");
+  fMallocExtension_ReleaseFreeMemory =
+    PL_dlsym(NULL, "MallocExtension_ReleaseFreeMemory");
 
   return set;
 }
 
+#ifdef HAVE_MALLINFO2
+#define mallinfo mallinfo2
+#define mallinfo_fname "mallinfo2"
+#ifndef HAVE_MALLINFO
+#define HAVE_MALLINFO
+#endif
+#else
+#define mallinfo_fname "mallinfo"
+#endif
+
+#ifdef HAVE_MALLINFO
+static int is_ptmalloc = FALSE;
+static struct mallinfo (*fmallinfo)(void) = NULL;
+static int             (*fmalloc_trim)(int pad) = NULL;
+
+static int
+initPTMalloc(void)
+{ static int done = FALSE;
+
+  if ( done )
+    return is_ptmalloc;
+  done = TRUE;
+
+  if ( (fmallinfo    = PL_dlsym(NULL, mallinfo_fname)) &&
+       (fmalloc_trim = PL_dlsym(NULL, "malloc_trim")) )
+  { struct mallinfo info = fmallinfo();
+
+    if ( info.uordblks > 100000 )
+    { PL_set_prolog_flag("malloc", PL_ATOM, "ptmalloc");
+      is_ptmalloc = TRUE;
+    }
+  }
+
+  return is_ptmalloc;
+}
+#else
+#define is_ptmalloc FALSE
+#define initPTMalloc() FALSE
+#define fmalloc_trim(pad) (void)0
+#endif
+
+
+int
+initMalloc(void)
+{ return ( initTCMalloc() ||
+	   initPTMalloc() ||
+	   FALSE
+	 );
+}
+
+
+/** trim_heap
+ *
+ * Release as much as possible memory to the system.
+ */
+
+static
+PRED_IMPL("trim_heap", 0, trim_heap, 0)
+{ if ( is_tcmalloc && fMallocExtension_ReleaseFreeMemory )
+    fMallocExtension_ReleaseFreeMemory();
+  else if ( is_ptmalloc )
+    fmalloc_trim(0);
+
+  return TRUE;
+}
+
 
 /** thread_idle(:Goal, +How)
  *
@@ -1880,7 +1921,7 @@ PRED_IMPL("thread_idle", 2, thread_idle, PL_FA_TRANSPARENT)
     return FALSE;
 
   if ( how == ATOM_short )
-  { trimStacks(TRUE PASS_LD);
+  { trimStacks(TRUE);
     if ( fMallocExtension_MarkThreadTemporarilyIdle &&
 	 fMallocExtension_MarkThreadBusy )
       fMallocExtension_MarkThreadTemporarilyIdle();
@@ -1921,4 +1962,5 @@ BeginPredDefs(alloc)
   PRED_DEF("garbage_collect_heap", 0, garbage_collect_heap, 0)
 #endif
   PRED_DEF("thread_idle", 2, thread_idle, PL_FA_TRANSPARENT)
+  PRED_DEF("trim_heap",   0, trim_heap,   0)
 EndPredDefs
