@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2020, VU University Amsterdam
+    Copyright (c)  2013-2021, VU University Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
@@ -37,6 +37,12 @@
 #include "pl-comp.h"
 #include "pl-dict.h"
 #include "pl-rsort.h"
+#include "pl-funct.h"
+#include "pl-prims.h"
+#include "pl-wam.h"
+#include "pl-fli.h"
+#include "pl-gc.h"
+#include "pl-gvar.h"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Dicts are associative arrays,  where  keys   are  either  atoms  or small
@@ -88,8 +94,9 @@ dict_functor(int pairs)
 		 *      LOW-LEVEL FUNCTIONS	*
 		 *******************************/
 
+#define get_dict_ex(t, dp, ex) LDFUNC(get_dict_ex, t, dp, ex)
 static int
-get_dict_ex(term_t t, Word dp, int ex ARG_LD)
+get_dict_ex(DECL_LD term_t t, Word dp, int ex)
 { Word p = valTermRef(t);
 
   deRef(p);
@@ -113,13 +120,14 @@ get_dict_ex(term_t t, Word dp, int ex ARG_LD)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-get_create_dict_ex(+t, -dict ARG_LD) extracts a dict  from t or raises a
+get_create_dict_ex(DECL_LD +t, -dict) extracts a dict  from t or raises a
 type error. The term reference dict contains   a  plain dict term handle
 and is never a reference.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define get_create_dict_ex(t, dt) LDFUNC(get_create_dict_ex, t, dt)
 static int
-get_create_dict_ex(term_t t, term_t dt ARG_LD)
+get_create_dict_ex(DECL_LD term_t t, term_t dt)
 { Word p = valTermRef(t);
 
   deRef(p);
@@ -148,7 +156,7 @@ dict_lookup_ptr() returns a pointer to the value for a given key
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Word
-dict_lookup_ptr(word dict, word name ARG_LD)
+dict_lookup_ptr(DECL_LD word dict, word name)
 { Functor data = valueTerm(dict);
   int arity = arityFunctor(data->definition);
   int l = 1, h = arity/2;
@@ -187,8 +195,9 @@ dict_lookup_ptr(word dict, word name ARG_LD)
    -2:    duplicate key
 */
 
+#define dict_ordered(data, count, dupl) LDFUNC(dict_ordered, data, count, dupl)
 static int
-dict_ordered(Word data, int count, int ex ARG_LD)
+dict_ordered(DECL_LD Word data, int count, Word dupl)
 { int ordered = TRUE;
   Word n1, n2;
 
@@ -207,9 +216,11 @@ dict_ordered(Word data, int count, int ex ARG_LD)
       if ( *n1 > *n2 )
 	ordered = FALSE;
       if ( *n1 == *n2 )
-      { if ( ex )
+      { if ( dupl )
+	{ *dupl = *n1;
+	} else
 	{ term_t t = PL_new_term_ref();
-	  *valTermRef(t) = linkVal(n1);
+	  *valTermRef(t) = linkValI(n1);
 	  PL_error(NULL, 0, NULL, ERR_DUPLICATE_KEY, t);
 	}
 	return -2;
@@ -221,26 +232,20 @@ dict_ordered(Word data, int count, int ex ARG_LD)
 }
 
 
-#if defined(O_PLMT) || defined(O_MULTIPLE_ENGINES)
-#define GET_LDARG(x) PL_local_data_t *__PL_ld = (x)
-#else
-#define GET_LDARG(x)
-#endif
-
 static int
 compare_dict_entry(const void *a, const void *b, void *arg)
-{ GET_LDARG(arg);
-  Word p = (Word)a+1;
+{ Word p = (Word)a+1;
   Word q = (Word)b+1;
-
-  deRef(p);
-  deRef(q);
+  WITH_LD(arg)
+  { deRef(p);
+    deRef(q);
+  }
   return (*p<*q ? -1 : *p>*q ? 1 : 0);
 }
 
 
 int
-dict_order(Word dict, int ex ARG_LD)
+dict_order(DECL_LD Word dict, Word dupl)
 { Functor data = (Functor)dict;
   int arity = arityFunctor(data->definition);
 
@@ -249,7 +254,7 @@ dict_order(Word dict, int ex ARG_LD)
   sort_r(data->arguments+1, arity/2, sizeof(word)*2,
 	 compare_dict_entry, LD);
 
-  return dict_ordered(data->arguments+1, arity/2, ex PASS_LD) == TRUE;
+  return dict_ordered(data->arguments+1, arity/2, dupl);
 }
 
 
@@ -264,13 +269,17 @@ typedef struct order_term_refs
 } order_term_refs;
 
 
+#define compare_term_refs(ip1, ip2, ctx) LDFUNC(compare_term_refs, ip1, ip2, ctx)
+static inline int compare_term_refs(DECL_LD const int *ip1, const int *ip2, order_term_refs *ctx);
+
 static int
-compare_term_refs(const void *a, const void *b, void *arg)
-{ const int *ip1 = a;
-  const int *ip2 = b;
-  order_term_refs *ctx = arg;
-  GET_LDARG(ctx->ld);
-  Word p = valTermRef(ctx->av[*ip1*2]);
+(compare_term_refs)(const void *a, const void *b, void *arg)
+{ return compare_term_refs(PASS_AS_LD(((order_term_refs*)arg)->ld) a, b, arg);
+}
+
+static inline int
+compare_term_refs(DECL_LD const int *ip1, const int *ip2, order_term_refs *ctx)
+{ Word p = valTermRef(ctx->av[*ip1*2]);
   Word q = valTermRef(ctx->av[*ip2*2]);
 
   assert(!isRef(*p));
@@ -281,7 +290,7 @@ compare_term_refs(const void *a, const void *b, void *arg)
 
 
 int
-dict_order_term_refs(term_t *av, int *indexes, int count ARG_LD)
+dict_order_term_refs(DECL_LD term_t *av, int *indexes, int count)
 { order_term_refs ctx;
 
   ctx.ld = LD;
@@ -305,22 +314,23 @@ dict_order_term_refs(term_t *av, int *indexes, int count ARG_LD)
 }
 
 
+#define assign_in_dict(dp, val) LDFUNC(assign_in_dict, dp, val)
 static int
-assign_in_dict(Word dp, Word val ARG_LD)
+assign_in_dict(DECL_LD Word dp, Word val)
 { deRef(val);
 
   if ( !canBind(*val) )
   { *dp = *val;
   } else if ( isAttVar(*val) )
-  { *dp = makeRef(val);
+  { *dp = makeRefG(val);
   } else
   { if ( dp < val )
     { if ( unlikely(tTop+1 >= tMax) )
 	return TRAIL_OVERFLOW;
       setVar(*dp);
-      Trail(val, makeRef(dp));
+      Trail(val, makeRefG(dp));
     } else
-    { *dp = makeRef(val);
+    { *dp = makeRefG(val);
     }
   }
 
@@ -328,8 +338,9 @@ assign_in_dict(Word dp, Word val ARG_LD)
 }
 
 
-int
-put_dict(word dict, int size, Word nv, word *new_dict ARG_LD)
+#define put_dict(dict, size, nv, new_dict) LDFUNC(put_dict, dict, size, nv, new_dict)
+static int
+put_dict(DECL_LD word dict, int size, Word nv, word *new_dict)
 { Functor data = valueTerm(dict);
   int arity = arityFunctor(data->definition);
   Word new, out, in, in_end, nv_end;
@@ -359,19 +370,19 @@ put_dict(word dict, int size, Word nv, word *new_dict ARG_LD)
     deRef2(nv+1, n_name);
 
     if ( *i_name == *n_name )
-    { if ( (rc=assign_in_dict(out++, nv PASS_LD)) != TRUE )
+    { if ( (rc=assign_in_dict(out++, nv)) != TRUE )
 	return rc;
       *out++ = *i_name;
-      if ( !modified && compareStandard(nv, in, TRUE PASS_LD) )
+      if ( !modified && compareStandard(nv, in, TRUE) )
 	modified = TRUE;
       in += 2;
       nv += 2;
     } else if ( *i_name < *n_name )
-    { *out++ = linkVal(in);
+    { *out++ = linkValI(in);
       *out++ = *i_name;
       in += 2;
     } else
-    { if ( (rc=assign_in_dict(out++, nv PASS_LD)) != TRUE )
+    { if ( (rc=assign_in_dict(out++, nv)) != TRUE )
 	return rc;
       *out++ = *n_name;
       nv += 2;
@@ -387,7 +398,7 @@ put_dict(word dict, int size, Word nv, word *new_dict ARG_LD)
     while(in < in_end)
     { Word i_name;
       deRef2(in+1, i_name);
-      *out++ = linkVal(in);
+      *out++ = linkValI(in);
       *out++ = *i_name;
       in += 2;
     }
@@ -397,7 +408,7 @@ put_dict(word dict, int size, Word nv, word *new_dict ARG_LD)
       int rc;
 
       deRef2(nv+1, n_name);
-      if ( (rc=assign_in_dict(out++, nv PASS_LD)) != TRUE )
+      if ( (rc=assign_in_dict(out++, nv)) != TRUE )
 	return rc;
       *out++ = *n_name;
       nv += 2;
@@ -405,7 +416,7 @@ put_dict(word dict, int size, Word nv, word *new_dict ARG_LD)
   }
 
   gTop = out;
-  new[1] = linkVal(&data->arguments[0]);
+  new[1] = linkValI(&data->arguments[0]);
   new[0] = dict_functor((out-(new+1))/2);
 
   *new_dict = consPtr(new, TAG_COMPOUND|STG_GLOBAL);
@@ -414,8 +425,9 @@ put_dict(word dict, int size, Word nv, word *new_dict ARG_LD)
 }
 
 
+#define del_dict(dict, key, new_dict) LDFUNC(del_dict, dict, key, new_dict)
 static int
-del_dict(word dict, word key, word *new_dict ARG_LD)
+del_dict(DECL_LD word dict, word key, word *new_dict)
 { Functor data = valueTerm(dict);
   int arity = arityFunctor(data->definition);
   Word new, out, in, in_end;
@@ -435,14 +447,14 @@ del_dict(word dict, word key, word *new_dict ARG_LD)
 
     deRef2(in+1, i_name);
     if ( *i_name != key )
-    { *out++ = linkVal(in);
+    { *out++ = linkValI(in);
       *out++ = *i_name;
     }
     in += 2;
   }
 
   gTop = out;
-  new[1] = linkVal(&data->arguments[0]); /* tag */
+  new[1] = linkValI(&data->arguments[0]); /* tag */
   new[0] = dict_functor((out-(new+1))/2); /* arity */
 
   *new_dict = consPtr(new, TAG_COMPOUND|STG_GLOBAL);
@@ -456,8 +468,9 @@ del_dict(word dict, word key, word *new_dict ARG_LD)
    and *_OVERFLOW on some memory overflow
 */
 
+#define partial_unify_dict(dict1, dict2) LDFUNC(partial_unify_dict, dict1, dict2)
 static int
-partial_unify_dict(word dict1, word dict2 ARG_LD)
+partial_unify_dict(DECL_LD word dict1, word dict2)
 { Functor d1 = valueTerm(dict1);
   Functor d2 = valueTerm(dict2);
   Word in1  = d1->arguments;
@@ -467,7 +480,7 @@ partial_unify_dict(word dict1, word dict2 ARG_LD)
   int rc;
 
   /* unify the tages */
-  if ( (rc=unify_ptrs(in1, in2, ALLOW_RETCODE PASS_LD)) != TRUE )
+  if ( (rc=unify_ptrs(in1, in2, ALLOW_RETCODE)) != TRUE )
     return rc;
 
   /* advance to first v+k entry */
@@ -480,7 +493,7 @@ partial_unify_dict(word dict1, word dict2 ARG_LD)
     deRef2(in1+1, n1);
     deRef2(in2+1, n2);
     if ( *n1 == *n2 )
-    { if ( (rc = unify_ptrs(in1, in2, ALLOW_RETCODE PASS_LD)) != TRUE )
+    { if ( (rc = unify_ptrs(in1, in2, ALLOW_RETCODE)) != TRUE )
 	return rc;
       in1 += 2;
       in2 += 2;
@@ -503,8 +516,9 @@ partial_unify_dict(word dict1, word dict2 ARG_LD)
    two pass process.
 */
 
+#define select_dict(del, from, new_dict) LDFUNC(select_dict, del, from, new_dict)
 static int
-select_dict(word del, word from, word *new_dict ARG_LD)
+select_dict(DECL_LD word del, word from, word *new_dict)
 { Functor dd = valueTerm(del);
   Functor fd = valueTerm(from);
   Word din  = dd->arguments;
@@ -515,7 +529,7 @@ select_dict(word del, word from, word *new_dict ARG_LD)
   int rc;
 
   /* unify the tags */
-  if ( (rc=unify_ptrs(din, fin, ALLOW_RETCODE PASS_LD)) != TRUE )
+  if ( (rc=unify_ptrs(din, fin, ALLOW_RETCODE)) != TRUE )
     return rc;
 
   /* advance to first v+k entry */
@@ -529,7 +543,7 @@ select_dict(word del, word from, word *new_dict ARG_LD)
     deRef2(fin+1, f);
 
     if ( *d == *f )
-    { if ( (rc = unify_ptrs(din, fin, ALLOW_RETCODE PASS_LD)) != TRUE )
+    { if ( (rc = unify_ptrs(din, fin, ALLOW_RETCODE)) != TRUE )
 	return rc;
       din += 2;
       fin += 2;
@@ -567,7 +581,7 @@ select_dict(word del, word from, word *new_dict ARG_LD)
       { din += 2;
 	fin += 2;
       } else
-      { *out++ = linkVal(fin);
+      { *out++ = linkValI(fin);
 	*out++ = *f;
 	fin += 2;
 	left--;
@@ -582,8 +596,9 @@ select_dict(word del, word from, word *new_dict ARG_LD)
 }
 
 
+#define get_name_ex(t, np) LDFUNC(get_name_ex, t, np)
 static int
-get_name_ex(term_t t, Word np ARG_LD)
+get_name_ex(DECL_LD term_t t, Word np)
 { Word p = valTermRef(t);
 
   deRef(p);
@@ -597,8 +612,9 @@ get_name_ex(term_t t, Word np ARG_LD)
 }
 
 
+#define get_name_value(p, name, value, m, flags) LDFUNC(get_name_value, p, name, value, m, flags)
 static int
-get_name_value(Word p, Word name, Word value, mark *m, int flags ARG_LD)
+get_name_value(DECL_LD Word p, Word name, Word value, mark *m, int flags)
 { const char *type;
 
   deRef(p);
@@ -615,7 +631,7 @@ get_name_value(Word p, Word name, Word value, mark *m, int flags ARG_LD)
       if ( is_dict_key(*np) )
       { *name = *np;
 	deRef2(&f->arguments[1], vp);
-	*value = linkVal(vp);
+	*value = linkValI(vp);
 
 	return TRUE;
       } else
@@ -631,7 +647,7 @@ get_name_value(Word p, Word name, Word value, mark *m, int flags ARG_LD)
 
       *name = nameFunctor(f->definition);
       deRef2(&f->arguments[0], vp);
-      *value = linkVal(vp);
+      *value = linkValI(vp);
       return TRUE;
     }
   }
@@ -663,10 +679,11 @@ PL_is_dict(term_t t)
   if ( isTerm(*p) )
   { Functor f = valueTerm(*p);
     FunctorDef fd = valueFunctor(f->definition);
+    word dupl;
 
     if ( fd->name == ATOM_dict &&
 	 fd->arity%2 == 1 &&
-	 dict_ordered(f->arguments+1, fd->arity/2, FALSE PASS_LD) == TRUE )
+	 dict_ordered(f->arguments+1, fd->arity/2, &dupl) == TRUE )
       return TRUE;
   }
 
@@ -677,6 +694,7 @@ PL_is_dict(term_t t)
 static int
 PL_get_dict_ex(term_t data, term_t tag, term_t dict, int flags)
 { GET_LD
+  word dupl;
 
   if ( PL_is_dict(data) )
   { PL_put_term(dict, data);
@@ -694,7 +712,7 @@ PL_get_dict_ex(term_t data, term_t tag, term_t dict, int flags)
 
     if ( unlikely(tTop+1 >= tMax) )
     { if ( !makeMoreStackSpace(TRAIL_OVERFLOW, ALLOW_GC|ALLOW_SHIFT) )
-       return FALSE;
+	return FALSE;
     }
     if ( (rc=ensureGlobalSpace(len*2+2, ALLOW_GC)) != TRUE )
       return raiseStackOverflow(rc);
@@ -705,11 +723,17 @@ PL_get_dict_ex(term_t data, term_t tag, term_t dict, int flags)
     if ( tag )
     { Word cp = valTermRef(tag);
 
-      *ap = linkVal(cp);		/* TBD: maybe move to another function */
-      if ( tagex(*ap) == (TAG_REFERENCE|STG_LOCAL) )
-      { deRef(cp)
-	setVar(*ap);
-	Trail(cp, makeRef(ap));
+      deRef(cp);
+      if ( needsRef(*cp) )
+      { if ( isVar(*cp) )
+	{ setVar(*ap);
+	  Trail(cp, makeRefG(ap));
+	} else
+	{ assert(onStack(global, cp));
+	  *ap = makeRefG(cp);
+	}
+      } else
+      { *ap = *cp;
       }
     } else
     { setVar(*ap);
@@ -721,20 +745,26 @@ PL_get_dict_ex(term_t data, term_t tag, term_t dict, int flags)
     while( isList(*tail) )
     { Word head = HeadList(tail);
 
-      if ( !get_name_value(head, ap+1, ap, &m, flags PASS_LD) )
+      if ( !get_name_value(head, ap+1, ap, &m, flags) )
 	return FALSE;
       ap += 2;
       tail = TailList(tail);
       deRef(tail);
     }
 
-    if ( dict_order(dp, TRUE PASS_LD) )
+    if ( (rc=dict_order(dp, &dupl)) == TRUE )
     { gTop = ap;
       *valTermRef(dict) = consPtr(dp, TAG_COMPOUND|STG_GLOBAL);
       DEBUG(CHK_SECURE, checkStacks(NULL));
       return TRUE;
     } else
-    { return FALSE;
+    { term_t ex;
+
+      assert(rc == -2);
+      Undo(m);
+      return ( (ex = PL_new_term_ref()) &&
+	       _PL_unify_atomic(ex, dupl) &&
+	       PL_error(NULL, 0, NULL, ERR_DUPLICATE_KEY, ex) );
     }
   }					/* TBD: {name:value, ...} */
 
@@ -755,13 +785,17 @@ typedef struct cmp_dict_index_data
   PL_local_data_t *ld;
 } cmp_dict_index_data;
 
+#define cmp_dict_index(ip1, ip2, ctx) LDFUNC(cmp_dict_index, ip1, ip2, ctx)
+static inline int cmp_dict_index(DECL_LD const int *ip1, const int *ip2, cmp_dict_index_data *ctx);
+
 static int
-cmp_dict_index(const void *a1, const void *a2, void *arg)
-{ int *ip1 = (int*)a1;
-  int *ip2 = (int*)a2;
-  cmp_dict_index_data *ctx = arg;
-  GET_LDARG(ctx->ld);
-  Word p = &ctx->data[*ip1*2+1];
+(cmp_dict_index)(const void *a1, const void *a2, void *arg)
+{ return cmp_dict_index(PASS_AS_LD(((cmp_dict_index_data*)arg)->ld) a1, a2, arg);
+}
+
+static inline int
+cmp_dict_index(DECL_LD const int *ip1, const int *ip2, cmp_dict_index_data *ctx)
+{ Word p = &ctx->data[*ip1*2+1];
   Word q = &ctx->data[*ip2*2+1];
   int rc;
 
@@ -789,12 +823,11 @@ cmp_dict_index(const void *a1, const void *a2, void *arg)
 
 
 int
-PL_for_dict(term_t dict,
-	   int (*func)(term_t key, term_t value, int last, void *closure),
+pl_for_dict(DECL_LD term_t dict,
+	   int LDFUNCP (*func)(DECL_LD term_t key, term_t value, int last, void *closure),
 	   void *closure,
 	   int flags)
-{ GET_LD
-  term_t av = PL_new_term_refs(2);
+{ term_t av = PL_new_term_refs(2);
   int i, arity, pairs;
   Word p = valTermRef(dict);
   int index_buf[256];
@@ -835,10 +868,10 @@ PL_for_dict(term_t dict,
 
     deRef(p);
     Functor f = valueTerm(*p);
-    *valTermRef(av+0) = linkVal(&f->arguments[in+1]);
-    *valTermRef(av+1) = linkVal(&f->arguments[in]);
+    *valTermRef(av+0) = linkValI(&f->arguments[in+1]);
+    *valTermRef(av+1) = linkValI(&f->arguments[in]);
 
-    if ( (rc=(*func)(av+0, av+1, ++i == pairs, closure)) != 0 )
+    if ( (rc=LDFUNCP(*func)(av+0, av+1, ++i == pairs, closure)) != 0 )
       break;
   }
 
@@ -927,7 +960,7 @@ fix_firstvars(Code start, Code end)
   }
 }
 
-int
+static int
 resortDictsInCodes(Code PC, Code end)
 {
   for( ; PC < end; PC = stepPC(PC) )
@@ -957,6 +990,11 @@ resortDictsInCodes(Code PC, Code end)
 	    return PL_no_memory();
 
 	  PC = stepPC(PC);		/* skip *_FUNCTOR */
+	  if ( fetchop(PC) == H_VOID_N ) /* deal with _{key:_, ...} */
+	  { assert(PC[1] == (code)2);
+	    PC[0] = encode(H_VOID);
+	    PC[1] = encode(H_VOID);
+	  }
 	  PC = skipArgs(PC, 1);		/* skip the type */
 	  fields_start = PC;
 
@@ -1047,8 +1085,9 @@ resortDictsInClause(Clause clause)
    Used by loadQlfTerm().  Term may not be cyclic.
 */
 
+#define resort_dicts_in_term(p) LDFUNC(resort_dicts_in_term, p)
 static void
-resort_dicts_in_term(Word p ARG_LD)
+resort_dicts_in_term(DECL_LD Word p)
 {
 right_arg:
   deRef(p);
@@ -1057,16 +1096,17 @@ right_arg:
   { Functor t = valueTerm(*p);
     FunctorDef fd = valueFunctor(t->definition);
     Word ea;
+    word dupl;
 
     if ( fd->name == ATOM_dict && fd->arity%2 == 1 &&
-	 dict_ordered(&t->arguments[1], fd->arity/2, FALSE PASS_LD) == FALSE )
+	 dict_ordered(&t->arguments[1], fd->arity/2, &dupl) == FALSE )
     { DEBUG(MSG_DICT, Sdprintf("Re-ordering dict\n"));
-      dict_order((Word)t, FALSE PASS_LD);
+      dict_order((Word)t, &dupl);
     }
 
     ea = &t->arguments[fd->arity-1];
     for(p=t->arguments; p<ea; p++)
-      resort_dicts_in_term(p PASS_LD);
+      resort_dicts_in_term(p);
 
     goto right_arg;
   }
@@ -1078,7 +1118,7 @@ resortDictsInTerm(term_t t)
 { GET_LD
   Word p = valTermRef(t);
 
-  resort_dicts_in_term(p PASS_LD);
+  resort_dicts_in_term(p);
 }
 
 
@@ -1104,10 +1144,11 @@ PRED_IMPL("is_dict", 1, is_dict, 0)
   if ( isTerm(*p) )
   { Functor f = valueTerm(*p);
     FunctorDef fd = valueFunctor(f->definition);
+    //word dupl;
 
     if ( fd->name == ATOM_dict &&
 	 fd->arity%2 == 1 /*&&
-	 dict_ordered(f->arguments+1, fd->arity/2, FALSE PASS_LD) == TRUE*/ )
+	 dict_ordered(f->arguments+1, fd->arity/2, &dupl) == TRUE*/ )
       return TRUE;
   }
 
@@ -1124,12 +1165,13 @@ PRED_IMPL("is_dict", 2, is_dict, 0)
   if ( isTerm(*p) )
   { Functor f = valueTerm(*p);
     FunctorDef fd = valueFunctor(f->definition);
+    //word dupl;
 
     if ( fd->name == ATOM_dict &&
 	 fd->arity%2 == 1 /*&&
-	 dict_ordered(f->arguments+1, fd->arity/2, FALSE PASS_LD) == TRUE*/ )
+	 dict_ordered(f->arguments+1, fd->arity/2,  &dupl) == TRUE*/ )
       return unify_ptrs(&f->arguments[0], valTermRef(A2),
-			ALLOW_GC|ALLOW_SHIFT PASS_LD);
+			ALLOW_GC|ALLOW_SHIFT);
   }
 
   return FALSE;
@@ -1153,15 +1195,15 @@ pl_get_dict(term_t PL__t0, int PL__ac, int ex, control_t PL__ctx)
   { case FRG_FIRST_CALL:
     { Word np = valTermRef(A1);
 
-      if ( !get_dict_ex(A2, &dict, !ex PASS_LD) )
+      if ( !get_dict_ex(A2, &dict, !ex) )
 	return FALSE;
 
       deRef(np);
       if ( is_dict_key(*np) )
       { Word vp;
 
-	if ( (vp=dict_lookup_ptr(dict, *np PASS_LD)) )
-	  return unify_ptrs(vp, valTermRef(A3), ALLOW_GC|ALLOW_SHIFT PASS_LD);
+	if ( (vp=dict_lookup_ptr(dict, *np)) )
+	  return unify_ptrs(vp, valTermRef(A3), ALLOW_GC|ALLOW_SHIFT);
 
 	if ( ex )
 	  return PL_error(NULL, 0, NULL, ERR_EXISTENCE3,
@@ -1197,7 +1239,7 @@ pl_get_dict(term_t PL__t0, int PL__ac, int ex, control_t PL__ctx)
 
 	  deRef2(&f->arguments[i+1], np);	/* TBD: check type */
 	  if ( unify_ptrs(&f->arguments[i], valTermRef(A3),
-			  ALLOW_GC|ALLOW_SHIFT PASS_LD) &&
+			  ALLOW_GC|ALLOW_SHIFT) &&
 	       _PL_unify_atomic(A1, *np) )
 	  { PL_close_foreign_frame(fid);
 
@@ -1244,11 +1286,11 @@ PRED_IMPL("get_dict", 5, get_dict, 0)
   word key;
   Word vp;
 
-  if ( !get_name_ex(A1, &key PASS_LD) ||
+  if ( !get_name_ex(A1, &key) ||
        !(*valTermRef(av+1) = key) ||
-       !get_create_dict_ex(A2, dt PASS_LD) ||
-       !(vp=dict_lookup_ptr(*valTermRef(dt), key PASS_LD)) ||
-       !unify_ptrs(vp, valTermRef(A3), ALLOW_GC|ALLOW_SHIFT PASS_LD) ||
+       !get_create_dict_ex(A2, dt) ||
+       !(vp=dict_lookup_ptr(*valTermRef(dt), key)) ||
+       !unify_ptrs(vp, valTermRef(A3), ALLOW_GC|ALLOW_SHIFT) ||
        !PL_put_term(av+0, A5) )
     return FALSE;
 
@@ -1257,7 +1299,7 @@ PRED_IMPL("get_dict", 5, get_dict, 0)
     int rc;
 
     if ( (rc = put_dict(*valTermRef(dt),
-			1, valTermRef(av), &new PASS_LD)) == TRUE )
+			1, valTermRef(av), &new)) == TRUE )
     { term_t t = dt+3;
 
       *valTermRef(t) = new;
@@ -1306,11 +1348,11 @@ typedef struct dict_pairs_ctx
   term_t tmp;
 } dict_pairs_ctx;
 
+#define put_pair(key, value, last, closure) LDFUNC(put_pair, key, value, last, closure)
 
 static int
-put_pair(term_t key, term_t value, int last, void *closure)
+put_pair(DECL_LD term_t key, term_t value, int last, void *closure)
 { dict_pairs_ctx *ctx = closure;
-  GET_LDARG(ctx->ld);
 
   if ( PL_cons_functor(ctx->tmp, FUNCTOR_minus2, key, value) &&
        PL_unify_list_ex(ctx->tail, ctx->head, ctx->tail) &&
@@ -1320,8 +1362,6 @@ put_pair(term_t key, term_t value, int last, void *closure)
   return -1;
 }
 
-#undef GET_LDARG
-
 
 static
 PRED_IMPL("dict_pairs", 3, dict_pairs, 0)
@@ -1330,7 +1370,7 @@ PRED_IMPL("dict_pairs", 3, dict_pairs, 0)
   if ( !PL_is_variable(A1) )
   { term_t dict = PL_new_term_ref();
 
-    if ( get_create_dict_ex(A1, dict PASS_LD) )
+    if ( get_create_dict_ex(A1, dict) )
     { dict_pairs_ctx ctx;
 
       ctx.ld = LD;
@@ -1370,8 +1410,8 @@ PRED_IMPL("put_dict", 3, put_dict, 0)
 
 retry:
   if ( (dt = PL_new_term_refs(2)) &&
-       get_create_dict_ex(A2, dt+0 PASS_LD) &&
-       get_create_dict_ex(A1, dt+1 PASS_LD) )
+       get_create_dict_ex(A2, dt+0) &&
+       get_create_dict_ex(A1, dt+1) )
   { Functor f2 = valueTerm(*valTermRef(dt+1));
     int arity = arityFunctor(f2->definition);
     word new;
@@ -1379,7 +1419,7 @@ retry:
 
     if ( (rc = put_dict(*valTermRef(dt+0),
 			arity/2, &f2->arguments[1],
-			&new PASS_LD)) == TRUE )
+			&new)) == TRUE )
     { term_t t = PL_new_term_ref();
 
       *valTermRef(t) = new;
@@ -1400,21 +1440,22 @@ retry:
 True when Dict is a copy of Dict0 with Name Value added or replaced.
 */
 
+#define put_dict4(key, dict, value, newdict) LDFUNC(put_dict4, key, dict, value, newdict)
 static foreign_t
-put_dict4(term_t key, term_t dict, term_t value, term_t newdict ARG_LD)
+put_dict4(DECL_LD term_t key, term_t dict, term_t value, term_t newdict)
 { term_t dt = PL_new_term_refs(3);
   term_t av = dt+1;
   fid_t fid = PL_open_foreign_frame();
 
 retry:
-  if ( get_create_dict_ex(dict, dt PASS_LD) &&
-       get_name_ex(key, valTermRef(av+1) PASS_LD) &&
+  if ( get_create_dict_ex(dict, dt) &&
+       get_name_ex(key, valTermRef(av+1)) &&
        PL_put_term(av, value) )
   { word new;
     int rc;
 
     if ( (rc = put_dict(*valTermRef(dt),
-			1, valTermRef(av), &new PASS_LD)) == TRUE )
+			1, valTermRef(av), &new)) == TRUE )
     { term_t t = PL_new_term_ref();
 
       *valTermRef(t) = new;
@@ -1436,7 +1477,7 @@ static
 PRED_IMPL("put_dict", 4, put_dict, 0)
 { PRED_LD
 
-  return put_dict4(A1, A2, A3, A4 PASS_LD);
+  return put_dict4(A1, A2, A3, A4);
 }
 
 
@@ -1448,8 +1489,9 @@ Backtrackable destructive assignment, similar to setarg/3.
 #define SETDICT_BACKTRACKABLE    0x1
 #define SETDICT_LINK		0x2
 
+#define setdict(key, dict, value, flags) LDFUNC(setdict, key, dict, value, flags)
 static int
-setdict(term_t key, term_t dict, term_t value, int flags ARG_LD)
+setdict(DECL_LD term_t key, term_t dict, term_t value, int flags)
 { word k, m;
   Word val;
 
@@ -1470,24 +1512,24 @@ retry:
     { if ( !(flags & SETDICT_LINK) )
       { term_t copy = PL_new_term_ref();
 
-	if ( !duplicate_term(value, copy PASS_LD) )
+	if ( !duplicate_term(value, copy) )
 	  return FALSE;
 	value = copy;
 	val = valTermRef(value);
 	deRef(val);
       }
-      freezeGlobal(PASS_LD1);
+      freezeGlobal();
     }
   }
 
-  if ( get_dict_ex(dict, &m, TRUE PASS_LD) &&
-       get_name_ex(key, &k PASS_LD) )
+  if ( get_dict_ex(dict, &m, TRUE) &&
+       get_name_ex(key, &k) )
   { Word vp;
 
-    if ( (vp=dict_lookup_ptr(m, k PASS_LD)) )
+    if ( (vp=dict_lookup_ptr(m, k)) )
     { if ( (flags&SETDICT_BACKTRACKABLE) )
 	TrailAssignment(vp);
-      unify_vp(vp, val PASS_LD);
+      unify_vp(vp, val);
       return TRUE;
     }
 
@@ -1503,21 +1545,21 @@ static
 PRED_IMPL("b_set_dict", 3, b_set_dict, 0)
 { PRED_LD
 
-  return setdict(A1, A2, A3, SETDICT_BACKTRACKABLE PASS_LD);
+  return setdict(A1, A2, A3, SETDICT_BACKTRACKABLE);
 }
 
 static
 PRED_IMPL("nb_set_dict", 3, nb_set_dict, 0)
 { PRED_LD
 
-  return setdict(A1, A2, A3, 0 PASS_LD);
+  return setdict(A1, A2, A3, 0);
 }
 
 static
 PRED_IMPL("nb_link_dict", 3, nb_link_dict, 0)
 { PRED_LD
 
-  return setdict(A1, A2, A3, SETDICT_LINK PASS_LD);
+  return setdict(A1, A2, A3, SETDICT_LINK);
 }
 
 
@@ -1535,16 +1577,16 @@ PRED_IMPL("del_dict", 4, del_dict, 0)
   fid_t fid = PL_open_foreign_frame();
 
 retry:
-  if ( get_create_dict_ex(A2, mt PASS_LD) &&
-       get_name_ex(A1, &key PASS_LD) )
+  if ( get_create_dict_ex(A2, mt) &&
+       get_name_ex(A1, &key) )
   { Word vp;
 
-    if ( (vp=dict_lookup_ptr(*valTermRef(mt), key PASS_LD)) &&
-	 unify_ptrs(vp, valTermRef(A3), ALLOW_GC|ALLOW_SHIFT PASS_LD) )
+    if ( (vp=dict_lookup_ptr(*valTermRef(mt), key)) &&
+	 unify_ptrs(vp, valTermRef(A3), ALLOW_GC|ALLOW_SHIFT) )
     { int rc;
       word new;
 
-      if ( (rc=del_dict(*valTermRef(mt), key, &new PASS_LD)) == TRUE )
+      if ( (rc=del_dict(*valTermRef(mt), key, &new)) == TRUE )
       { term_t t = PL_new_term_ref();
 
 	*valTermRef(t) = new;
@@ -1574,9 +1616,9 @@ PRED_IMPL("select_dict", 3, select_dict, 0)
   word r;
 
 retry:
-  if ( get_create_dict_ex(A1, dt+0 PASS_LD) &&
-       get_create_dict_ex(A2, dt+1 PASS_LD) )
-  { int rc = select_dict(*valTermRef(dt+0), *valTermRef(dt+1), &r PASS_LD);
+  if ( get_create_dict_ex(A1, dt+0) &&
+       get_create_dict_ex(A2, dt+1) )
+  { int rc = select_dict(*valTermRef(dt+0), *valTermRef(dt+1), &r);
 
     switch(rc)
     { case TRUE:
@@ -1606,9 +1648,9 @@ PRED_IMPL(":<", 2, select_dict, 0)
   term_t dt = PL_new_term_refs(2);
 
 retry:
-  if ( get_create_dict_ex(A1, dt+0 PASS_LD) &&
-       get_create_dict_ex(A2, dt+1 PASS_LD) )
-  { int rc = select_dict(*valTermRef(dt+0), *valTermRef(dt+1), NULL PASS_LD);
+  if ( get_create_dict_ex(A1, dt+0) &&
+       get_create_dict_ex(A2, dt+1) )
+  { int rc = select_dict(*valTermRef(dt+0), *valTermRef(dt+1), NULL);
 
     switch(rc)
     { case TRUE:
@@ -1633,9 +1675,9 @@ PRED_IMPL(">:<", 2, punify_dict, 0)
   term_t dt = PL_new_term_refs(2);
 
 retry:
-  if ( get_create_dict_ex(A1, dt+0 PASS_LD) &&
-       get_create_dict_ex(A2, dt+1 PASS_LD) )
-  { int rc = partial_unify_dict(*valTermRef(dt+0), *valTermRef(dt+1) PASS_LD);
+  if ( get_create_dict_ex(A1, dt+0) &&
+       get_create_dict_ex(A2, dt+1) )
+  { int rc = partial_unify_dict(*valTermRef(dt+0), *valTermRef(dt+1));
 
     switch(rc)
     { case TRUE:
@@ -1666,10 +1708,10 @@ PL_get_dict_key(atom_t key, term_t dict, term_t value)
 
   if ( !is_dict_key(key) )
     return -1;
-  if ( !get_dict_ex(dict, &d, FALSE PASS_LD) )
+  if ( !get_dict_ex(dict, &d, FALSE) )
     return FALSE;
-  if ( (vp=dict_lookup_ptr(d, key PASS_LD)) )
-  { *valTermRef(value) = linkVal(vp);
+  if ( (vp=dict_lookup_ptr(d, key)) )
+  { *valTermRef(value) = linkValI(vp);
     return TRUE;
   }
 
