@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1997-2020, University of Amsterdam
+    Copyright (c)  1997-2021, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -37,6 +38,9 @@
 #ifndef PL_GLOBAL_H_INCLUDED
 #define PL_GLOBAL_H_INCLUDED
 #include "pl-allocpool.h"
+#include "pl-mutex.h"
+#include "pl-thread.h"
+#include "pl-gmp.h"
 
 #ifndef GLOBAL			/* global variables */
 #define GLOBAL extern
@@ -107,16 +111,11 @@ struct PL_global_data
 
   struct
   { sig_handler handlers[MAXSIGNAL];	/* How Prolog preceives signals */
-#ifdef HAVE_SIGNAL
     int		sig_alert;		/* our alert signal */
 #endif
   } signals;
-#ifdef O_LOGICAL_UPDATE
-  volatile ggen_t _generation;		/* generation of the database */
-#ifdef ATOMIC_GENERATION_HACK
-  volatile gen_t _last_generation;	/* see pl-inline.h, global_generation() */
-#endif
-#endif
+
+  volatile gen_t _generation;		/* generation of the database */
 
   struct
   { int		os_argc;		/* main(int argc, char **argv) */
@@ -160,6 +159,8 @@ struct PL_global_data
     int		engines_finished;	/* # engines threads */
     double	thread_cputime;		/* Total CPU time of threads */
 #endif
+    int		errors;			/* Printed error messages */
+    int		warnings;		/* Printed warning messages */
   } statistics;
 
 #ifdef O_PROFILE
@@ -279,6 +280,7 @@ struct PL_global_data
       struct event_list *onbreak;	/* breakpoint change */
       struct event_list *onframefinish; /* Debugged frame finished */
 #ifdef O_PLMT
+      struct event_list *onthreadstart;	/* thread start hook */
       struct event_list *onthreadexit;	/* thread exit hook */
 #endif
       struct event_list *onuntable;	/* Untable after reload */
@@ -313,6 +315,7 @@ struct PL_global_data
     Procedure	is2;			/* is/2 */
     Procedure	strict_equal2;		/* ==/2 */
     Procedure	not_strict_equal2;	/* \==/2 */
+    Procedure   arg3;			/* arg/3 */
     Procedure	exception_hook4;
     Procedure	print_message2;
     Procedure	foreign_registered2;	/* $foreign_registered/2 */
@@ -443,7 +446,6 @@ struct PL_local_data
 #endif
   int		in_arithmetic;		/* doing arithmetic */
   int		in_print_message;	/* Inside printMessage() */
-  gen_t		gen_reload;		/* reload generation */
   void *	glob_info;		/* pl-glob.c */
   IOENC		encoding;		/* default I/O encoding */
   struct PL_local_data *next_free;	/* see maybe_free_local_data() */
@@ -454,6 +456,12 @@ struct PL_local_data
     int		is_sync;		/* current signal is synchronous */
 #ifndef __unix__
     int		forced;			/* Forced signal */
+#endif
+#ifdef HAVE_SIGNAL
+#ifdef O_C_STACK_GUARDED
+    jmp_buf	context;		/* Guarded for C-stack overflows */
+    int		sig_critical;		/* context is valid */
+    void       *alt_stack;		/* The stack */
 #endif
   } signal;
 
@@ -524,6 +532,8 @@ struct PL_local_data
     double	last_walltime;		/* Last Wall time (m-secs since start) */
     double	user_cputime;		/* User saved CPU time */
     double	system_cputime;		/* Kernel saved CPU time */
+    int		errors;			/* Printed error messages */
+    int		warnings;		/* Printed warning messages */
   } statistics;
 
 #ifdef O_GMP
@@ -614,6 +624,8 @@ struct PL_local_data
     alloc_pool *node_pool;		/* Node allocation pool for tries */
     int	has_scheduling_component;	/* A leader was created */
     int in_answer_completion;		/* Running answer completion */
+    int in_assert_propagation;		/* Running propagate_assert/1 */
+    int flags;				/* Global flags (TF_*) */
     term_t delay_list;			/* Global delay list */
     term_t idg_current;			/* Current node in IDG (trie symbol) */
     struct
@@ -698,6 +710,9 @@ struct PL_local_data
   pl_shift_status_t shift_status;	/* Stack shifter status */
   pl_debugstatus_t _debugstatus;	/* status of the debugger */
   struct btrace *btrace_store;		/* C-backtraces */
+#if O_DEBUG
+  pl_internaldebugstatus_t internal_debug; /* status of C-level debug flags */
+#endif
 
 #ifdef O_PLMT
   struct
@@ -709,10 +724,25 @@ struct PL_local_data
     struct _thread_sig   *sig_tail;	/* Tail of signal queue */
     DefinitionChain local_definitions;	/* P_THREAD_LOCAL predicates */
     simpleMutex scan_lock;		/* Hold for asynchronous scans */
+    thread_wait_for *waiting_for;	/* thread_wait/2 info */
     alert_channel alert;		/* How to alert the thread */
   } thread;
 #endif
 
+  struct
+  { gen_t	      gen_start;	/* Global start generation */
+    gen_t	      gen_base;		/* Local  start generation */
+    gen_t	      gen_max;		/* Transaction max gen */
+    gen_t	      gen_nest;		/* Start of nested generation */
+    gen_t	      generation;	/* Local current generation */
+    Table	      clauses;		/* Affected clauses */
+    Table	      predicates;	/* Pred --> last modified */
+    struct tbl_trail *table_trail;	/* Affected tables */
+    term_t	      id;		/* Default the goal */
+    struct tr_stack  *stack;		/* Nested transaction stack */
+    unsigned int      flags;		/*  */
+  } transaction;
+
 #ifdef O_LOCALE
   struct
   { PL_locale *current;			/* Current locale */
@@ -724,11 +754,22 @@ struct PL_local_data
     int64_t	cgc_inferences;		/* Inferences at last cgc consider */
   } clauses;
 
+  struct
+  { gen_t	generation;		/* reload generation */
+    int		nesting;		/* reload nesting */
+  } reload;
+
   struct
   { DefinitionChain nesting;		/* Nesting chain in the autoloader */
     Definition	loop;			/* We are looping on this def */
   } autoload;
 
+  struct
+  { term_t undo_list;			/* Stacked undo goals */
+    Buffer scheduled;
+    int    running;
+  } undo;
+
   struct
   { intptr_t _total_marked;		/* # marked global cells */
     intptr_t _trailcells_deleted;	/* # garbage trailcells */
@@ -773,6 +814,12 @@ GLOBAL PL_local_data_t  PL_local_data;
 #ifdef O_MULTIPLE_ENGINES
 GLOBAL PL_local_data_t *PL_current_engine_ptr;
 #endif
+/* See _LD_WITH_FALLBACK().  These should never be compiled,
+   but are on clang in debug mode
+ */
+GLOBAL char *__PL_ld;
+GLOBAL struct {char *engine;} *PL__ctx;
+
 
 #define GD (&PL_global_data)
 #define CD (&PL_code_data)
