@@ -33,9 +33,13 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-#include "pl-incl.h"
+#include "pl-index.h"
 #include "pl-comp.h"
 #include "pl-rsort.h"
+#include "pl-util.h"
+#include "pl-proc.h"
+#include "pl-fli.h"
+#include "pl-wam.h"
 #include <math.h>
 
 		 /*******************************
@@ -82,8 +86,16 @@ typedef struct index_context
   iarg_t	position[MAXINDEXDEPTH+1]; /* Keep track of argument position */
 } index_context, *IndexContext;
 
+#if USE_LD_MACROS
+#define	bestHash(av, ac, clist, min_speedup, hints, ctx)	LDFUNC(bestHash, av, ac, clist, min_speedup, hints, ctx)
+#define	setClauseChoice(chp, cref, generation)			LDFUNC(setClauseChoice, chp, cref, generation)
+#define	first_clause_guarded(argv, argc, clist, ctx)		LDFUNC(first_clause_guarded, argv, argc, clist, ctx)
+#endif /*USE_LD_MACROS*/
+
+#define LDFUNC_DECLARATIONS
+
 static int	bestHash(Word av, size_t ac, ClauseList clist, float min_speedup,
-			 hash_hints *hints, IndexContext ctx ARG_LD);
+			 hash_hints *hints, IndexContext ctx);
 static ClauseIndex hashDefinition(ClauseList clist, hash_hints *h,
 				  IndexContext ctx);
 static void	replaceIndex(Definition def, ClauseList cl,
@@ -92,7 +104,7 @@ static void	deleteIndexP(Definition def, ClauseList cl, ClauseIndex *cip);
 static void	deleteIndex(Definition def, ClauseList cl, ClauseIndex ci);
 static void	insertIndex(Definition def, ClauseList clist, ClauseIndex ci);
 static void	setClauseChoice(ClauseChoice chp, ClauseRef cref,
-				gen_t generation ARG_LD);
+				gen_t generation);
 static int	addClauseToIndex(ClauseIndex ci, Clause cl, ClauseRef where);
 static void	addClauseToListIndexes(Definition def, ClauseList cl,
 				       Clause clause, ClauseRef where);
@@ -100,12 +112,14 @@ static void	insertIntoSparseList(ClauseRef cref,
 				     ClauseRef *headp, ClauseRef *tailp,
 				     ClauseRef where);
 static ClauseRef first_clause_guarded(Word argv, size_t argc, ClauseList clist,
-				      IndexContext ctx ARG_LD);
+				      IndexContext ctx);
 static Code	skipToTerm(Clause clause, const iarg_t *position);
 static void	unalloc_index_array(void *p);
 static void	wait_for_index(const ClauseIndex ci);
 static void	completed_index(ClauseIndex ci);
 
+#undef LDFUNC_DECLARATIONS
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Compute the index in the hash-array from   a machine word and the number
 of buckets. This used to be simple, but now that our tag bits are on the
@@ -129,8 +143,9 @@ hashIndex(word key, int buckets)
 }
 
 
+#define canIndex(w) LDFUNC(canIndex, w)
 static inline int
-canIndex(word w ARG_LD)
+canIndex(DECL_LD word w)
 { for(;;)
   { switch(tag(w))
     { case TAG_VAR:
@@ -146,8 +161,9 @@ canIndex(word w ARG_LD)
 }
 
 
+#define indexOfWord(w) LDFUNC(indexOfWord, w)
 static inline word
-indexOfWord(word w ARG_LD)
+indexOfWord(DECL_LD word w)
 { for(;;)
   { switch(tag(w))
     { case TAG_VAR:
@@ -163,12 +179,8 @@ indexOfWord(word w ARG_LD)
       case TAG_FLOAT:
       { Word p = addressIndirect(w);
 	size_t n = wsizeofInd(*p);
-	word k;
 
-	k = MurmurHashAligned2(p+1, n*sizeof(*p), MURMUR_SEED);
-	k &= ~((word)STG_GLOBAL);	/* avoid confusion with functor_t */
-	if ( !k ) k = 1;		/* avoid no-key */
-	return k;
+	return murmur_key(p+1, n*sizeof(*p));
       }
       case TAG_COMPOUND:
 	w = *valPtr(w);			/* functor_t */
@@ -188,12 +200,13 @@ getIndexOfTerm(term_t t)
 { GET_LD
   word w = *valTermRef(t);
 
-  return indexOfWord(w PASS_LD);
+  return indexOfWord(w);
 }
 
 
+#define nextClauseArg1(chp, generation) LDFUNC(nextClauseArg1, chp, generation)
 static inline ClauseRef
-nextClauseArg1(ClauseChoice chp, gen_t generation ARG_LD)
+nextClauseArg1(DECL_LD ClauseChoice chp, gen_t generation)
 { ClauseRef cref = chp->cref;
   word key = chp->key;
 
@@ -211,7 +224,7 @@ nextClauseArg1(ClauseChoice chp, gen_t generation ARG_LD)
 	  }
 	}
 	if ( --maxsearch == 0 )
-	{ setClauseChoice(chp, cref, generation PASS_LD);
+	{ setClauseChoice(chp, cref, generation);
 	  return result;
 	}
       }
@@ -242,8 +255,9 @@ argument we are processing.
 TBD: Keep a flag telling whether there are non-indexable clauses.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define nextClauseFromBucket(ci, argv, ctx) LDFUNC(nextClauseFromBucket, ci, argv, ctx)
 static ClauseRef
-nextClauseFromBucket(ClauseIndex ci, Word argv, IndexContext ctx ARG_LD)
+nextClauseFromBucket(DECL_LD ClauseIndex ci, Word argv, IndexContext ctx)
 { ClauseRef cref;
   word key = ctx->chp->key;
 
@@ -278,13 +292,13 @@ nextClauseFromBucket(ClauseIndex ci, Word argv, IndexContext ctx ARG_LD)
 	  DEBUG(MSG_INDEX_DEEP,
 		Sdprintf("Recursive index for %s at level %d\n",
 			 keyName(cref->d.key), ctx->depth));
-	  return first_clause_guarded(argv, argc, cl, ctx PASS_LD);
+	  return first_clause_guarded(argv, argc, cl, ctx);
 	}
 
 	ctx->chp->key = 0;		/* See (*) */
 	for(cr=cl->first_clause; cr; cr=cr->next)
 	{ if ( visibleClauseCNT(cr->value.clause, ctx->generation) )
-	  { setClauseChoice(ctx->chp, cr->next, ctx->generation PASS_LD);
+	  { setClauseChoice(ctx->chp, cr->next, ctx->generation);
 	    return cr;
 	  }
 	}
@@ -314,7 +328,7 @@ nextClauseFromBucket(ClauseIndex ci, Word argv, IndexContext ctx ARG_LD)
       { if ( ((!cref->d.key || key == cref->d.key) &&
 	      visibleClauseCNT(cref->value.clause, ctx->generation)) ||
 	     --maxsearch == 0 )
-	{ setClauseChoice(ctx->chp, cref, ctx->generation PASS_LD);
+	{ setClauseChoice(ctx->chp, cref, ctx->generation);
 
 	  return result;
 	}
@@ -336,34 +350,25 @@ nextClauseFromBucket(ClauseIndex ci, Word argv, IndexContext ctx ARG_LD)
 */
 
 static void
-setClauseChoice(ClauseChoice chp, ClauseRef cref, gen_t generation ARG_LD)
-{ while ( cref && !GLOBALLY_VISIBLE_CLAUSE(cref->value.clause, generation) )
-  { cref = cref->next;
-    LD->clauses.erased_skipped++;
-  }
+setClauseChoice(DECL_LD ClauseChoice chp, ClauseRef cref, gen_t generation)
+{ while ( cref && !visibleClauseCNT(cref->value.clause, generation) )
+    cref = cref->next;
 
   chp->cref = cref;
 }
 
 
+#define indexKeyFromArgv(ci, argv) LDFUNC(indexKeyFromArgv, ci, argv)
 static inline word
-murmur_key(void *ptr, size_t n)
-{ word k = MurmurHashAligned2(ptr, n, MURMUR_SEED);
-  if ( !k ) k = 1;
-  return k;
-}
-
-
-static inline word
-indexKeyFromArgv(ClauseIndex ci, Word argv ARG_LD)
+indexKeyFromArgv(DECL_LD ClauseIndex ci, Word argv)
 { if ( likely(ci->args[1] == 0) )
-  { return indexOfWord(argv[ci->args[0]-1] PASS_LD);
+  { return indexOfWord(argv[ci->args[0]-1]);
   } else
   { word key[MAX_MULTI_INDEX];
     int  harg;
 
     for(harg=0; ci->args[harg]; harg++)
-    { if ( !(key[harg] = indexOfWord(argv[ci->args[harg]-1] PASS_LD)) )
+    { if ( !(key[harg] = indexOfWord(argv[ci->args[harg]-1])) )
 	return 0;
     }
 
@@ -409,14 +414,14 @@ TBD:
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static ClauseRef
-first_clause_guarded(Word argv, size_t argc, ClauseList clist,
-		     IndexContext ctx ARG_LD)
+first_clause_guarded(DECL_LD Word argv, size_t argc, ClauseList clist,
+		     IndexContext ctx)
 { ClauseRef cref;
   ClauseIndex *cip;
   hash_hints hints;
   ClauseChoice chp = ctx->chp;
 
-#define STATIC_RELOADING() (LD->gen_reload && false(ctx->predicate, P_DYNAMIC))
+#define STATIC_RELOADING() (LD->reload.generation && false(ctx->predicate, P_DYNAMIC))
 
   if ( unlikely(argc == 0) )
     goto simple;			/* TBD: alt supervisor */
@@ -434,7 +439,7 @@ retry:
       if ( ISDEADCI(ci) )
 	continue;
 
-      if ( (k=indexKeyFromArgv(ci, argv PASS_LD)) )
+      if ( (k=indexKeyFromArgv(ci, argv)) )
       { best_index = ci;
 	chp->key = k;
 	break;
@@ -453,7 +458,7 @@ retry:
 		       predicateName(ctx->predicate)));
 
 	if ( bestHash(argv, argc, clist, best_index->speedup,
-		      &hints, ctx PASS_LD) )
+		      &hints, ctx) )
 	{ ClauseIndex ci;
 
 	  DEBUG(MSG_JIT, Sdprintf("[%d] Found better at args %s\n",
@@ -461,7 +466,7 @@ retry:
 				  iargsName(hints.args, NULL)));
 
 	  if ( (ci=hashDefinition(clist, &hints, ctx)) )
-	  { chp->key = indexKeyFromArgv(ci, argv PASS_LD);
+	  { chp->key = indexKeyFromArgv(ci, argv);
 	    assert(chp->key);
 	    best_index = ci;
 	  } else
@@ -477,18 +482,18 @@ retry:
 
       hi = hashIndex(chp->key, best_index->buckets);
       chp->cref = best_index->entries[hi].head;
-      return nextClauseFromBucket(best_index, argv, ctx PASS_LD);
+      return nextClauseFromBucket(best_index, argv, ctx);
     }
   }
 
   if ( unlikely(clist->number_of_clauses == 0) )
     return NULL;
 
-  if ( (chp->key = indexOfWord(argv[0] PASS_LD)) &&
+  if ( (chp->key = indexOfWord(argv[0])) &&
        (clist->number_of_clauses <= 10 || STATIC_RELOADING()) )
   { chp->cref = clist->first_clause;
 
-    cref = nextClauseArg1(chp, ctx->generation PASS_LD);
+    cref = nextClauseArg1(chp, ctx->generation);
     if ( !cref ||
 	 !(chp->cref && chp->cref->d.key == chp->key &&
 	   cref->d.key == chp->key) )
@@ -498,7 +503,7 @@ retry:
   }
 
   if ( !STATIC_RELOADING() &&
-       bestHash(argv, argc, clist, 0.0, &hints, ctx PASS_LD) )
+       bestHash(argv, argc, clist, 0.0, &hints, ctx) )
   { ClauseIndex ci;
 
     if ( (ci=hashDefinition(clist, &hints, ctx)) )
@@ -509,11 +514,11 @@ retry:
       if ( ci->invalid )
 	goto retry;
 
-      chp->key = indexKeyFromArgv(ci, argv PASS_LD);
+      chp->key = indexKeyFromArgv(ci, argv);
       assert(chp->key);
       hi = hashIndex(chp->key, ci->buckets);
       chp->cref = ci->entries[hi].head;
-      return nextClauseFromBucket(ci, argv, ctx PASS_LD);
+      return nextClauseFromBucket(ci, argv, ctx);
     } else
     { goto retry;
     }
@@ -521,14 +526,14 @@ retry:
 
   if ( chp->key )
   { chp->cref = clist->first_clause;
-    return nextClauseArg1(chp, ctx->generation PASS_LD);
+    return nextClauseArg1(chp, ctx->generation);
   }
 
 simple:
   for(cref = clist->first_clause; cref; cref = cref->next)
   { if ( visibleClauseCNT(cref->value.clause, ctx->generation) )
     { chp->key = 0;
-      setClauseChoice(chp, cref->next, ctx->generation PASS_LD);
+      setClauseChoice(chp, cref->next, ctx->generation);
       break;
     }
   }
@@ -536,10 +541,9 @@ simple:
   return cref;
 }
 
-int acquired = 0;
 
 ClauseRef
-firstClause(Word argv, LocalFrame fr, Definition def, ClauseChoice chp ARG_LD)
+firstClause(DECL_LD Word argv, LocalFrame fr, Definition def, ClauseChoice chp)
 { ClauseRef cref;
   index_context ctx;
 
@@ -549,13 +553,13 @@ firstClause(Word argv, LocalFrame fr, Definition def, ClauseChoice chp ARG_LD)
   ctx.depth       = 0;
   ctx.position[0] = END_INDEX_POS;
 
+  MEMORY_ACQUIRE();			/* sync with retract_clause() */
   acquire_def(def);
   cref = first_clause_guarded(argv,
 			      def->functor->arity,
 			      &def->impl.clauses,
-			      &ctx
-			      PASS_LD);
-#define CHK_STATIC_RELOADING() (LD->gen_reload && false(def, P_DYNAMIC))
+			      &ctx);
+#define CHK_STATIC_RELOADING() (LD->reload.generation && false(def, P_DYNAMIC))
   DEBUG(CHK_SECURE, assert(!cref || !chp->cref ||
 			   visibleClause(chp->cref->value.clause,
 					 generationFrame(fr)) ||
@@ -567,29 +571,30 @@ firstClause(Word argv, LocalFrame fr, Definition def, ClauseChoice chp ARG_LD)
 
 
 ClauseRef
-nextClause__LD(ClauseChoice chp, Word argv, LocalFrame fr, Definition def ARG_LD)
+nextClause(DECL_LD ClauseChoice chp, Word argv, LocalFrame fr, Definition def)
 { gen_t generation = generationFrame(fr);
   ClauseRef cref;
 
   (void)argv;				/* we want to use these later */
   (void)def;				/* to create secondary indexes */
 
+  MEMORY_ACQUIRE();
   acquire_def(def);
   if ( !chp->key )			/* not indexed */
   { for(cref=chp->cref; cref; cref = cref->next)
     { if ( visibleClauseCNT(cref->value.clause, generation) )
-      { setClauseChoice(chp, cref->next, generation PASS_LD);
+      { setClauseChoice(chp, cref->next, generation);
 	break;
       }
     }
   } else
-  { cref = nextClauseArg1(chp, generation PASS_LD);
+  { cref = nextClauseArg1(chp, generation);
   }
   release_def(def);
 
   DEBUG(CHK_SECURE,
 	assert(!cref || !chp->cref ||
-	       chp->cref->value.clause->generation.erased > generation));
+	       visibleClause(chp->cref->value.clause, generation)));
 
   return cref;
 }
@@ -1029,40 +1034,46 @@ deleteClauseBucket(ClauseBucket ch, Clause clause, word key, int is_list)
 (*) The  clause  list   may   have   been   marked    as   "dirty"   by
 deleteActiveClauseFromBucket()  even though  it  does  not  contain the
 clause being deleted. We reset the count when we have finished scanning
-the list.
+the list to the number of erased clauses remaining in the list.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-gcClauseList(ClauseList clist, DirtyDefInfo ddi, gen_t start)
+gcClauseList(ClauseList clist, DirtyDefInfo ddi, gen_t start, Buffer tr_starts)
 { ClauseRef cref=clist->first_clause, prev = NULL;
+  size_t left = 0;
 
   while(cref && clist->erased_clauses)
   { Clause cl = cref->value.clause;
 
-    if ( true(cl, CL_ERASED) && ddi_is_garbage(ddi, start, cl) )
-    { ClauseRef c = cref;
+    if ( true(cl, CL_ERASED) )
+    { if ( ddi_is_garbage(ddi, start, tr_starts, cl) )
+      { ClauseRef c = cref;
 
-      clist->erased_clauses--;
+	clist->erased_clauses--;
 
-      cref = cref->next;
-      if ( !prev )
-      { clist->first_clause = c->next;
-	if ( !c->next )
-	  clist->last_clause = NULL;
+	cref = cref->next;
+	if ( !prev )
+	{ clist->first_clause = c->next;
+	  if ( !c->next )
+	    clist->last_clause = NULL;
+	} else
+	{ prev->next = c->next;
+	  if ( c->next == NULL)
+	    clist->last_clause = prev;
+	}
+
+	lingerClauseRef(c);
+	continue;
       } else
-      { prev->next = c->next;
-	if ( c->next == NULL)
-	  clist->last_clause = prev;
+      { left++;
       }
-
-      lingerClauseRef(c);
-    } else
-    { prev = cref;
-      cref = cref->next;
     }
+
+    prev = cref;
+    cref = cref->next;
   }
 
-  clist->erased_clauses = 0; /* see (*) */
+  clist->erased_clauses = left; /* see (*) */
 }
 
 
@@ -1073,7 +1084,8 @@ the number of indexable entries that have been removed from the bucket.
 
 static int
 gcClauseBucket(Definition def, ClauseBucket ch,
-	       unsigned int dirty, int is_list, DirtyDefInfo ddi, gen_t start)
+	       unsigned int dirty, int is_list, DirtyDefInfo ddi,
+	       gen_t start, Buffer tr_starts)
 { ClauseRef cref = ch->head, prev = NULL;
   int deleted = 0;
 
@@ -1082,8 +1094,9 @@ gcClauseBucket(Definition def, ClauseBucket ch,
     { ClauseList cl = &cref->value.clauses;
 
       if ( cl->erased_clauses )
-      { gcClauseList(cl, ddi, start);
-	dirty--;
+      { gcClauseList(cl, ddi, start, tr_starts);
+	if ( !cl->erased_clauses )
+	  dirty--;
 
 	if ( cl->first_clause == NULL )
 	  goto delete;
@@ -1091,7 +1104,7 @@ gcClauseBucket(Definition def, ClauseBucket ch,
     } else
     { Clause cl = cref->value.clause;
 
-      if ( true(cl, CL_ERASED) && ddi_is_garbage(ddi, start, cl) )
+      if ( true(cl, CL_ERASED) && ddi_is_garbage(ddi, start, tr_starts, cl) )
       { ClauseRef c;
 
 	dirty--;
@@ -1130,7 +1143,7 @@ gcClauseBucket(Definition def, ClauseBucket ch,
 	  { for(cref=ch->head; cref; cref=cref->next)
 	    { Clause cl = cref->value.clause;
 	      assert( false(cl, CL_ERASED) ||
-		      !ddi_is_garbage(ddi, start, cl)
+		      !ddi_is_garbage(ddi, start, tr_starts, cl)
 		    );
 	    }
 	  }
@@ -1148,7 +1161,7 @@ See also deleteActiveClauseFromIndexes() comment
 
 static void
 cleanClauseIndex(Definition def, ClauseList cl, ClauseIndex ci,
-		 DirtyDefInfo ddi, gen_t start)
+		 DirtyDefInfo ddi, gen_t start, Buffer tr_starts)
 { if ( cl->number_of_clauses < ci->resize_below )
   { deleteIndex(def, cl, ci);
   } else
@@ -1158,7 +1171,8 @@ cleanClauseIndex(Definition def, ClauseList cl, ClauseIndex ci,
 
       for(; n; n--, ch++)
       { if ( ch->dirty )
-	{ ci->size -= gcClauseBucket(def, ch, ch->dirty, ci->is_list, ddi, start);
+	{ ci->size -= gcClauseBucket(def, ch, ch->dirty, ci->is_list,
+				     ddi, start, tr_starts);
 	  if ( !ch->dirty && --ci->dirty == 0 )
 	    break;
 	}
@@ -1176,7 +1190,8 @@ references erased before generation `active` from the indexes.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
-cleanClauseIndexes(Definition def, ClauseList cl, DirtyDefInfo ddi, gen_t start)
+cleanClauseIndexes(Definition def, ClauseList cl, DirtyDefInfo ddi,
+		   gen_t start, Buffer tr_starts)
 { ClauseIndex *cip;
 
   if ( (cip=cl->clause_indexes) )
@@ -1185,7 +1200,7 @@ cleanClauseIndexes(Definition def, ClauseList cl, DirtyDefInfo ddi, gen_t start)
 
       if ( ISDEADCI(ci) )
 	continue;
-      cleanClauseIndex(def, cl, ci, ddi, start);
+      cleanClauseIndex(def, cl, ci, ddi, start, tr_starts);
     }
   }
 }
@@ -2335,6 +2350,7 @@ skipToTerm(Clause clause, const iarg_t *position)
 
     if ( an > 0 )
       pc = skipArgs(pc, an);
+  again:
     c = decode(*pc);
     switch(c)
     { case H_FUNCTOR:
@@ -2348,6 +2364,9 @@ skipToTerm(Clause clause, const iarg_t *position)
 	var[1] = encode(H_VOID);
 	return var;
       }
+      case I_CHP:
+	pc = stepPC(pc);
+        goto again;
 #if defined(O_DEBUG) || defined(O_MAINTENANCE)
       case H_FIRSTVAR:
       case H_VAR:
@@ -2522,8 +2541,8 @@ expected speedup is
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-bestHash(Word av, size_t ac, ClauseList clist, float min_speedup,
-	 hash_hints *hints, IndexContext ctx ARG_LD)
+bestHash(DECL_LD Word av, size_t ac, ClauseList clist, float min_speedup,
+	 hash_hints *hints, IndexContext ctx)
 { int i;
   assessment_set aset;
   hash_assessment *a;
@@ -2544,7 +2563,7 @@ bestHash(Word av, size_t ac, ClauseList clist, float min_speedup,
 
 					/* Step 1: find instantiated args */
   for(i=0; i<ac; i++)
-  { if ( canIndex(av[i] PASS_LD) )
+  { if ( canIndex(av[i]) )
       instantiated[ninstantiated++] = i;
   }
 
@@ -2791,8 +2810,9 @@ unify_clause_index(term_t t, ClauseIndex ci)
 }
 
 
+#define add_deep_indexes(ci, head, tail) LDFUNC(add_deep_indexes, ci, head, tail)
 static int
-add_deep_indexes(ClauseIndex ci, term_t head, term_t tail ARG_LD)
+add_deep_indexes(DECL_LD ClauseIndex ci, term_t head, term_t tail)
 { size_t i;
 
   for(i=0; i<ci->buckets; i++)
@@ -2814,7 +2834,7 @@ add_deep_indexes(ClauseIndex ci, term_t head, term_t tail ARG_LD)
 		 !unify_clause_index(head, ci) )
 	      return FALSE;
 	    if ( ci->is_list &&
-		 !add_deep_indexes(ci, head, tail PASS_LD) )
+		 !add_deep_indexes(ci, head, tail) )
 	      return FALSE;
 	  }
 	}
@@ -2829,7 +2849,7 @@ add_deep_indexes(ClauseIndex ci, term_t head, term_t tail ARG_LD)
 bool
 unify_index_pattern(Procedure proc, term_t value)
 { GET_LD
-  Definition def = getProcDefinition__LD(proc->definition PASS_LD);
+  Definition def = getProcDefinition(proc);
   ClauseIndex *cip;
   int rc = FALSE;
   int found = 0;
@@ -2850,7 +2870,7 @@ unify_index_pattern(Procedure proc, term_t value)
 	   !unify_clause_index(head, ci) )
 	goto out;
       if ( ci->is_list )
-      { if ( !add_deep_indexes(ci, head, tail PASS_LD) )
+      { if ( !add_deep_indexes(ci, head, tail) )
 	  goto out;
       }
     }
