@@ -45,7 +45,27 @@ option  parsing,  initialisation  and  handling  of errors and warnings.
 #include "pl-arith.h"
 #include "pl-zip.h"
 #include "pl-prof.h"
+#include "pl-read.h"
+#include "pl-prims.h"
+#include "pl-setup.h"
+#include "pl-fli.h"
+#include "pl-wic.h"
+#include "pl-pro.h"
+#include "pl-trace.h"
+#include "pl-proc.h"
+#include "pl-modul.h"
+#include "pl-flag.h"
+#include "pl-rec.h"
+#include "pl-term.h"
+#include "pl-funct.h"
+#include "pl-ext.h"
+#include "pl-srcfile.h"
+#include "pl-load.h"
+#include "pl-nt.h"
+#include "os/pl-prologflag.h"
 #include "os/pl-ctype.h"
+#include "os/pl-utf8.h"
+#include "os/pl-cstack.h"
 #include <errno.h>
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
@@ -400,14 +420,23 @@ memarea_limit(const char *s)
 }
 
 
+static int
+on_error_style(const char *s)
+{ return ( strcmp(s, "print") == 0 ||
+	   strcmp(s, "halt") == 0 ||
+	   strcmp(s, "status") == 0 );
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 When detected to run under a  GNU-Emacs   shell  or using M-x run-prolog
-from GNU-Emacs, don't pretend we can manipulate the TTY settings.
+from GNU-Emacs, don't pretend we  can   manipulate  the TTY settings. On
+Windows, do pretend we have a tty, so the prompt is displayed.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-setupGNUEmacsInferiorMode()
-{ char envbuf[4];
+setupGNUEmacsInferiorMode(void)
+{ char envbuf[80];
   char *s;
   int val;
 
@@ -418,6 +447,11 @@ setupGNUEmacsInferiorMode()
 
     clearPrologFlagMask(PLFLAG_TTY_CONTROL);
     val = TRUE;
+#ifdef __WINDOWS__
+    Sinput->flags  |= SIO_ISATTY;
+    Soutput->flags |= SIO_ISATTY;
+    Serror->flags  |= SIO_ISATTY;
+#endif
   } else
   { val = FALSE;
   }
@@ -457,7 +491,11 @@ initPaths(int argc, const char **argv)
 #endif
   }
 
-  systemDefaults.startup = store_string(PrologPath(DEFSTARTUP, plp, sizeof(plp)));
+#ifdef DEFSTARTUP
+  systemDefaults.startup = store_string(DEFSTARTUP);
+#else
+  systemDefaults.startup = NULL;
+#endif
 
 #ifdef O_XOS
   if ( systemDefaults.home )
@@ -506,6 +544,8 @@ cleanupPaths(void)
   cleanupStringP(&GD->options.topLevel);
   cleanupStringP(&GD->options.initFile);
   cleanupStringP(&GD->options.saveclass);
+  cleanupStringP(&GD->options.on_error);
+  cleanupStringP(&GD->options.on_warning);
   cleanupStringP(&GD->os.myhome);
 #ifdef __WINDOWS__
   cleanupStringP(&GD->paths.module);
@@ -546,7 +586,8 @@ initDefaults(void)
   else
     setPrologFlagMask(PLFLAG_TTY_CONTROL);
 
-  setPrologFlagMask(PLFLAG_DEBUGINFO|PLFLAG_GCTHREAD);
+  setPrologFlagMask(PLFLAG_DEBUGINFO);
+  setPrologFlagMask(PLFLAG_GCTHREAD);
 }
 
 
@@ -611,6 +652,8 @@ initDefaultOptions(void)
   GD->options.initFile	       = store_string(systemDefaults.startup);
   GD->options.scriptFiles      = NULL;
   GD->options.saveclass	       = store_string("none");
+  GD->options.on_error	       = store_string("print");
+  GD->options.on_warning       = store_string("print");
 
   if ( systemDefaults.goal )
     opt_append(&GD->options.goals, systemDefaults.goal);
@@ -684,6 +727,12 @@ parseCommandLineOptions(int argc0, char **argv0, char **argvleft, int compile)
 	    GD->options.silent = TRUE;
 	} else
 	  return -1;
+      } else if ( (rc=is_bool_opt(s, "debug_on_interrupt", &b)) )
+      { if ( rc == TRUE )
+	{ if ( b )
+	    setPrologFlagMask(PLFLAG_DEBUG_ON_INTERRUPT);
+	} else
+	  return -1;
       } else if ( (rc=is_bool_opt(s, "debug", &b)) )
       { if ( rc == TRUE )
 	{ if ( !b )
@@ -695,6 +744,7 @@ parseCommandLineOptions(int argc0, char **argv0, char **argvleft, int compile)
 	{ if ( !b )
 	  { clearPrologFlagMask(PLFLAG_SIGNALS);
 	    clearPrologFlagMask(PLFLAG_GCTHREAD);
+	    GD->options.nosignals = TRUE;
 	  }
 	} else
 	  return -1;
@@ -767,6 +817,16 @@ parseCommandLineOptions(int argc0, char **argv0, char **argvleft, int compile)
 #endif
       } else if ( (optval=is_longopt(s, "dump-runtime-variables")) )
       { GD->options.config = store_string(optval);
+      } else if ( (optval=is_longopt(s, "on-error")) )
+      { if ( on_error_style(optval) )
+	  GD->options.on_error = store_string(optval);
+	else
+	  return -1;
+      } else if ( (optval=is_longopt(s, "on-warning")) )
+      { if ( on_error_style(optval) )
+	  GD->options.on_warning = store_string(optval);
+	else
+	  return -1;
       } else if ( !compile )
       { argvleft[argcleft++] = argv[0];
       }
@@ -939,6 +999,53 @@ PL_is_initialised(int *argc, char ***argv)
 }
 
 
+int
+PL_winitialise(int argc, wchar_t **wargv)
+{ char **argv;
+  tmp_buffer b;
+  int i;
+  size_t sz;
+  char *abuf, *op;
+
+  if ( !(argv = malloc((argc+1)*sizeof(*argv))) )
+    return FALSE;
+
+  initBuffer(&b);
+  for(i=0; i<argc; i++)
+  { wchar_t *w;
+
+    for(w = wargv[i]; *w; w++)
+    { if ( *w < 0x7f )
+      { char c = *w;
+	addBuffer(&b, c, char);
+      } else
+      { char wb[6];
+	char *e;
+
+	e = utf8_put_char(wb, *w);
+	addMultipleBuffer(&b, wb, e-wb, char);
+      }
+    }
+    addBuffer(&b, 0, char);
+  }
+
+  sz = entriesBuffer(&b, char);
+  if ( !(abuf = malloc(sz)) )
+    return FALSE;
+  memcpy(abuf, baseBuffer(&b, char), sz);
+  discardBuffer(&b);
+  op = abuf;
+
+  for(i=0; i<argc; i++)
+  { argv[i] = op;
+    op += strlen(op)+1;
+  }
+  argv[i] = NULL;
+
+  return PL_initialise(argc, argv);
+}
+
+
 int
 PL_initialise(int argc, char **argv)
 { int n;
@@ -949,6 +1056,11 @@ PL_initialise(int argc, char **argv)
   if ( GD->initialised )
     succeed;
 
+  /* Initialize debug flag early, if first argument */
+  if (argc > 2 && strcmp(argv[1],"-d") == 0)
+    /* One's complement tells p_d_f_s not to bail on error, just return */
+    prolog_debug_from_string(argv[2], ~TRUE);
+
   initAlloc();
   initPrologThreads();			/* initialise thread system */
   SinitStreams();
@@ -1042,6 +1154,11 @@ PL_initialise(int argc, char **argv)
   initialiseForeign(GD->cmdline.os_argc, /* PL_initialise_hook() functions */
 		    GD->cmdline.os_argv);
   setAccessLevel(ACCESS_LEVEL_SYSTEM);
+  if ( GD->options.nosignals )
+  { GET_LD
+    clearPrologFlagMask(PLFLAG_SIGNALS);
+    clearPrologFlagMask(PLFLAG_GCTHREAD);
+  }
 
   if ( GD->bootsession )
   { IOSTREAM *s = SopenZIP(GD->resources.DB, "$prolog/state.qlf", RC_WRONLY);
@@ -1147,11 +1264,14 @@ usage(void)
     "    -s file                  Script source file\n",
     "    -p alias=path            Define file search path 'alias'\n",
     "    -O                       Optimised compilation\n",
+    "    --on-error=style         One of print, halt or status\n",
+    "    --on-warning=style       One of print, halt or status\n",
     "    --tty[=bool]             (Dis)allow tty control\n",
     "    --packs[=bool]           Do (not) attach add-ons\n",
     "    --signals[=bool]         Do (not) modify signal handling\n",
     "    --threads[=bool]         Do (not) allow for threads\n",
     "    --debug[=bool]           Do (not) generate debug info\n",
+    "    --debug-on-interrupt[=bool] Trap the debugger on interrupt\n",
     "    --quiet[=bool] (-q)      Do (not) suppress informational messages\n",
     "    --traditional            Disable extensions of version 7\n",
     "    --home=DIR               Use DIR as SWI-Prolog home\n",
@@ -1222,7 +1342,8 @@ version(void)
 
 static int
 abi_version(void)
-{ setupProlog();
+{ initDefaultOptions();
+  setupProlog();
   Sprintf(PLPKGNAME "-abi-%d-%d-%08x-%08x\n",
 	  PL_FLI_VERSION,
 	  PL_QLF_LOADVERSION,
@@ -1273,7 +1394,7 @@ struct on_halt
   OnHalt	next;
 };
 
-void
+static void
 register_halt(OnHalt *where, halt_function f, void *arg)
 { if ( !GD->os.halting )
   { OnHalt h = allocHeapOrHalt(sizeof(struct on_halt));
@@ -1331,9 +1452,7 @@ with -DGC_DEBUG.
 
 int
 cleanupProlog(int rval, int reclaim_memory)
-{ GET_LD
-
-  if ( GD->cleaning != CLN_NORMAL )
+{ if ( GD->cleaning != CLN_NORMAL )
     return FALSE;
 
 #ifdef __WINDOWS__
@@ -1348,12 +1467,14 @@ cleanupProlog(int rval, int reclaim_memory)
   }
 
 #ifdef O_PLMT
-  if ( !LD )
+  if ( !GLOBAL_LD )
   { PL_thread_attach_engine(NULL);
-    LD = GLOBAL_LD;
-    if ( !LD )
-      goto emergency;
   }
+  GET_LD
+  if ( !LD )
+    goto emergency;
+#else
+  GET_LD
 #endif
 
   GD->cleaning = CLN_PROLOG;
@@ -1391,7 +1512,8 @@ cleanupProlog(int rval, int reclaim_memory)
   resetProfiler();			/* don't do profiling anymore */
 #endif
 #ifdef O_PLMT
-  exitPrologThreads();
+  if ( !exitPrologThreads() )		/* reclaim memory while a thread */
+    reclaim_memory = FALSE;		/* runs is likely to crash */
 
 emergency:
 #endif
@@ -1429,7 +1551,7 @@ emergency:
 #endif
 
   if ( reclaim_memory )
-  { freeStacks(PASS_LD1);
+  { freeStacks();
 #ifdef O_PLMT
     cleanupLocalDefinitions(LD);
 #endif
@@ -1558,14 +1680,20 @@ vsysError(const char *fm, va_list args)
     unblockSignals(&LD->gc.saved_sigmask);
   }
 
+  Sfprintf(Serror, "\n");
+  save_backtrace("SYSERROR");
+  print_backtrace_named("SYSERROR");
+
 #if defined(O_DEBUGGER)
   Sfprintf(Serror, "\n\nPROLOG STACK:\n");
   PL_backtrace(10, 0);
   Sfprintf(Serror, "]\n");
 #endif /*O_DEBUGGER*/
 
-  if ( GD->bootsession )
-    PL_abort_process();
+  if ( !(true(Sinput, SIO_ISATTY) &&
+	 true(Serror, SIO_ISATTY)) ||
+       GD->bootsession )
+    PL_abort_process();			/* non-interactive or booting */
 
 action:
 #ifdef HAVE_GETPID
