@@ -33,6 +33,11 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
+#include "pl-incl.h"
+
+#ifndef _PL_INIT_H
+#define _PL_INIT_H
+
 		/********************************
 		*           STRUCTURES		*
 		********************************/
@@ -76,22 +81,35 @@ typedef struct
   bool		silent;			/* -q: quiet operation */
   bool		traditional;		/* --traditional: no version 7 exts */
   bool		nothreads;		/* --no-threads */
+  bool		nosignals;		/* --no-signals */
+  char *	on_error;
+  char *	on_warning;
   int		xpce;			/* --no-pce */
 #ifdef __WINDOWS__
   bool		win_app;		/* --win_app: be Windows application */
 #endif
 } pl_options_t;
 
-COMMON(int)	opt_append(opt_list **l, const char *s);
+		 /*******************************
+		 *    FUNCTION DECLARATIONS	*
+		 *******************************/
+
+int		startProlog(int argc, char **argv);
+bool		sysError(const char *fm, ...);
+void		fatalError(const char *fm, ...) NORETURN;
+bool		warning(const char *fm, ...);
+void		vfatalError(const char *fm, va_list args) NORETURN;
+bool		vwarning(const char *fm, va_list args);
+int		cleanupProlog(int status, int reclaim);
+int		run_on_halt(OnHalt *handlers, int rval);
+int		setTraditional(void);
+int		opt_append(opt_list **l, const char *s);
 
 
 		/********************************
 		*           PARAMETERS		*
 		********************************/
 
-#ifndef DEFSTARTUP
-#define DEFSTARTUP "init.pl"
-#endif
 #ifndef SYSTEMHOME
 #define SYSTEMHOME "/usr/lib/swipl"
 #endif
@@ -122,3 +140,5 @@ COMMON(int)	opt_append(opt_list **l, const char *s);
 #define PLHOMEVAR_1	"SWI_HOME_DIR"
 #define PLHOMEVAR_2	"SWIPL"
 #define PLHOMEFILE	"swipl.home"
+
+#endif /*_PL_INIT_H*/
