@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2012, University of Amsterdam
+    Copyright (c)  1985-2020, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -44,7 +45,6 @@
 #include <winsock2.h>
 #include <windows.h>
 #include "os/SWI-Stream.h"
-#define PL_ARITY_AS_SIZE 1
 #include "SWI-Prolog.h"
 #include <signal.h>
 
@@ -54,10 +54,12 @@
 #ifndef O_ANSI_COLORS
 #define O_ANSI_COLORS 1
 #endif
+#ifndef HAVE_WMAIN
+#define HAVE_WMAIN 1
+#endif
 
 #else /* non-Windows version */
 
-#define PL_ARITY_AS_SIZE
 #include "SWI-Prolog.h"
 
 #endif
@@ -92,14 +94,36 @@ consoleHandlerRoutine(DWORD id)
 }
 #endif
 
+		 /*******************************
+		 *	      TCMALLOC		*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+If  we  link  the  main   program    against   an  alternative  malloc()
+implementation we better ensure the main   program  depends on malloc(),
+otherwise the linker may still  decide   to  put  the alternative malloc
+library further down in the  link   dependencies.  We should also ensure
+this dependency is not optimized away.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+void *_PL_tc_malloc_base;
+
+static void
+force_malloc_dependency(void)
+{ _PL_tc_malloc_base = malloc(1);
+}
 
 		 /*******************************
 		 *		MAIN		*
 		 *******************************/
 
-
 int
+#if HAVE_WMAIN
+#define PL_initialise PL_winitialise
+wmain(int argc, wchar_t **argv)
+#else
 main(int argc, char **argv)
+#endif
 {
 #if O_CTRLC
   main_thread_id = GetCurrentThreadId();
@@ -110,6 +134,8 @@ main(int argc, char **argv)
   PL_w32_wrap_ansi_console();	/* decode ANSI color sequences (ESC[...m) */
 #endif
 
+  force_malloc_dependency();
+
   if ( !PL_initialise(argc, argv) )
     PL_halt(1);
 
@@ -123,3 +149,13 @@ main(int argc, char **argv)
 }
 
 
+		 /*******************************
+		 *   AddressSanitizer support   *
+		 *******************************/
+
+#ifdef __SANITIZE_ADDRESS__
+const char*
+__asan_default_options()
+{ return "detect_leaks=0";
+}
+#endif
