@@ -34,8 +34,15 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-#include "pl-incl.h"
+#include "pl-modul.h"
 #include "pl-comp.h"
+#include "pl-util.h"
+#include "pl-proc.h"
+#include "pl-funct.h"
+#include "pl-srcfile.h"
+#include "pl-fli.h"
+#include "pl-prims.h"
+#include "pl-wam.h"
 #undef LD
 #define LD LOCAL_LD
 
@@ -64,8 +71,9 @@ unallocProcedureSymbol(void *name, void *value)
 }
 
 
+#define _lookupModule(name) LDFUNC(_lookupModule, name)
 static Module
-_lookupModule(atom_t name ARG_LD)
+_lookupModule(DECL_LD atom_t name)
 { Module m, super;
 
   if ( (m = lookupHTable(GD->tables.modules, (void*)name)) )
@@ -125,14 +133,14 @@ _lookupModule(atom_t name ARG_LD)
 
 
 Module
-lookupModule__LD(atom_t name ARG_LD)
+lookupModule(DECL_LD atom_t name)
 { Module m;
 
   if ( (m = lookupHTable(GD->tables.modules, (void*)name)) )
     return m;
 
   PL_LOCK(L_MODULE);
-  m = _lookupModule(name PASS_LD);
+  m = _lookupModule(name);
   PL_UNLOCK(L_MODULE);
 
   return m;
@@ -140,7 +148,7 @@ lookupModule__LD(atom_t name ARG_LD)
 
 
 Module
-isCurrentModule__LD(atom_t name ARG_LD)
+isCurrentModule(DECL_LD atom_t name)
 { return lookupHTable(GD->tables.modules, (void*)name);
 }
 
@@ -162,7 +170,7 @@ temporary modules.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Module
-acquireModule__LD(atom_t name ARG_LD)
+acquireModule(DECL_LD atom_t name)
 { Module m;
 
   PL_LOCK(L_MODULE);
@@ -174,6 +182,15 @@ acquireModule__LD(atom_t name ARG_LD)
   return m;
 }
 
+void
+acquireModulePtr(DECL_LD Module m)
+{ if ( m && m->class == ATOM_temporary )
+  { PL_LOCK(L_MODULE);
+    m->references++;
+    PL_UNLOCK(L_MODULE);
+  }
+}
+
 void
 releaseModule(Module m)
 { if ( m->class == ATOM_temporary )
@@ -181,6 +198,10 @@ releaseModule(Module m)
     if ( --m->references == 0 &&
 	 true(m, M_DESTROYED) )
     { unlinkSourceFilesModule(m);
+#ifdef O_PLMT
+      if ( m->wait )
+	free_wait_area(m->wait);
+#endif
       GD->statistics.modules--;
       PL_unregister_atom(m->name);
       unallocModule(m);
@@ -264,8 +285,8 @@ initModules(void)
 
     GD->tables.modules = newHTable(MODULEHASHSIZE);
     GD->tables.modules->free_symbol = unallocModuleSymbol;
-    GD->modules.system = _lookupModule(ATOM_system PASS_LD);
-    GD->modules.user   = _lookupModule(ATOM_user PASS_LD);
+    GD->modules.system = _lookupModule(ATOM_system);
+    GD->modules.user   = _lookupModule(ATOM_user);
   }
   PL_UNLOCK(L_MODULE);
 }
@@ -599,8 +620,9 @@ setSuperModule(Module m, Module s)
 }
 
 
+#define set_module(m, prop) LDFUNC(set_module, m, prop)
 static int
-set_module(Module m, term_t prop ARG_LD)
+set_module(DECL_LD Module m, term_t prop)
 { atom_t pname;
   size_t arity;
 
@@ -615,7 +637,7 @@ set_module(Module m, term_t prop ARG_LD)
       if ( !PL_get_atom_ex(arg, &mname) )
 	return FALSE;
 
-      return setSuperModule(m, _lookupModule(mname PASS_LD));
+      return setSuperModule(m, _lookupModule(mname));
     } else if ( pname == ATOM_class )
     { atom_t class;
 
@@ -674,13 +696,13 @@ PRED_IMPL("set_module", 1, set_module, PL_FA_TRANSPARENT)
   Word p;
   int rc;
 
-  if ( !(p=stripModuleName(valTermRef(A1), &mname PASS_LD)) )
+  if ( !(p=stripModuleName(valTermRef(A1), &mname)) )
     return FALSE;
-  *valTermRef(prop) = linkVal(p);
+  *valTermRef(prop) = linkValNoG(p);
 
   PL_LOCK(L_MODULE);
-  m = mname ? _lookupModule(mname PASS_LD) : MODULE_parse;
-  rc = set_module(m, prop PASS_LD);
+  m = mname ? _lookupModule(mname) : MODULE_parse;
+  rc = set_module(m, prop);
   PL_UNLOCK(L_MODULE);
 
   return rc;
@@ -721,7 +743,7 @@ return value is the plain term or NULL if `term` is cyclic.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Word
-stripModuleName(Word term, atom_t *name ARG_LD)
+stripModuleName(DECL_LD Word term, atom_t *name)
 { int depth = 100;
   deRef(term);
   atom_t nm = 0;
@@ -735,7 +757,7 @@ stripModuleName(Word term, atom_t *name ARG_LD)
     nm = *mp;
     term = argTermP(*term, 1);
     deRef(term);
-    if ( --depth == 0 && !is_acyclic(term PASS_LD) )
+    if ( --depth == 0 && !is_acyclic(term) )
     { term_t t = pushWordAsTermRef(term);
       PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_acyclic_term, t);
       popTermRef();
@@ -757,11 +779,11 @@ remaining term.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 Word
-stripModule(Word term, Module *module, int flags ARG_LD)
+stripModule(DECL_LD Word term, Module *module, int flags)
 { atom_t mname = 0;
   Word rc;
 
-  if ( (rc=stripModuleName(term, &mname PASS_LD)) )
+  if ( (rc=stripModuleName(term, &mname)) )
   { if ( mname )
     { if ( unlikely(flags&SM_NOCREATE) )
       { Module m;
@@ -895,8 +917,9 @@ PRED_IMPL("delete_import_module", 2, delete_import_module, 0)
 }
 
 
+#define get_existing_source_file(file, sfp) LDFUNC(get_existing_source_file, file, sfp)
 static int
-get_existing_source_file(term_t file, SourceFile *sfp ARG_LD)
+get_existing_source_file(DECL_LD term_t file, SourceFile *sfp)
 { SourceFile sf;
   atom_t a;
 
@@ -949,7 +972,7 @@ PRED_IMPL("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
 	return FALSE;
       }
 
-      if ( !get_existing_source_file(file, &sf PASS_LD) )
+      if ( !get_existing_source_file(file, &sf) )
 	return FALSE;			/* given, but non-existing file */
 
       if ( sf )
@@ -1061,7 +1084,7 @@ PRED_IMPL("$current_source_module", 1, current_source_module, 0)
   return PL_unify_atom(A1, LD->modules.source->name);
 }
 
-
+static
 PRED_IMPL("$set_source_module", 1, set_source_module, 0)
 { PRED_LD
   atom_t name;
@@ -1107,10 +1130,11 @@ typedef struct defm_target
 } defm_target;
 
 
+#define find_modules_with_defs(m, count, targets, tmp, mlist, l) LDFUNC(find_modules_with_defs, m, count, targets, tmp, mlist, l)
 static int
-find_modules_with_defs(Module m, int count, defm_target targets[],
+find_modules_with_defs(DECL_LD Module m, int count, defm_target targets[],
 		       term_t tmp, term_t mlist,
-		       int l ARG_LD)
+		       int l)
 { ListCell c;
   int i;
   int found = FALSE;
@@ -1151,7 +1175,7 @@ find_modules_with_defs(Module m, int count, defm_target targets[],
   for(c = m->supers; c; c=c->next)
   { Module s = c->value;
 
-    if ( !find_modules_with_defs(s, count, targets, tmp, mlist, l-1 PASS_LD) )
+    if ( !find_modules_with_defs(s, count, targets, tmp, mlist, l-1) )
       return FALSE;
   }
 
@@ -1187,9 +1211,9 @@ PRED_IMPL("$def_modules", 2, def_modules, PL_FA_TRANSPARENT)
   atom_t mname = 0;
   Word mp;
 
-  if ( !(mp=stripModuleName(valTermRef(A1), &mname PASS_LD)) )
+  if ( !(mp=stripModuleName(valTermRef(A1), &mname)) )
     return FALSE;
-  *valTermRef(ttail) = linkVal(mp);
+  *valTermRef(ttail) = linkValNoG(mp);
 
   if ( mname )
   { Module m2;
@@ -1214,7 +1238,7 @@ PRED_IMPL("$def_modules", 2, def_modules, PL_FA_TRANSPARENT)
   if ( !PL_get_nil_ex(ttail) )
     return FALSE;
 
-  if ( !find_modules_with_defs(m, tcount, targets, tmp, tail, 100 PASS_LD) )
+  if ( !find_modules_with_defs(m, tcount, targets, tmp, tail, 100) )
     return FALSE;
 
   return PL_unify_nil(tail);
@@ -1284,7 +1308,7 @@ declareModule(atom_t name, atom_t class, atom_t super,
     clearHTable(module->public);
   }
   if ( super )
-    rc = setSuperModule(module, _lookupModule(super PASS_LD));
+    rc = setSuperModule(module, _lookupModule(super));
 
   PL_UNLOCK(L_MODULE);
 
@@ -1342,8 +1366,9 @@ PRED_IMPL("$declare_module", 6, declare_module, 0)
 }
 
 
+#define unify_export_list(public, module) LDFUNC(unify_export_list, public, module)
 static int
-unify_export_list(term_t public, Module module ARG_LD)
+unify_export_list(DECL_LD term_t public, Module module)
 { term_t head = PL_new_term_ref();
   term_t list = PL_copy_term_ref(public);
   int rval = TRUE;
@@ -1417,7 +1442,7 @@ PRED_IMPL("$module_property", 2, module_property, 0)
     else
       fail;
   } else if ( pname == ATOM_exports )
-  { return unify_export_list(a, m PASS_LD);
+  { return unify_export_list(a, m);
   } else if ( pname == ATOM_class )
   { return PL_unify_atom(a, m->class);
   } else if ( pname == ATOM_size )
@@ -1452,8 +1477,9 @@ exportProcedure(Module module, Procedure proc)
   return TRUE;
 }
 
+#define export_pi1(pi, module) LDFUNC(export_pi1, pi, module)
 static int
-export_pi1(term_t pi, Module module ARG_LD)
+export_pi1(DECL_LD term_t pi, Module module)
 { functor_t fd;
   Procedure proc;
 
@@ -1474,8 +1500,9 @@ export_pi1(term_t pi, Module module ARG_LD)
   }
 }
 
+#define export_pi(pi, module, depth) LDFUNC(export_pi, pi, module, depth)
 static int
-export_pi(term_t pi, Module module, int depth ARG_LD)
+export_pi(DECL_LD term_t pi, Module module, int depth)
 { if ( !PL_strip_module(pi, &module, pi) )
     return FALSE;
 
@@ -1486,13 +1513,13 @@ export_pi(term_t pi, Module module, int depth ARG_LD)
       return PL_type_error("acyclic_term", pi);
 
     _PL_get_arg(1, pi, a1);
-    if ( !export_pi(a1, module, depth PASS_LD) )
+    if ( !export_pi(a1, module, depth) )
       return FALSE;
     PL_reset_term_refs(a1);
     _PL_get_arg(2, pi, pi);
   }
 
-  return export_pi1(pi, module PASS_LD);
+  return export_pi1(pi, module);
 }
 
 
@@ -1502,7 +1529,7 @@ PRED_IMPL("export", 1, export, PL_FA_TRANSPARENT)
 { PRED_LD
   Module module = NULL;
 
-  return export_pi(A1, module, 0 PASS_LD);
+  return export_pi(A1, module, 0);
 }
 
 
@@ -1612,8 +1639,9 @@ atomToImportStrength(atom_t a)
 }
 
 
+#define import(pred, strength) LDFUNC(import, pred, strength)
 static int
-import(term_t pred, term_t strength ARG_LD)
+import(DECL_LD term_t pred, term_t strength)
 { Module source = NULL;
   Module destination = contextModule(environment_frame);
   functor_t fd;
@@ -1731,14 +1759,14 @@ static
 PRED_IMPL("import", 1, import, PL_FA_TRANSPARENT)
 { PRED_LD
 
-  return import(A1, 0 PASS_LD);
+  return import(A1, 0);
 }
 
 static
 PRED_IMPL("$import", 2, import, PL_FA_TRANSPARENT)
 { PRED_LD
 
-  return import(A1, A2 PASS_LD);
+  return import(A1, A2);
 }
 
 /** '$destroy_module'(+Module) is det.
