@@ -3,7 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2017, University of Amsterdam
+    Copyright (c)  1985-2021, University of Amsterdam
+			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -32,7 +34,9 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-#include "pl-incl.h"
+#include "pl-op.h"
+#include "pl-fli.h"
+#include "pl-modul.h"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 In traditional Prolog, the operator table is global.  In modern systems
@@ -131,11 +135,11 @@ defOperator(Module m, atom_t name, int type, int priority, int force)
   int t = (type & OP_MASK);		/* OP_PREFIX, ... */
   int must_reg = FALSE;
 
-  DEBUG(7, Sdprintf(":- op(%d, %s, %s) in module %s\n",
-		    priority,
-		    PL_atom_chars(operatorTypeToAtom(type)),
-		    PL_atom_chars(name),
-		    PL_atom_chars(m->name)));
+  DEBUG(MSG_OPERATOR, Sdprintf(":- op(%d, %s, %s) in module %s\n",
+			       priority,
+			       PL_atom_chars(operatorTypeToAtom(type)),
+			       PL_atom_chars(name),
+			       PL_atom_chars(m->name)));
 
   assert(t>=OP_PREFIX && t<=OP_POSTFIX);
 
@@ -232,10 +236,11 @@ currentOperator(Module m, atom_t name, int kind, int *type, int *priority)
       *priority = op->priority[kind];
 
       DEBUG(MSG_OPERATOR,
-	    Sdprintf("currentOperator(%s) --> %s %d\n",
+	    Sdprintf("current_op(%s) --> %s %d (module %s)\n",
 		     PL_atom_chars(name),
 		     PL_atom_chars(operatorTypeToAtom(*type)),
-		     *priority));
+		     *priority,
+		     PL_atom_chars(m->name)));
 
       succeed;
     }
@@ -363,9 +368,9 @@ PRED_IMPL("op", 3, op, PL_FA_TRANSPARENT|PL_FA_ISO)
   }
 
   if ( !PL_get_atom_ex(type, &tp) )
-    fail;
+    return FALSE;
   if ( !PL_get_integer_ex(pri, &p) )
-    fail;
+    return FALSE;
   if ( !((p >= 0 && p <= OP_MAXPRIORITY) || (p == -1 && m != MODULE_user)) )
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_operator_priority, pri);
   if ( !(t = atomToOperatorType(tp)) )
@@ -377,17 +382,17 @@ PRED_IMPL("op", 3, op, PL_FA_TRANSPARENT|PL_FA_ISO)
   { term_t l = PL_copy_term_ref(name);
     term_t e = PL_new_term_ref();
 
-    while( PL_get_list(l, e, l) )
+    while( PL_get_list_ex(l, e, l) )
     { if ( !PL_get_atom(e, &nm) )
 	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, e);
       if ( !defOperator(m, nm, t, p, FALSE) )
 	return FALSE;
     }
-    if ( !PL_get_nil(l) )
-      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
+    if ( !PL_get_nil_ex(l) )
+      return FALSE;
   }
 
-  succeed;
+  return TRUE;
 }
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -477,10 +482,11 @@ typedef struct
 } op_enum;
 
 
+#define current_op(m, inherit, prec, type, name, h) LDFUNC(current_op, m, inherit, prec, type, name, h)
 static word
-current_op(Module m, int inherit,
+current_op(DECL_LD Module m, int inherit,
 	   term_t prec, term_t type, term_t name,
-	   control_t h ARG_LD)
+	   control_t h)
 { op_enum *e;
   Buffer b;
   int mx;
@@ -522,6 +528,11 @@ current_op(Module m, int inherit,
       e->index = 0;
 
       scanVisibleOperators(m, nm, p, t, b, inherit);
+      DEBUG(MSG_OPERATOR,
+	    if ( nm ) Sdprintf("Scanned for %s in %s: %d hits\n",
+			       PL_atom_chars(nm),
+			       PL_atom_chars(m->name),
+			       (int)entriesBuffer(b, opdef)));
       break;
     }
     case FRG_REDO:
@@ -571,17 +582,25 @@ current_op(Module m, int inherit,
   fail;
 }
 
+#define get_op_module(a3, name, m) LDFUNC(get_op_module, a3, name, m)
 static int
-get_op_module(term_t a3, term_t name, Module *m ARG_LD)
+get_op_module(DECL_LD term_t a3, term_t name, Module *m)
 { atom_t mname = 0;
   Word p;
+  word w;
 
-  if ( !(p=stripModuleName(valTermRef(a3), &mname PASS_LD)) )
+  if ( !(p=stripModuleName(valTermRef(a3), &mname)) ||
+       !(w=linkValG(p)) )
     return FALSE;
-  *valTermRef(name) = linkVal(p);
+  *valTermRef(name) = w;
 
-  if ( mname && !(*m=isCurrentModule(mname)) )
+  if ( *m && (*m)->name == mname )
+    return TRUE;
+  if ( mname && !(*m=acquireModule(mname)) )
+  { DEBUG(MSG_OPERATOR, Sdprintf("Could not acquire module %s\n",
+				 PL_atom_chars(mname)));
     *m = MODULE_user;
+  }
 
   return TRUE;
 }
@@ -591,16 +610,22 @@ static
 PRED_IMPL("current_op", 3, current_op,
 	  PL_FA_NONDETERMINISTIC|PL_FA_TRANSPARENT|PL_FA_ISO)
 { PRED_LD
-  Module m = MODULE_parse;
+  Module mp = MODULE_parse;
+  Module m  = mp;
   term_t name = A3;
+  word rc;
 
   if ( CTX_CNTRL != FRG_CUTTED )
   { if ( !(name = PL_new_term_ref()) ||
-	 !get_op_module(A3, name, &m PASS_LD) )
+	 !get_op_module(A3, name, &m) )
       return FALSE;
   }
 
-  return current_op(m, TRUE, A1, A2, name, PL__ctx PASS_LD);
+  rc = current_op(m, TRUE, A1, A2, name, PL__ctx);
+  if ( m != mp )
+    releaseModule(m);
+
+  return rc;
 }
 
 /** '$local_op'(?Precedence, ?Type, ?Name) is nondet.
@@ -619,11 +644,11 @@ PRED_IMPL("$local_op", 3, local_op, PL_FA_NONDETERMINISTIC|PL_FA_TRANSPARENT)
 
   if ( CTX_CNTRL != FRG_CUTTED )
   { if ( !(name = PL_new_term_ref()) ||
-	 !get_op_module(A3, name, &m PASS_LD) )
+	 !get_op_module(A3, name, &m) )
       return FALSE;
   }
 
-  return current_op(m, FALSE, A1, A2, name, PL__ctx PASS_LD);
+  return current_op(m, FALSE, A1, A2, name, PL__ctx);
 }
 
 
@@ -651,6 +676,8 @@ static const opdef operators[] = {
   OP(ATOM_colon,		 OP_XFY, 600),	/* : */
   OP(ATOM_prove,		 OP_FX,	 1200),	/* :- */
   OP(ATOM_prove,		 OP_XFX, 1200),
+  OP(ATOM_ssu_commit,		 OP_XFX, 1200), /* => */
+//OP(ATOM_ssu_choice,		 OP_XFX, 1200), /* ?=> */
   OP(ATOM_semicolon,		 OP_XFY, 1100),	/* ; */
   OP(ATOM_bar,			 OP_XFY, 1105),	/* | */
   OP(ATOM_smaller,		 OP_XFX, 700),	/* < */
@@ -694,6 +721,7 @@ static const opdef operators[] = {
   OP(ATOM_multifile,		 OP_FX,	 1150),	/* multifile */
   OP(ATOM_meta_predicate,	 OP_FX,	 1150),	/* meta_predicate */
   OP(ATOM_public,		 OP_FX,	 1150),	/* public */
+  OP(ATOM_table,		 OP_FX,	 1150),	/* table */
   OP(ATOM_xor,			 OP_YFX, 400),	/* xor */
 
   OP(NULL_ATOM,			 0,	 0)
