@@ -35,9 +35,9 @@
 #ifndef PL_PRIVITF_H_INCLUDED
 #define PL_PRIVITF_H_INCLUDED
 
-COMMON(int)	PL_get_char(term_t c, int *p, int eof);
-COMMON(int)	PL_unify_char(term_t chr, int c, int mode);
-COMMON(int)	PL_unify_predicate(term_t head, predicate_t pred, int how);
+int	PL_get_char(term_t c, int *p, int eof);
+int	PL_unify_char(term_t chr, int c, int mode);
+int	PL_unify_predicate(term_t head, predicate_t pred, int how);
 
 
 
@@ -71,19 +71,20 @@ typedef struct list_ctx
   Word gstore;
 } list_ctx;
 
+#define addAtomicList(ctx, value) LDFUNC(addAtomicList, ctx, value)
 static inline void
-addAtomicList__LD(list_ctx *ctx, word value ARG_LD)
+addAtomicList(DECL_LD list_ctx *ctx, word value)
 { ctx->gstore[0] = consPtr(&ctx->gstore[1], TAG_COMPOUND|STG_GLOBAL);
   ctx->gstore[1] = FUNCTOR_dot2;
   ctx->gstore[2] = value;
   ctx->gstore += 3;
 }
 
-#define addSmallIntList(ctx, i) addAtomicList__LD(ctx, consInt(i)    PASS_LD)
-#define addCharList(ctx, c)     addAtomicList__LD(ctx, codeToAtom(c) PASS_LD)
+#define addSmallIntList(ctx, i) addAtomicList(ctx, consInt(i))
+#define addCharList(ctx, c)     addAtomicList(ctx, codeToAtom(c))
 
-COMMON(int)	allocList(size_t maxcells, list_ctx *ctx);
-COMMON(int)	unifyList(term_t term, list_ctx *ctx);
-COMMON(int)	unifyDiffList(term_t head, term_t tail, list_ctx *ctx);
+int	allocList(size_t maxcells, list_ctx *ctx);
+int	unifyList(term_t term, list_ctx *ctx);
+int	unifyDiffList(term_t head, term_t tail, list_ctx *ctx);
 
 #endif /*PL_PRIVITF_H_INCLUDED*/
