@@ -36,14 +36,31 @@
 
 /*#define O_DEBUG 1*/
 
-#define GLOBAL SO_LOCAL			/* allocate global variables here */
-#include "pl-incl.h"
+#define GLOBAL			/* allocate global variables here */
+#include "pl-setup.h"
 #include "pl-comp.h"
 #include "pl-arith.h"
 #include "os/pl-cstack.h"
+#include "os/pl-ctype.h"
+#include "os/pl-prologflag.h"
 #include "pl-dbref.h"
 #include "pl-trie.h"
 #include "pl-tabling.h"
+#include "pl-undo.h"
+#include "pl-fli.h"
+#include "pl-funct.h"
+#include "pl-modul.h"
+#include "pl-rec.h"
+#include "pl-flag.h"
+#include "pl-ext.h"
+#include "pl-op.h"
+#include "pl-trace.h"
+#include "pl-read.h"
+#include "pl-wam.h"
+#include "pl-gc.h"
+#include "pl-proc.h"
+#include "pl-pro.h"
+#include "pl-gvar.h"
 #include <sys/stat.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -75,7 +92,7 @@ setupProlog(void)
   DEBUG(1, Sdprintf("Starting Heap Initialisation\n"));
 
 #ifdef O_LOGICAL_UPDATE
-  next_global_generation();
+  next_generation(NULL);
 #endif
 
   LD->critical = 0;
@@ -100,7 +117,7 @@ setupProlog(void)
   GD->combined_stack.gc		 = TRUE;
   GD->combined_stack.overflow_id = STACK_OVERFLOW;
 
-  initPrologLocalData(PASS_LD1);
+  initPrologLocalData();
 
   DEBUG(1, Sdprintf("Atoms ...\n"));
   initAtoms();
@@ -125,8 +142,8 @@ setupProlog(void)
   initFlags();
   DEBUG(1, Sdprintf("Foreign Predicates ...\n"));
   initBuildIns();
-  DEBUG(1, Sdprintf("TCMalloc binding ...\n"));
-  initTCMalloc();
+  DEBUG(1, Sdprintf("Malloc binding ...\n"));
+  initMalloc();
   DEBUG(1, Sdprintf("Operators ...\n"));
   initOperators();
   DEBUG(1, Sdprintf("GMP ...\n"));
@@ -157,7 +174,7 @@ setupProlog(void)
 
 
 void
-initPrologLocalData(ARG1_LD)
+initPrologLocalData(DECL_LD)
 {
 #ifdef O_LIMIT_DEPTH
   LD->depth_info.limit = DEPTH_NO_LIMIT;
@@ -171,6 +188,7 @@ initPrologLocalData(ARG1_LD)
 
 #ifdef O_PLMT
   simpleMutexInit(&LD->thread.scan_lock);
+  LD->transaction.gen_base = GEN_INFINITE;
 #endif
 
   updateAlerted(LD);
@@ -482,7 +500,8 @@ dispatch_signal(int sig, int sync)
   }
 
   if ( (LD->critical || (true(sh, PLSIG_SYNC) && !sync)) &&
-       !is_fatal_signal(sig) )
+       !is_fatal_signal(sig) &&
+       sig != SIGINT )
   { PL_raise(sig);			/* wait for better times! */
     return;
   }
@@ -495,7 +514,7 @@ dispatch_signal(int sig, int sync)
   }
 
   if ( !sync )
-    blockGC(0 PASS_LD);
+    blockGC(0);
   LD->signal.current = sig;
   LD->signal.is_sync = sync;
 
@@ -563,7 +582,7 @@ dispatch_signal(int sig, int sync)
   lTop = (LocalFrame)valTermRef(lTopSave);
 
   if ( !sync )
-    unblockGC(0 PASS_LD);
+    unblockGC(0);
 
 					/* we cannot return.  First try */
 					/* longjmp.  If that fails, crash */
@@ -695,21 +714,79 @@ terminate_handler(int sig)
 #endif
 }
 
+void
+terminate_on_signal(int signo)
+{ PL_signal(signo, terminate_handler);
+}
+
 static void
 initTerminationSignals(void)
 {
 #ifdef SIGTERM
-  PL_signal(SIGTERM, terminate_handler);
+  terminate_on_signal(SIGTERM);
 #endif
 #ifdef SIGABRT
-  PL_signal(SIGABRT, terminate_handler);
+  terminate_on_signal(SIGABRT);
 #endif
 #ifdef SIGQUIT
-  PL_signal(SIGQUIT, terminate_handler);
+  terminate_on_signal(SIGQUIT);
 #endif
 }
 #endif /*HAVE_SIGNAL*/
 
+#ifdef O_C_STACK_GUARDED
+static void
+alt_segv_handler(int sig)
+{ GET_LD
+  (void)sig;
+
+  DEBUG(MSG_SIGNAL,
+	Sdprintf("Got C-stack overflow; critical = %d\n",
+		 LD->signal.sig_critical));
+
+  if ( LD->signal.sig_critical )
+  { longjmp(LD->signal.context, TRUE);
+    /*NORETURN*/
+  }
+
+  sigCrashHandler(sig);
+}
+#endif
+
+int
+initGuardCStack(void)
+{
+#ifdef O_C_STACK_GUARDED
+  GET_LD
+  stack_t ss = {0};
+
+  if ( (LD->signal.alt_stack = malloc(SIGSTKSZ)) )
+  { ss.ss_sp = LD->signal.alt_stack;
+    ss.ss_size = SIGSTKSZ;
+    ss.ss_flags = 0;
+
+    if ( sigaltstack(&ss, NULL) == 0)
+    { DEBUG(MSG_SIGNAL, Sdprintf("Setup altstack (%zd bytes)\n", SIGSTKSZ));
+
+      if ( LD == &PL_local_data )	/* main thread, only need this once */
+      { struct sigaction sa = {0};
+
+	sa.sa_flags = SA_ONSTACK;
+	sa.sa_handler = alt_segv_handler;
+	sigemptyset(&sa.sa_mask);
+
+	if ( sigaction(SIGSEGV, &sa, NULL) == 0 )
+	{ DEBUG(MSG_SIGNAL, Sdprintf("Setup SEGV on altstack\n"));
+	  return TRUE;
+	}
+      }
+    }
+  }
+#endif
+
+  return FALSE;
+}
+
 static void
 agc_handler(int sig)
 { GET_LD
@@ -781,6 +858,7 @@ initSignals(void)
     set_sighandler(SIGPIPE, SIG_IGN);
 #endif
     initTerminationSignals();
+    initGuardCStack();
 #endif /*HAVE_SIGNAL*/
     initBackTrace();
     for( ; sn->name; sn++)
@@ -1061,15 +1139,16 @@ int
 PL_handle_signals(void)
 { GET_LD
 
-  if ( !is_signalled(PASS_LD1) )
+  if ( !is_signalled() )
     return 0;
   else
-    return handleSignals(PASS_LD1);
+    return handleSignals();
 }
 
 #ifndef __unix__
+#define handleSigInt(_) LDFUNC(handleSigInt, _)
 static int
-handleSigInt(ARG1_LD)
+handleSigInt(DECL_LD)
 { int intmask = 1<<(SIGINT-1);
 
   if ( LD->signal.forced == SIGINT && LD->signal.pending[0] & intmask )
@@ -1090,18 +1169,18 @@ handleSigInt(ARG1_LD)
 #endif
 
 int
-handleSignals(ARG1_LD)
+handleSignals(DECL_LD)
 { int done = 0;
   int i;
 
-  if ( !is_signalled(PASS_LD1) )
+  if ( !is_signalled() )
     return 0;
   if ( !HAS_LD )
     return 0;
   if ( exception_term )
     return -1;
 #ifndef __unix__				/* on Unix we ask to signal twice */
-  if ( (done=handleSigInt(PASS_LD1)) )
+  if ( (done=handleSigInt()) )
     return done;
 #endif
   if ( LD->critical )
@@ -1173,7 +1252,7 @@ PRED_IMPL("prolog_alert_signal", 2, prolog_alert_signal, 0)
 
 
 int
-endCritical__LD(ARG1_LD)
+f_endCritical(DECL_LD)
 { if ( exception_term )
     return FALSE;
 
@@ -1321,9 +1400,9 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
       succeed;
     }
 
-    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_signal_handler, sig);
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_signal_handler, new);
   } else
-    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_signal_handler, sig);
+    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_signal_handler, new);
 
   succeed;
 }
@@ -1416,6 +1495,7 @@ emptyStacks(void)
     LD->attvar.gc_attvars   = PL_new_term_ref();
     DEBUG(3, Sdprintf("attvar.tail at %p\n", valTermRef(LD->attvar.tail)));
 #endif
+    LD->undo.undo_list      = init_undo_list(); /* must be first.  See __do_undo() */
     LD->tabling.delay_list  = init_delay_list();
     LD->tabling.idg_current = PL_new_term_ref();
 #ifdef O_GVAR
@@ -1482,7 +1562,7 @@ allocStacks(void)
   if ( !gBase || !tBase || !aBase )
   { if ( gBase )
       *gBase++ = MARK_MASK;		/* compensate for freeStacks */
-    freeStacks(PASS_LD1);
+    freeStacks();
     return FALSE;
   }
 
@@ -1504,7 +1584,7 @@ allocStacks(void)
 
 
 void
-freeStacks(ARG1_LD)
+freeStacks(DECL_LD)
 { if ( gBase )
   { gBase--;
     stack_free(gBase);
@@ -1573,7 +1653,7 @@ trimStacks() reclaims all unused space on the stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
-trimStacks(int resize ARG_LD)
+trimStacks(DECL_LD int resize)
 { LD->trim_stack_requested = FALSE;
 
   if ( resize )
@@ -1605,7 +1685,7 @@ static
 PRED_IMPL("trim_stacks", 0, trim_stacks, 0)
 { PRED_LD
 
-  trimStacks(TRUE PASS_LD);
+  trimStacks(TRUE);
 
   succeed;
 }
@@ -1659,6 +1739,11 @@ freePrologLocalData(PL_local_data_t *ld)
     free_alloc_pool(ld->tabling.node_pool);
 
   clearThreadTablingData(ld);
+
+#ifdef O_C_STACK_GUARDED
+  if ( ld->signal.alt_stack )
+    free(ld->signal.alt_stack);
+#endif
 }
 
 
@@ -1676,7 +1761,7 @@ set_stack_limit(size_t limit)
                sizeStack(global) +
                sizeStack(trail) )
   { garbageCollect(GC_USER);
-    trimStacks(TRUE PASS_LD);
+    trimStacks(TRUE);
 
     if ( limit < sizeStack(local) +
 	         sizeStack(global) +
