@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2016, University of Amsterdam
+    Copyright (c)  1985-2020, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -33,12 +34,23 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
-#include "pl-incl.h"
+#include "pl-write.h"
+#include "pl-arith.h"
 #include "pl-dict.h"
+#include "pl-fli.h"
+#include "pl-read.h"
+#include "pl-attvar.h"
+#include "pl-op.h"
+#include "pl-wam.h"
+#include "pl-prims.h"
+#include "pl-modul.h"
+#include "pl-setup.h"
 #include <math.h>
 #include "os/pl-dtoa.h"
 #include "os/pl-ctype.h"
+#include "os/pl-prologflag.h"
 #include <stdio.h>			/* sprintf() */
+#include <errno.h>
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
 #endif
@@ -65,6 +77,13 @@ typedef struct
   term_t prec_opt;			/* term in write options with prec */
 } write_options;
 
+#if USE_LD_MACROS
+#define	enterPortray(_)		LDFUNC(enterPortray, _)
+#define	leavePortray(_)		LDFUNC(leavePortray, _)
+#endif /*USE_LD_MACROS*/
+
+#define LDFUNC_DECLARATIONS
+
 static bool	writeTerm2(term_t term, int prec,
 			   write_options *options, bool arg) WUNUSED;
 static bool	writeTerm(term_t t, int prec,
@@ -74,11 +93,13 @@ static bool	writeArgTerm(term_t t, int prec,
 static int	PutToken(const char *s, IOSTREAM *stream);
 static int	writeAtom(atom_t a, write_options *options);
 static int	callPortray(term_t arg, int prec, write_options *options);
-static int	enterPortray(ARG1_LD);
-static void	leavePortray(ARG1_LD);
+static int	enterPortray(void);
+static void	leavePortray(void);
+
+#undef LDFUNC_DECLARATIONS
 
 char *
-var_name_ptr__LD(Word p, char *name ARG_LD)
+var_name_ptr(DECL_LD Word p, char *name)
 { size_t iref;
 
   deRef(p);
@@ -138,8 +159,9 @@ Return:	TRUE:  processes
 	-1:    error
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define writeNumberVar(t, options) LDFUNC(writeNumberVar, t, options)
 static int
-writeNumberVar(term_t t, write_options *options ARG_LD)
+writeNumberVar(DECL_LD term_t t, write_options *options)
 { Word p = valTermRef(t);
   Functor f;
 
@@ -161,7 +183,7 @@ writeNumberVar(term_t t, write_options *options ARG_LD)
 
   p = &f->arguments[0];
   deRef(p);
-  if ( isInteger(*p) )
+  if ( isTaggedInt(*p) || isBignum(*p) )
   { int64_t n = valInteger(*p);
     char buf[32];			/* Max is H354745078340568300 */
 
@@ -220,6 +242,16 @@ wr_is_symbol(int c, write_options *options)
 	    (options->flags & PL_WRT_BACKQUOTE_IS_SYMBOL)) );
 }
 
+static int
+code_requires_quoted(int c, IOSTREAM *fd, int flags)
+{ if ( c > 0x7f && (flags&PL_WRT_QUOTE_NON_ASCII) )
+    return TRUE;
+  if ( fd && Scanrepresent(c, fd) != 0 )
+    return TRUE;
+
+  return FALSE;
+}
+
 static int
 atomType(atom_t a, write_options *options)
 { Atom atom = atomValue(a);
@@ -227,6 +259,7 @@ atomType(atom_t a, write_options *options)
   size_t len = atom->length;
   IOSTREAM *fd = options ? options->out : NULL;
   Module m = options ? options->module : MODULE_user;
+  int flags = options ? options->flags : 0;
 
   if ( len == 0 )
     return AT_QUOTE;
@@ -234,7 +267,7 @@ atomType(atom_t a, write_options *options)
   if ( isLower(*s) || (true(m, M_VARPREFIX) && isAlpha(*s)) )
   { do
     { for( ++s;
-	   --len > 0 && isAlpha(*s) && (!fd || Scanrepresent(*s, fd)==0);
+	   --len > 0 && isAlpha(*s) && !code_requires_quoted(*s, fd, flags);
 	   s++)
 	;
     } while ( len >= 2 &&
@@ -256,7 +289,7 @@ atomType(atom_t a, write_options *options)
 
     for( ;
 	 left > 0 && wr_is_symbol(*s, options) &&
-	 (!fd || Scanrepresent(*s, fd)==0);
+	 !code_requires_quoted(*s, fd, flags);
 	 s++, left--)
       ;
     if ( left > 0 )
@@ -290,17 +323,20 @@ unquoted_atomW(atom_t atom, IOSTREAM *fd, int flags)
   if ( !f_is_prolog_atom_start(*s) )
   { for( ; len > 0; s++, len--)
     { if ( !f_is_prolog_symbol(*s) ||
-	   (fd && Scanrepresent(*s, fd)<0) )
+	   code_requires_quoted(*s, fd, flags) )
 	return FALSE;
     }
     return TRUE;
   }
 
+  if ( code_requires_quoted(*s, fd, flags) )
+    return FALSE;
+
   do
   { for( ++s;
 	 ( --len > 0 &&
 	   f_is_prolog_identifier_continue(*s) &&
-	   (!fd || Scanrepresent(*s, fd)==0)
+	   !code_requires_quoted(*s, fd, flags)
 	 );
 	 s++)
       ;
@@ -338,7 +374,7 @@ unquoted_atom(atom_t a)
 
 static bool
 Putc(int c, IOSTREAM *s)
-{ return Sputcode(c, s) == EOF ? FALSE : TRUE;
+{ return Sputcode(c, s) != EOF;
 }
 
 
@@ -411,7 +447,7 @@ needSpace(int c, IOSTREAM *s)
   if ( s->lastc == EOF )
     return FALSE;
 
-  if ( (s->lastc&C_PREFIX_SIGN) && (isDigit(c) || isSymbolW(c)) )
+  if ( (s->lastc&C_PREFIX_SIGN) && (isDigit(c) || f_is_prolog_symbol(c)) )
     return TRUE;
   if ( (s->lastc&C_PREFIX_OP) && ( c == '(' || c == '{' ) )
     return TRUE;				/* avoid op(...) */
@@ -420,8 +456,9 @@ needSpace(int c, IOSTREAM *s)
 
   s->lastc &= ~C_MASK;
 
-  if ( ((isAlphaW(s->lastc) && isAlphaW(c)) ||
-	(isSymbolW(s->lastc) && isSymbolW(c)) ||
+  if ( ((f_is_prolog_identifier_continue(s->lastc) &&
+	 f_is_prolog_identifier_continue(c)) ||
+	(f_is_prolog_symbol(s->lastc) && f_is_prolog_symbol(c)) ||
 	(c == '(' && !isPunctW(s->lastc)) ||
 	(c == '\'' && (isDigit(s->lastc))) ||
 	(isquote(c) && s->lastc == c)
@@ -501,11 +538,11 @@ PutCloseBrace(IOSTREAM *s)
 static bool
 putQuoted(int c, int quote, int flags, IOSTREAM *stream)
 { if ( (flags & PL_WRT_CHARESCAPES) )
-  { if ( c == ' ' ||
-	 (!(c < 0xff && !isGraph(c)) && c != quote && c != '\\') )
+  { if ( !unicode_quoted_escape(c) && c != quote && c != '\\' &&
+	 (!stream || Scanrepresent(c, stream) == 0) )
     { TRY(Putc(c, stream));
     } else
-    { char esc[8];
+    { char esc[10];			/* Longest is UXXXXXXXX */
 
       esc[1] = EOS;
 
@@ -538,10 +575,14 @@ putQuoted(int c, int quote, int flags, IOSTREAM *stream)
 	    esc[0] = '\\';
 	    break;
 	  default:
-	    if ( c <= 0xff )
-	      Ssprintf(esc, "%03o\\", c);
-	    else
-	      assert(0);			/* to be done */
+	    if ( (flags & PL_WRT_CHARESCAPES_UNICODE) )
+	    { if ( c <= 0xffff )
+		Ssprintf(esc, "u%04X", c);
+	      else
+		Ssprintf(esc, "U%08X", c);
+	    } else
+	    { Ssprintf(esc, "x%X\\", c);
+	    }
 	}
       }
       if ( !Putc('\\', stream) ||
@@ -596,7 +637,7 @@ writeAttVar(term_t av, write_options *options)
 
     Sputcode('{', options->out);
     a = PL_new_term_ref();
-    PL_get_attr__LD(av, a PASS_LD);
+    PL_get_attr(av, a);
     if ( !writeTerm(a, 1200, options) )
       return FALSE;
     Sputcode('}', options->out);
@@ -614,9 +655,9 @@ writeAttVar(term_t av, write_options *options)
       pred = _PL_predicate("portray_attvar", 1, "$attvar",
 			   &GD->procedures.portray_attvar1);
 
-    if ( !enterPortray(PASS_LD1) )
+    if ( !enterPortray() )
       return FALSE;
-    if ( !saveWakeup(&wstate, TRUE PASS_LD) )
+    if ( !saveWakeup(&wstate, TRUE) )
       return FALSE;
     old = Scurout;
     Scurout = options->out;
@@ -624,8 +665,8 @@ writeAttVar(term_t av, write_options *options)
     if ( rc != TRUE && !PL_exception(0) )
       rc = TRUE;
     Scurout = old;
-    restoreWakeup(&wstate PASS_LD);
-    leavePortray(PASS_LD1);
+    restoreWakeup(&wstate);
+    leavePortray();
 
     return rc;
   }
@@ -812,7 +853,9 @@ static int
 writeString(term_t t, write_options *options)
 { GET_LD
   PL_chars_t txt;
+  int rc = TRUE;
 
+  PL_STRINGS_MARK();
   PL_get_text(t, &txt, CVT_STRING);
 
   if ( true(options, PL_WRT_QUOTED) )
@@ -824,26 +867,33 @@ writeString(term_t t, write_options *options)
     else
       quote = '"';
 
-    TRY(Putc(quote, options->out));
+    if ( !(rc=Putc(quote, options->out)) )
+      goto out;
 
     for(i=0; i<txt.length; i++)
     { int chr = get_chr_from_text(&txt, i);
 
-      TRY(putQuoted(chr, quote, options->flags, options->out));
+      if ( !(rc=putQuoted(chr, quote, options->flags, options->out)) )
+	goto out;
     }
 
-    return Putc(quote, options->out);
+    rc = Putc(quote, options->out);
   } else
   { unsigned int i;
 
     for(i=0; i<txt.length; i++)
     { int chr = get_chr_from_text(&txt, i);
 
-      TRY(Putc(chr, options->out));
+      if ( !(rc=Putc(chr, options->out)) )
+	break;
     }
   }
+  PL_STRINGS_RELEASE();
 
-  succeed;
+out:
+  PL_free_text(&txt);
+
+  return rc;
 }
 
 #endif /*O_STRING*/
@@ -973,7 +1023,7 @@ format_special_float(double f, char *buf)
       return writeNaN(f, buf);
     case FP_NINF:
     case FP_PINF:
-      return writeINF(n, buf);
+      return writeINF(f, buf);
     case FP_NDENORM:		/* pos/neg denormalized non-zero */
     case FP_PDENORM:
     case FP_NNORM:			/* pos/neg normalized non-zero */
@@ -1083,10 +1133,44 @@ format_float(double f, char *buf)
   return buf;
 }
 
+#ifdef O_GMP
+#define writeMPZ(mpz, options) LDFUNC(writeMPZ, mpz, options)
+static int
+writeMPZ(DECL_LD mpz_t mpz, write_options *options)
+{ char tmp[1024];
+  char *buf;
+  size_t sz = mpz_sizeinbase(mpz, 10) + 2;
+  int rc;
+
+  if ( sz <= sizeof(tmp) )
+    buf = tmp;
+  else
+    buf = PL_malloc(sz);
+
+  /* mpz_get_str() can perform large intermediate allocations */
+  EXCEPTION_GUARDED({ LD->gmp.persistent++;
+		      mpz_get_str(buf, 10, mpz);
+		      LD->gmp.persistent--;
+		    },
+		    { LD->gmp.persistent--;
+		      rc = PL_rethrow();
+		    })
+  rc = PutToken(buf, options->out);
+  if ( buf != tmp )
+    PL_free(buf);
+
+  return rc;
+}
+#endif
 
 static int
 WriteNumber(Number n, write_options *options)
-{ switch(n->type)
+{
+#ifdef O_GMP
+  GET_LD
+#endif
+
+  switch(n->type)
   { case V_INTEGER:
     { char buf[32];
 
@@ -1095,30 +1179,17 @@ WriteNumber(Number n, write_options *options)
     }
 #ifdef O_GMP
     case V_MPZ:
-    { GET_LD
-      char tmp[1024];
-      char *buf;
-      size_t sz = mpz_sizeinbase(n->value.mpz, 10) + 2;
-      bool rc;
+      return writeMPZ(n->value.mpz, options);
+    case V_MPQ:
+    { mpz_t num, den;			/* num/den */
+      char sep = true(options, PL_WRT_RAT_NATURAL) ? '/' : 'r';
 
-      if ( sz <= sizeof(tmp) )
-	buf = tmp;
-      else
-	buf = PL_malloc(sz);
-
-      /* mpz_get_str() can perform large intermediate allocations :-( */
-      EXCEPTION_GUARDED({ LD->gmp.persistent++;
-			  mpz_get_str(buf, 10, n->value.mpz);
-			  LD->gmp.persistent--;
-			},
-			{ LD->gmp.persistent--;
-			  rc = PL_rethrow();
-			})
-      rc = PutToken(buf, options->out);
-      if ( buf != tmp )
-	PL_free(buf);
-
-      return rc;
+      num[0] = *mpq_numref(n->value.mpq);
+      den[0] = *mpq_denref(n->value.mpq);
+      return ( writeMPZ(num, options) &&
+	       Sputcode(sep, options->out) != EOF &&
+	       (options->out->lastc = EOF) &&
+	       writeMPZ(den, options) );
     }
 #endif
     case V_FLOAT:
@@ -1170,8 +1241,9 @@ writePrimitive(term_t t, write_options *options)
 }
 
 
+#define pl_nl(stream) LDFUNC(pl_nl, stream)
 static int
-pl_nl__LD(term_t stream ARG_LD)
+pl_nl(DECL_LD term_t stream)
 { IOSTREAM *s;
 
   if ( getTextOutputStream(stream, &s) )
@@ -1187,14 +1259,14 @@ static
 PRED_IMPL("nl", 1, nl, PL_FA_ISO)
 { PRED_LD
 
-  return pl_nl__LD(A1 PASS_LD);
+  return pl_nl(A1);
 }
 
 static
 PRED_IMPL("nl", 0, nl, PL_FA_ISO)
 { PRED_LD
 
-  return pl_nl__LD(0 PASS_LD);
+  return pl_nl(0);
 }
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1236,7 +1308,7 @@ put_write_options(term_t opts_in, write_options *options)
 
 
 static int
-enterPortray(ARG1_LD)
+enterPortray(DECL_LD)
 { if ( LD->IO.portray_nesting >= MAX_PORTRAY_NESTING )
     return PL_resource_error("portray_nesting");
   LD->IO.portray_nesting++;
@@ -1245,7 +1317,7 @@ enterPortray(ARG1_LD)
 
 
 static void
-leavePortray(ARG1_LD)
+leavePortray(DECL_LD)
 { LD->IO.portray_nesting--;
 }
 
@@ -1264,7 +1336,7 @@ callPortray(term_t arg, int prec, write_options *options)
   { pred = _PL_predicate("call", 3, "user", &GD->procedures.call3);
   } else
   { pred = _PL_predicate("portray", 1, "user", &GD->procedures.portray);
-    if ( !pred->definition->impl.any )
+    if ( !pred->definition->impl.any.defined )
       return FALSE;
   }
 
@@ -1274,9 +1346,9 @@ callPortray(term_t arg, int prec, write_options *options)
     int rval;
     term_t av;
 
-    if ( !enterPortray(PASS_LD1) )
+    if ( !enterPortray() )
       return -1;
-    if ( !saveWakeup(&wstate, TRUE PASS_LD) )
+    if ( !saveWakeup(&wstate, TRUE) )
       return -1;
     Scurout = options->out;
     if ( options->portray_goal )
@@ -1294,8 +1366,8 @@ callPortray(term_t arg, int prec, write_options *options)
     if ( !rval && PL_exception(0) )
       rval = -1;
     Scurout = old;
-    restoreWakeup(&wstate PASS_LD);
-    leavePortray(PASS_LD1);
+    restoreWakeup(&wstate);
+    leavePortray();
 
     return rval;
   }
@@ -1345,7 +1417,7 @@ writeList(term_t list, write_options *options)
   term_t head = PL_new_term_ref();
   term_t l    = PL_copy_term_ref(list);
 
-  if ( false(options, PL_WRT_DOTLISTS) )
+  if ( false(options, PL_WRT_DOTLISTS|PL_WRT_NO_LISTS) )
   { TRY(Putc('[', options->out));
     for(;;)
     { PL_get_list(l, head, l);
@@ -1370,8 +1442,15 @@ writeList(term_t list, write_options *options)
 
     for(;;)
     { PL_get_list(l, head, l);
-      if ( !PutToken(".", options->out) ||
-	   !Putc('(', options->out) ||
+      if ( true(options, PL_WRT_DOTLISTS) )
+      { if ( !PutToken(".", options->out) )
+	  return FALSE;
+      } else
+      { if ( !writeAtom(ATOM_dot, options) )
+	  return FALSE;
+      }
+
+      if ( !Putc('(', options->out) ||
 	   !writeArgTerm(head, 999, options, TRUE) ||
 	   !PutComma(options) )
 	return FALSE;
@@ -1410,8 +1489,9 @@ writeList(term_t list, write_options *options)
 }
 
 
+#define isBlockOp(t, arg, functor) LDFUNC(isBlockOp, t, arg, functor)
 static int
-isBlockOp(term_t t, term_t arg, atom_t functor ARG_LD)
+isBlockOp(DECL_LD term_t t, term_t arg, atom_t functor)
 { if ( functor == ATOM_nil || functor == ATOM_curl )
   { _PL_get_arg(1, t, arg);
     if ( (functor == ATOM_nil  && PL_is_pair(arg)) ||
@@ -1422,9 +1502,10 @@ isBlockOp(term_t t, term_t arg, atom_t functor ARG_LD)
   return FALSE;
 }
 
+#define writeDictPair(name, value, last, closure) LDFUNC(writeDictPair, name, value, last, closure)
 
 static int
-writeDictPair(term_t name, term_t value, int last, void *closure)
+writeDictPair(DECL_LD term_t name, term_t value, int last, void *closure)
 { write_options *options = closure;
 
   if ( writeTerm(name, 1200, options) &&
@@ -1472,7 +1553,7 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
   { return writePrimitive(t, options);
   } else
   { if ( true(options, PL_WRT_NUMBERVARS|PL_WRT_VARNAMES) )
-    { switch( writeNumberVar(t, options PASS_LD) )
+    { switch( writeNumberVar(t, options) )
       { case -1:
 	  return FALSE;
 	case TRUE:
@@ -1523,7 +1604,7 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 	return FALSE;
 
       if ( arity == 1 ||
-	  (arity == 2 && isBlockOp(t, arg, functor PASS_LD)) )
+	  (arity == 2 && isBlockOp(t, arg, functor)) )
       {
 					  /* op <term> */
 	if ( currentOperator(options->module, functor, OP_PREFIX,
@@ -1571,7 +1652,11 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 	  if ( arity == 1 )
 	  { TRY(writeAtom(functor, options));
 	  } else
-	  { _PL_get_arg(1, t, arg);
+	  { if ( functor == ATOM_curl &&
+		 (PL_is_atom(arg) || PL_is_variable(arg)) )
+	      TRY(Putc(' ', out));
+	    _PL_get_arg(1, t, arg);
+
 	    TRY(writeTerm(arg, 1200, options));
 	  }
 	  if (op_pri > prec)
@@ -1580,7 +1665,7 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 	  succeed;
 	}
       } else if ( arity == 2 ||
-		 (arity == 3 && isBlockOp(t, arg, functor PASS_LD)) )
+		 (arity == 3 && isBlockOp(t, arg, functor)) )
       {					  /* <term> op <term> */
 	if ( currentOperator(options->module, functor, OP_INFIX,
 			     &op_type, &op_pri) )
@@ -1696,7 +1781,10 @@ writeTopTerm(term_t term, int prec, write_options *options)
   Slock(options->out);
   if ( (!(options->flags&PL_WRT_NO_CYCLES) && options->max_depth) ||
        PL_is_acyclic(term) )
-  { rc = writeTerm(term, prec, options);
+  { C_STACK_OVERFLOW_GUARDED(
+	rc,
+	writeTerm(term, prec, options),
+        (void)0);
   } else
   { fid_t fid;
     term_t template, substitutions, cycles, at_term;
@@ -1716,7 +1804,10 @@ writeTopTerm(term_t term, int prec, write_options *options)
 			  PL_TERM, template,
 			  PL_TERM, cycles) )
       return FALSE;
-    rc = writeTerm(at_term, prec, options);
+    C_STACK_OVERFLOW_GUARDED(
+	rc,
+	writeTerm(at_term, prec, options),
+	(void)0);
     PL_discard_foreign_frame(fid);
   }
   Sunlock(options->out);
@@ -1725,8 +1816,9 @@ writeTopTerm(term_t term, int prec, write_options *options)
 }
 
 
+#define bind_varnames(names) LDFUNC(bind_varnames, names)
 static int
-bind_varnames(term_t names ARG_LD)
+bind_varnames(DECL_LD term_t names)
 { term_t tail, head, var, namet;
   int check_cycle_after = 1000;
 
@@ -1797,28 +1889,31 @@ writeBlobMask(atom_t a)
 
 
 static const opt_spec write_term_options[] =
-{ { ATOM_quoted,	    OPT_BOOL },
-  { ATOM_ignore_ops,	    OPT_BOOL },
-  { ATOM_dotlists,	    OPT_BOOL },
-  { ATOM_brace_terms,	    OPT_BOOL },
-  { ATOM_numbervars,        OPT_BOOL },
-  { ATOM_portray,           OPT_BOOL },
-  { ATOM_portrayed,         OPT_BOOL },
-  { ATOM_portray_goal,      OPT_TERM },
-  { ATOM_character_escapes, OPT_BOOL },
-  { ATOM_max_depth,	    OPT_INT  },
-  { ATOM_module,	    OPT_ATOM },
-  { ATOM_back_quotes,	    OPT_ATOM },
-  { ATOM_attributes,	    OPT_ATOM },
-  { ATOM_priority,	    OPT_INT },
-  { ATOM_partial,	    OPT_BOOL },
-  { ATOM_spacing,	    OPT_ATOM },
-  { ATOM_blobs,		    OPT_ATOM },
-  { ATOM_cycles,	    OPT_BOOL },
-  { ATOM_variable_names,    OPT_TERM },
-  { ATOM_nl,		    OPT_BOOL },
-  { ATOM_fullstop,	    OPT_BOOL },
-  { NULL_ATOM,		    0 }
+{ { ATOM_quoted,		    OPT_BOOL },
+  { ATOM_quote_non_ascii,	    OPT_BOOL },
+  { ATOM_ignore_ops,		    OPT_BOOL },
+  { ATOM_dotlists,		    OPT_BOOL },
+  { ATOM_brace_terms,		    OPT_BOOL },
+  { ATOM_numbervars,		    OPT_BOOL },
+  { ATOM_portray,		    OPT_BOOL },
+  { ATOM_portrayed,		    OPT_BOOL },
+  { ATOM_portray_goal,		    OPT_TERM },
+  { ATOM_character_escapes,	    OPT_BOOL },
+  { ATOM_character_escapes_unicode, OPT_BOOL },
+  { ATOM_max_depth,		    OPT_INT  },
+  { ATOM_module,		    OPT_ATOM },
+  { ATOM_back_quotes,		    OPT_ATOM },
+  { ATOM_attributes,		    OPT_ATOM },
+  { ATOM_priority,		    OPT_INT },
+  { ATOM_partial,		    OPT_BOOL },
+  { ATOM_spacing,		    OPT_ATOM },
+  { ATOM_blobs,			    OPT_ATOM },
+  { ATOM_cycles,		    OPT_BOOL },
+  { ATOM_variable_names,	    OPT_TERM },
+  { ATOM_nl,			    OPT_BOOL },
+  { ATOM_fullstop,		    OPT_BOOL },
+  { ATOM_no_lists,		    OPT_BOOL },
+  { NULL_ATOM,			    0 }
 };
 
 foreign_t
@@ -1827,12 +1922,14 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   bool quoted     = FALSE;
   bool ignore_ops = FALSE;
   bool dotlists   = FALSE;
-  bool braceterms = FALSE;
+  bool braceterms = TRUE;
   bool numbervars = -1;			/* not set */
   bool portray    = FALSE;
   term_t gportray = 0;
   atom_t bq       = 0;
   bool charescape = -1;			/* not set */
+  bool charescape_unicode = -1;
+  bool quote_non_ascii = FALSE;
   atom_t mname    = ATOM_user;
   atom_t attr     = ATOM_nil;
   atom_t blobs    = ATOM_nil;
@@ -1841,6 +1938,7 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   bool cycles     = TRUE;
   bool nl         = FALSE;
   bool fullstop   = FALSE;
+  bool no_lists   = FALSE;
   term_t varnames = 0;
   int local_varnames;
   IOSTREAM *s = NULL;
@@ -1851,11 +1949,13 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   options.spacing = ATOM_standard;
 
   if ( !scan_options(opts, 0, ATOM_write_option, write_term_options,
-		     &quoted, &ignore_ops, &dotlists, &braceterms,
+		     &quoted, &quote_non_ascii, &ignore_ops, &dotlists, &braceterms,
 		     &numbervars, &portray, &portray, &gportray,
-		     &charescape, &options.max_depth, &mname,
+		     &charescape, &charescape_unicode,
+		     &options.max_depth, &mname,
 		     &bq, &attr, &priority, &partial, &options.spacing,
-		     &blobs, &cycles, &varnames, &nl, &fullstop) )
+		     &blobs, &cycles, &varnames, &nl, &fullstop,
+		     &no_lists) )
     fail;
 
   if ( attr == ATOM_nil )
@@ -1900,6 +2000,11 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   if ( charescape == TRUE ||
        (charescape == -1 && true(options.module, M_CHARESCAPE)) )
     options.flags |= PL_WRT_CHARESCAPES;
+  if ( charescape_unicode == TRUE ||
+       (charescape_unicode == -1 && truePrologFlag(PLFLAG_CHARESCAPE_UNICODE)) )
+    options.flags |= PL_WRT_CHARESCAPES_UNICODE;
+  if ( true(options.module, RAT_NATURAL) )
+    options.flags |= PL_WRT_RAT_NATURAL;
   if ( gportray )
   { options.portray_goal = gportray;
     if ( !put_write_options(opts, &options) ||
@@ -1911,13 +2016,15 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   if ( numbervars == -1 )
     numbervars = (portray ? TRUE : FALSE);
 
-  if ( quoted )     options.flags |= PL_WRT_QUOTED;
-  if ( ignore_ops ) options.flags |= PL_WRT_IGNOREOPS;
-  if ( dotlists )   options.flags |= PL_WRT_DOTLISTS;
-  if ( braceterms ) options.flags |= PL_WRT_BRACETERMS;
-  if ( numbervars ) options.flags |= PL_WRT_NUMBERVARS;
-  if ( portray )    options.flags |= PL_WRT_PORTRAY;
-  if ( !cycles )    options.flags |= PL_WRT_NO_CYCLES;
+  if ( quoted )          options.flags |= PL_WRT_QUOTED;
+  if ( quote_non_ascii ) options.flags |= PL_WRT_QUOTE_NON_ASCII;
+  if ( ignore_ops )      options.flags |= PL_WRT_IGNOREOPS;
+  if ( dotlists )        options.flags |= PL_WRT_DOTLISTS;
+  if ( !braceterms )     options.flags |= PL_WRT_BRACETERMS;
+  if ( numbervars )      options.flags |= PL_WRT_NUMBERVARS;
+  if ( portray )         options.flags |= PL_WRT_PORTRAY;
+  if ( !cycles )         options.flags |= PL_WRT_NO_CYCLES;
+  if ( no_lists )        options.flags |= PL_WRT_NO_LISTS;
   if ( bq )
   { unsigned int flags = 0;
 
@@ -1933,7 +2040,7 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
 
   BEGIN_NUMBERVARS(local_varnames);
   if ( varnames )
-  { if ( (rc=bind_varnames(varnames PASS_LD)) )
+  { if ( (rc=bind_varnames(varnames)) )
       options.flags |= PL_WRT_VARNAMES;
     else
       goto out;
@@ -1944,10 +2051,12 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   options.out = s;
   if ( !partial )
     PutOpenToken(EOF, s);		/* reset this */
-  if ( (options.flags & PL_WRT_QUOTED) && !(s->flags&SIO_REPPL) )
-  { s->flags |= SIO_REPPL;
+  if ( (options.flags & PL_WRT_QUOTED) && !(s->flags&(SIO_REPPL|SIO_REPPLU)) )
+  { unsigned int flag = truePrologFlag(PLFLAG_CHARESCAPE_UNICODE) ? SIO_REPPLU
+								  : SIO_REPPL;
+    s->flags |= flag;
     rc = writeTopTerm(term, priority, &options);
-    s->flags &= ~SIO_REPPL;
+    s->flags &= ~flag;
   } else
   { rc = writeTopTerm(term, priority, &options);
   }
@@ -2071,10 +2180,11 @@ pl_write_canonical2(term_t stream, term_t term)
   options.singletons = PL_is_acyclic(term);
   options.numbered_check = FALSE;
 
-  rc = ( numberVars(term, &options, 0 PASS_LD) != NV_ERROR &&
+  rc = ( numberVars(term, &options, 0) != NV_ERROR &&
 	 do_write2(stream, term,
-		   PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS|
-		   PL_WRT_NODOTINATOM, TRUE)
+		   PL_WRT_QUOTED|PL_WRT_QUOTE_NON_ASCII|
+		   PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS|
+		   PL_WRT_NODOTINATOM|PL_WRT_BRACETERMS, TRUE)
        );
 
   END_NUMBERVARS(TRUE);
@@ -2126,21 +2236,36 @@ PRED_IMPL("$put_token", 2, put_token, 0)
   fail;
 }
 
-/** '$put_quoted_codes'(+Stream, +Quote, +Codes, +Options)
+/** '$put_quoted'(+Stream, +Quote, +Codes, +Options)
 
 Emit Codes using the escaped character  syntax,   but  does not emit the
-start and end-code itself. Options is  currently ignored. It is intended
-to provide additional preferences, so as using \uXXXX, \UXXXXXXXX, etc.
+start and end-code itself.
 */
 
+static const opt_spec put_quoted_options[] =
+{ { ATOM_character_escapes_unicode, OPT_BOOL },
+  { NULL_ATOM,			    0 }
+};
+
+
 static
 PRED_IMPL("$put_quoted", 4, put_quoted_codes, 0)
-{ IOSTREAM *out;
+{ PRED_LD
+  IOSTREAM *out;
   pl_wchar_t *w;
   size_t i, len;
   int quote;
   int flags = PL_WRT_CHARESCAPES;
   int rc = TRUE;
+  bool charescape_unicode = -1;
+
+  if ( !scan_options(A4, 0, ATOM_write_option, put_quoted_options,
+		     &charescape_unicode) )
+    return FALSE;
+
+  if ( charescape_unicode == TRUE ||
+       (charescape_unicode == -1 && truePrologFlag(PLFLAG_CHARESCAPE_UNICODE)) )
+    flags |= PL_WRT_CHARESCAPES_UNICODE;
 
   if ( !PL_get_stream_handle(A1, &out) ||
        !PL_get_char_ex(A2, &quote, FALSE) ||
@@ -2172,7 +2297,9 @@ Swrite_lss(void *handle, char *buf, size_t size)
   (void)buf;
 
   if ( lss->stream->position->charno > lss->length )
+  { errno = EINVAL;
     return -1;
+  }
 
   return size;
 }
